#include "AnalysisPipeline.h"
#include "RendererGLUT.h"
#include "VisWindow.h"
#include "ParamParser.h"
#include "ModelDatabase.h"

void DownloadAllModels(ParamParser & params);

void ProcessIndividualModels(ParamParser & params);

void CreateFinalDB(ParamParser & params);

VKString RenamedModel(std::map<VKString, int> & nameCounters,
					  VKString modelName);

int main(int argc, char** argv )
{
	bool valid;
	if (argc<2)
	{
		std::cout<<"Usage: "<<argv[0]<<" settings.txt [optionalSettings]"<<std::endl;
		exit(0);
	}
	
	const char ** constArgv=(const char**)argv;
	std::cout<<"Trying to read: "<<argv[1]<<std::endl;
	ParamParser params(argv[1], &(constArgv[2]), argc-2);
	
	VKString mode = params.GetStrValue("PruneDB", "Process");
	if (mode=="DownloadAllModels")
		DownloadAllModels(params);
	else if (mode=="ProcessIndividualModels")
		ProcessIndividualModels(params);
	else if (mode=="CreateFinalDB")
		CreateFinalDB(params);
	return 0;
}

void DownloadAllModels(ParamParser & params)
{
	assert(false);
}

void ProcessIndividualModels(ParamParser & params)
{
	assert(false);
}

void CreateFinalDB(ParamParser & params)
{
	bool testOnlyNoWrite = false;
	double DUPLICATE_THRESHOLD = 0.9999;
	bool renameModels = true;;
	std::map<VKString, int> prefixToCounter;
#define SHD_PREC float
	VKString dbdir = params.GetStrValue("PruneDB", "INPUT_Dir");
	VKString dbFile = params.GetStrValue("PruneDB", "INPUT_DBFile");
	VKString dbFilePath = dbdir+"/"+dbFile;
	std::cout<<"Reading: "<<dbFilePath.c_str()<<std::endl;
	assert(ModelDatabase::FileExists(dbFilePath));
	ParamParser dbParams(dbFilePath, NULL, 0);
	
	int nShapeDescr = 0;
	SHD_PREC * aveShapeDescriptor = NULL;
	
	std::vector<SHD_PREC*> perShapeDescriptors;
	std::vector<VKString> modelNames; 
	std::vector<VKString> invalidNames;
	
	std::cout<<"Finding Mean Shape: "<<std::endl;
	std::map<VKString, std::map<VKString, VKStringList> >::iterator iter;
		
	for (iter=dbParams.m_ParamModules.begin(); 
		 iter != dbParams.m_ParamModules.end(); iter++)
	{
		VKString modelName = iter->second["ModelName"][0];
		
		VKString shdFile = dbdir+"/Models/"+modelName+"/Geometry/"+modelName+".shd";
		FILE * file = fopen(shdFile.c_str(), "rb");
		if (file==NULL)
		{
			std::cout<<"[WARNING] No descriptor: "<<shdFile.c_str()<<std::endl;
			invalidNames.push_back(modelName);
			continue;
		}
		
		int nDescr;
		fread(&nDescr, sizeof(int), 1, file);
		assert(nDescr>0);
		if (aveShapeDescriptor==NULL)
		{
			aveShapeDescriptor = new SHD_PREC[nDescr];
			nShapeDescr = nDescr;
			for (int i=0; i<nDescr; i++)
				aveShapeDescriptor[i] = 0;
		}
		assert(nShapeDescr == nDescr);
		
		SHD_PREC * currDescr = new SHD_PREC[nDescr];
		fread(currDescr, sizeof(SHD_PREC), nDescr, file);
		fclose(file);
		
		for (int i=0; i<nDescr; i++)
			aveShapeDescriptor[i] += currDescr[i];
		
		modelNames.push_back(modelName);		
		perShapeDescriptors.push_back(currDescr);
	}
	
	assert(perShapeDescriptors.size()>0 && aveShapeDescriptor!=NULL);
	for (int i=0; i<nShapeDescr; i++)
		aveShapeDescriptor[i] /= (int)perShapeDescriptors.size();
	
	std::vector<Sortable> sortedDescriptors;
		
	// distance to mean?
	if (testOnlyNoWrite)
	{
		std::cout<<"AveShape=[ ";
		for (int i=0; i<nShapeDescr; i++)
			std::cout<<aveShapeDescriptor[i]<<" "<<std::flush;
		std::cout<<std::endl;
		std::cout<<"Distances to average: "<<std::endl;
	}
	for (int i=0; i<(int)perShapeDescriptors.size(); i++)
	{
		SHD_PREC * descr=perShapeDescriptors[i];
		double dist = 0;
		for (int j=0; j<nShapeDescr; j++)
			dist += pow((double)(aveShapeDescriptor[j] - descr[j]), 2.);
		
		if (testOnlyNoWrite)
			std::cout<<"\t"<<modelNames[i].c_str()<<" : "<<dist<<std::endl;
		
		sortedDescriptors.push_back(Sortable(dist, NULL, i));
	}
	
//	// average distance to other
//	std::cout<<"PerModelAnaysis: "<<std::flush;
//	for (int i=0; i<(int)perShapeDescriptors.size(); i++)
//	{
//		if (i%100==0)
//			std::cout<<i<<" "<<std::flush;
//		SHD_PREC * descr=perShapeDescriptors[i];
//		SHD_PREC corrVal = 0;
//		for (int j=0; j<(int)perShapeDescriptors.size(); j++)
//		{
//			if (i==j)
//				continue;
//			SHD_PREC * descr2=perShapeDescriptors[j];
//			
//			for (int k=0; k<nShapeDescr; k++)
//				corrVal += descr2[k]*descr[k];
//		}
//		sortedDescriptors.push_back(Sortable(-corrVal, NULL, i));
//	}
//	std::cout<<std::endl;
		
	std::sort(sortedDescriptors.begin(), sortedDescriptors.end());
	
	int nShapesToKeep = params.GetIntValue("CreateFinalDB", "INPUT_NKeep");
	
	std::cout<<"Writing closest to mean: "<<std::endl;	
	int takeFirst = vkMin(nShapesToKeep, (int)sortedDescriptors.size());
	std::ofstream textStream;
	if (!testOnlyNoWrite)
	{
		textStream.open(dbFilePath.c_str());
		assert(textStream.is_open());
	}
	VKStringList mvPrevNames;
	VKStringList mvNewNames;
	if (testOnlyNoWrite)
		std::cout<<"Closest Shapes("<<takeFirst<<"):"<<std::endl;
	for (int i=0; i<takeFirst; i++)
	{
		if (i>=sortedDescriptors.size())
			break;
		int id = sortedDescriptors[i].id;

		assert(id>=0 && id < (int)modelNames.size());
		VKString modelName = modelNames[id];
		VKString modelModule = VKString("Model")+VKString::number(i)+":";
		
		if (testOnlyNoWrite)
		{
			std::cout<<"\t["<<i<<"] "<<modelName.c_str();
			std::cout<<" : "<<sortedDescriptors[i].value<<std::endl;
		}
		
		bool sameAsPrev = false;
		if (i>0)
		{
			SHD_PREC * prevDescr = perShapeDescriptors[sortedDescriptors[i-1].id];
			SHD_PREC * currDescr = perShapeDescriptors[id];
			double corrVal = 0;
			for (int j=0; j<nShapeDescr; j++)
				corrVal += prevDescr[j]*currDescr[j];
			if (testOnlyNoWrite)
			{
				std::cout<<"[TEST] Corr["<<modelNames[id].c_str()<<", ";
				std::cout<<modelNames[sortedDescriptors[i-1].id].c_str();
				std::cout<<"] = "<<corrVal<<std::endl;
			}
			if (corrVal>DUPLICATE_THRESHOLD)
				sameAsPrev = true;
		}
		
		if (sameAsPrev)
		{
			std::cout<<"Excluding Dupl: "<<modelName.c_str()<<std::endl;
			invalidNames.push_back(modelName);
			sortedDescriptors.erase(sortedDescriptors.begin()+(i));
			i--;
			continue; 
		}
		
		// write new database.txt
		if (!testOnlyNoWrite)
		{
			VKString prevName = modelName;
			if (renameModels)
				modelName = RenamedModel(prefixToCounter, modelName);
			if (prevName!=modelName)
			{
				mvPrevNames.push_back(prevName);
				mvNewNames.push_back(modelName);
			}
			textStream<<modelModule.c_str()<<"\n";
			textStream<<"\tModelName\t"<<modelName.c_str()<<"\n";
			textStream<<"\tModelID\t"<<i<<"\n";
		}
	}
	if (!testOnlyNoWrite)
		textStream.close();
	
	// delete models
	for (int i=takeFirst; i<(int)sortedDescriptors.size(); i++)
	{
		// delete model
		int id = sortedDescriptors[i].id;
		assert(id>=0 && id < (int)modelNames.size());
		VKString modelName = modelNames[id];
		invalidNames.push_back(modelName);
	}
	
	for (int i=0; i<(int)invalidNames.size(); i++)
	{
		VKString delFile = (dbdir+"/Models/"+invalidNames[i]);
//		if (testOnlyNoWrite)
//			std::cout<<"[TEST] Would be deleting: \n\t"<<delFile.c_str()<<std::endl;
//		else
		if (!testOnlyNoWrite)
			system((VKString("rm -r ") + delFile).c_str());	
	}
	
	if (!testOnlyNoWrite)
	{
		for (int i=0; i<mvPrevNames.count(); i++)
		{
			 VKString mvDir1 = dbdir+"/Models/"+mvPrevNames[i];
			 VKString mvDir2 = dbdir+"/Models/_"+mvPrevNames[i];
			 system((VKString("mv ")+mvDir1+" "+mvDir2).c_str());	
		}
		
		assert(mvPrevNames.count()==mvNewNames.count());
		for (int i=0; i<mvPrevNames.count(); i++)
		{
			// copy directory
			VKString mvDir1 = dbdir+"/Models/_"+mvPrevNames[i];
			VKString mvDir2 = dbdir+"/Models/"+mvNewNames[i];
			system((VKString("mv ")+mvDir1+" "+mvDir2).c_str());
					
			// geometry file
			VKString geom1 = mvDir2+"/Geometry/"+mvPrevNames[i]+".ply";
			VKString geom2 = mvDir2+"/Geometry/"+mvNewNames[i]+".ply";
			system((VKString("mv ")+geom1+" "+geom2).c_str());

			// shd file
			VKString shd1 = mvDir2+"/Geometry/"+mvPrevNames[i]+".shd";
			VKString shd2 = mvDir2+"/Geometry/"+mvNewNames[i]+".shd";
			system((VKString("mv ")+shd1+" "+shd2).c_str());
			
			// *model3d file
			VKString m3dFile1 = mvDir2+"/"+mvPrevNames[i]+".model3d";
			VKString m3dFile2 = mvDir2+"/"+mvNewNames[i]+".model3d";
			system((VKString("mv ")+m3dFile1+" "+m3dFile2).c_str());

			std::ofstream textStream(m3dFile2.c_str());
			assert(textStream.is_open());
			
			textStream<<"ModelData:\n";
			textStream<<"\tGeometry\t"<<geom2.c_str()<<"\n";
			textStream<<"\tGeomType\tIRREGULAR_POLYGON_SOUP"<<"\n\n";
			
			textStream.close();
		}
	}
	
	for (int i=0; i<(int)perShapeDescriptors.size(); i++)
		delete [] perShapeDescriptors[i];
	
	delete [] aveShapeDescriptor;
}

VKString RenamedModel(std::map<VKString, int> & nameCounters,
					  VKString modelName)
{
	VKString prefix = modelName.left(modelName.length()-4);
	std::map<VKString, int>::iterator iter = nameCounters.find(prefix);
	int counter = 0;
	if (iter==nameCounters.end())
	{
		nameCounters[prefix] = 1;
		counter = 0;
	}
	else
	{
		counter = iter->second;
		iter->second++;
	}
	VKString intStr = VKString::number(counter);
	while (intStr.length() < 4)
		intStr = VKString("0") + intStr;
	assert(intStr.length() == 4);
	return prefix + intStr;
}
