/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __AnalysisPipeline_H
#define __AnalysisPipeline_H

#include "VKString.h"
#include "ParamParser.h"
#include "TimeProfiler.h"
#include <vector>
#include <iostream>
#include <assert.h>
#include "CollectionDiffuse.h"

class DatabaseVisWindow;
class ModelDatabase;
class PipelineStepApplication;
class CollectionDiffuse;

class AnalysisPipeline
{
	public:
		friend class PipelineStepApplication;
	
		AnalysisPipeline(ParamParser & params);
		virtual ~AnalysisPipeline();
		
		void RunPipeline();
		bool Process(const VKString & moduleName);
		
		void DatabaseLoader(const VKString & moduleName);
		void ShapeDescriptorLoader(const VKString & moduleName);
		void PointDescriptorLoader(const VKString & moduleName);
		void PointSetLoader(const VKString & moduleName);
		void DistanceLoader(const VKString & moduleName);
		void CorrCollectionLoader(const VKString & moduleName,
								  bool initGuessExists = false,
								  const VKString & iterCollectionName = "",
								  const VKString & inputCorrsName = "",
								  const VKString & inputCorrsValue = "", 
								  bool forceProcess = false);	
	
		void CreatePairwiseCollection(const VKString & moduleName, 
									  const VKString & collName);

		VKString CorrSinglePairLoader(const VKString & modelFrom, 
									  const VKString & modelTo, 
									  const VKString & moduleName, 
									  int * nIDsMin=NULL, int * nIDsMax=NULL,
									  bool initGuessExists = false,
									  const VKString & iterCollectionName = "",
									  const VKString & initGuessCorrs = "", 
									  const VKString & initGuessCorrVals = "");
		
		void CorrsIterative(const VKString & moduleName);
	
		void GetModelPairs(const VKString & moduleName, VKStringList & modelPairs);
		void FillClosestPairs(const VKString & moduleName, VKStringList & modelPairs);
	
		void ExecuteApp(const VKString & moduleName);
		PipelineStepApplication * GetPipelineApp(const VKString & appName);
		
//		void CorrCollectionLoaderExternal(ParamParser & collectionFile, 
//										  const VKString & mapName,
//										  const VKString & collectionName,
//										  const VKString & collectionModuleName,
//										  const VKString & corrsDir);

		void CorrValueLoader(const VKString & moduleName, 
							 bool anotherCollection = false, 
							 const VKString & anotherCollectionName="");
		VKString CorrValueLoader(const VKString & collectionName, 
								 int corrID,
								 const VKString & moduleName);
		
//		void AddCorrValue(const VKString & moduleName,
//						  const VKString & collectionName);
	
		void DatabaseSaver(const VKString & moduleName);
	
		virtual CollectionDiffuse::CorrMtxParams ReadCorrMtxParams(const VKString & moduleName);
		
		//virtual void Draw();
		virtual void InitRenderingParams(DatabaseVisWindow * window, ParamParser * drawparams);
	
		ModelDatabase * GetDB();
		ModelDatabase * GetMainDB();
		void AddDB(const VKString & dbName, ModelDatabase * db);
		ModelDatabase * GetAdditionalDB(int id);
		int NumAdditionalDBs();
		ModelDatabase * GetAdditionalDBByName(const VKString & dbName);
		void ChangeDB(bool next);
		void SwitchToDB(const VKString & dbName);
		void SwitchToMainDB();
		
		ParamParser & m_Params;
	
		VKStringList m_sAppNames;
    
        virtual ParamParser & RenderingParams();
	
		static void Pause(const VKString & msg);
		static void WaitForFile(const VKString & fileStr, 
								int checkEveySeconds, 
								int maxSeconds);
	
	protected:
		int m_iSelectedDB;
		std::vector<ModelDatabase * > m_AdditionalDBs;
		std::map<VKString, int > m_DBNameToID;
	
		std::map<VKString, PipelineStepApplication *> m_Apps;
	
		VKStringList m_ProcessList;
	
		ParamParser * m_pDrawParams;
		DatabaseVisWindow * m_pWindow;
	
		ModelDatabase * m_pDatabase;
		VKString m_sDatabaseName;
		bool valid;
	
		TimeProfiler m_Timer;
};

#endif

