/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __AppBenchmark_H
#define __AppBenchmark_H

#include "PipelineStepApplication.h"
#include "AnalysisPipeline.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "ModelValueLabel.h"

class AppBenchmark : public PipelineStepApplication
{
	public:
		enum BenchmarkType
		{
			BENCHMARK_NONE,
			BENCHMARK_CORRS
		};
	
		enum CorrsEvaluationType
		{
			EVAL_CORRS_NONE,
			EVAL_CORRS_VAL_AS_PENALTY
		};
		
		struct PerMapStats
		{
			PerMapStats();
			std::vector<double> m_EvalCorrValues;
			
			bool m_bFilled;
			PointSet * m_pPntsFrom;
			Model3D * m_pModelFrom;
			Model3D * m_pModelTo;
		};
	
		enum BenchDraw
		{
			BENCHDRAW_NONE,
			BENCHDRAW_ERROR,
			
		};
	
		AppBenchmark(AnalysisPipeline & pipeline, const VKString & moduleName);
		
		virtual bool CaptureMouseButtonEvent(DatabaseVisWindow * window, int x, int y, 
											 VisWindow::MouseButton mouseButtonID, 
											 VisWindow::MouseButtonState newState);
		virtual bool CaptureMouseMotionEvent(DatabaseVisWindow * window, 
											 int x, int y, int dx, int dy);
		
		virtual bool CaptureKeyPressed(DatabaseVisWindow * window, 
									   const char key, int mouseX, int mouseY);

		virtual VKString GetAppType();
		virtual void Execute();
		virtual void AddPairStats(int fromModel, int toModel);
		virtual void FillPerMapStats(PerMapStats & perMapStats);
		virtual void GetPerPointErrors(const ModelPoint & pnt,
									   std::vector<ModelPoint> & pnts,
									   std::vector<double> & errors);
		virtual void Draw(DatabaseVisWindow * window);
		
		virtual void WriteResultsChild(std::ofstream & textStream);		
		virtual bool ReadResultsChild(std::ifstream & textStream);
	
		virtual double WritePairsTable(DatabaseVisWindow * window);
	
		virtual void DynamicGraphIterations(DatabaseVisWindow * window);

		virtual VKString GetMeshStr(VKStringList & t);
		virtual VKString GetMapStr(PerMapStats & perMapStats);
		virtual void WriteFuzzyMapForPair(PerMapStats & perMapStats);
		virtual void WriteCoarseMapForPair(PerMapStats & perMapStats);
		virtual void WriteFineMapForPair(PerMapStats & perMapStats);
		virtual void WriteErrorsToConcatFile();
		
	protected:
		bool m_bWritingImages;	
	
	// benchmark info
		BenchmarkType m_Benchmark;
		CorrsEvaluationType m_CorrEvaluation;
	
		CorrCollection * m_pGroundTruthCollection;
		CorrValue * m_pGroundTruthValue;
	
		CorrCollection * m_pCorrCollection;
		CorrValue * m_pCorrValue;
		VKString m_sDistanceName;
		bool m_bWriteErrors;
		bool m_bTakeBest;
		bool m_bWriteFuzzyForPair;
		bool m_bWriteCoarseForPair;
		bool m_bWriteFineForPair;
	
	// correspondence data
		std::vector<PerMapStats> m_EvalCorrStats;
		std::map<Model3D *, std::map<Model3D*, int> > m_SrcModelToPairs;
	
	// drawing
		void ClearDrawableErrors();
		void FillDrawAtPnt(DatabaseVisWindow * window, 
						   const ModelPoint & selPnt, bool sort);
		void FillDrawMeshErr(DatabaseVisWindow * window,
							 Model3D * fromModel, int drawableLink);
		BenchDraw m_DrawType;
		int m_DrawableLinkMain;	
		ModelPoint m_DrawableForPnt;
		Model3D * m_pMainModel;
		Model3D * m_pModelTo;
		std::map<Model3D*, ModelPerVertexValues* > m_DrawableError;
		std::map<Model3D*, ModelPerVertexValues*> m_CachedPerVertexVals;
		double m_fMaxErr;
		double m_fMaxErrRender;
		CollectionIdentityPointsets * GetColorInterpolator(Model3D * model);
		std::map<Model3D*, CollectionIdentityPointsets *> m_ColorInterpolators;
		
		bool m_bIterated;
		
};
	

#endif

