/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __AppCreateDB_H
#define __AppCreateDB_H

#include "PipelineStepApplication.h"
#include "AnalysisPipeline.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "ModelValueLabel.h"

struct PSBClassTree
{
	VKString m_NodeStr;
	std::vector<PSBClassTree*> m_pChildren;
	std::vector<int> m_Models;
};

class AppCreateDB : public PipelineStepApplication
{
	public: 
        enum CreateDBState
        {
            CREATEDB_PROCESSING,
            CREATEDB_KEYWORD,
            CREATEDB_NUM_MODELS,
            CREATEDB_INPUTDB,
            CREATEDB_SEARCHING
        };
        
        enum CreateDBInputDB
        {
            GOOGLE_WAREHOUSE,
            PRINCETON_SHAPE_BENCHMARK,
            CORRESPONDENCE_BENCHMARK
        };
    
        enum ProcessingState
        {
            PROCESS_SELECT_STATE,
            PROCESS_DELETE_WHOLE,
            PROCESS_SELECT_CONN_COMP,
            PROCESS_SELECT_FACE,
            PROCESS_ERASE_FACE,
			PROCESS_GROUND_TRUTH,
			PROCESS_GROUND_TRUTH_NEW_LABEL
        };
    
        enum DrawCCFlags
        {
            DRAW_CC     = 1,
            DRAW_MESH   = 2
        };
    
		AppCreateDB(AnalysisPipeline & pipeline, const VKString & moduleName);

        virtual bool CaptureMouseButtonEvent(DatabaseVisWindow * window, int x, int y, 
                                             VisWindow::MouseButton mouseButtonID, 
                                             VisWindow::MouseButtonState newState);
        virtual bool CaptureMouseMotionEvent(DatabaseVisWindow * window, 
                                             int x, int y, int dx, int dy);
    
        virtual void SelectFace(DatabaseVisWindow * window, int x, int y);
    
		virtual bool CaptureKeyPressed(DatabaseVisWindow * window, 
									   const char key, int mouseX, int mouseY);

		virtual VKString GetAppType();
		virtual void Execute();
		virtual void Draw(DatabaseVisWindow * window);
        virtual void DrawMesh(DatabaseVisWindow * window, int subwindowID);
		
		virtual void WriteDBCopy();
        virtual void WriteDB();
	
		virtual void WriteResultsChild(std::ofstream & textStream);	
		virtual bool ReadResultsChild(std::ifstream & textStream);
	
		virtual void DownloadGoogleWarehouse(const VKString & keyword, int nModels, 
											 bool loadToMem, bool keepInMem);
		static VKString DownloadGoogleWarehouse(const VKString & keyword, 
												int nModels, 
												const VKString & dataDir, 
												const VKString & googleDir,
												bool redownloadGoogleData,
												ModelDatabase * db,
												bool loadToMem, bool keepInMem);
        virtual void GetPrincetonShapeBenchmark(const VKString & keyword, int nModels);
        virtual void GetCorrespondenceBenchmark(const VKString & keyword, int nModels);
    
        static VKString InputDBName(CreateDBInputDB db);
        static VKString AppStateName(CreateDBState state);
		
		virtual void ReorderModels();
	
		virtual void SelectGroundTruthPoint(DatabaseVisWindow * window, int x, int y);
		virtual void AddGroundTruthPointLabel(const VKString & strLabel);
		virtual void PrintGroundTruthPointLabels();
		virtual void WriteGroundTruthPointLabels();
		virtual void UndoGroundTruthPointLabel();
	
		std::vector<int> m_GTLabelsHistory;
		VKStringList m_GroundTruthLabels;
		std::map<int, std::vector<int> > m_ModelToGTVertexID;		// deleteme
		VKString m_sNewGroundTruthLabel;
		VKString m_sGroundTruthPntSetName;
	
	protected:
        void DeleteSelectedFaces();
        void DeleteFace(int face);
        void CreateNewModelFromSelectedFaces();
        void AddSelection(int seedFace);
        void MainProcessedModel(DatabaseVisWindow * window, int fromSubwindow);
    
		void FillPSBKeywordToFilesMap();
		void PrintPSBKeywords(int maxDepth);
		void PrintPSBNode(PSBClassTree & node, int currDepth, int maxDepth);
		std::map<VKString, std::vector<int> > m_PSBKeywordToFiles;
		PSBClassTree m_ClassRoot;
	
		void FillCorrsKeywordToFilesMap();
		std::vector<VKString> m_PathToModel;
		VKString m_sCorrsHelpStr;
		std::map<VKString, std::vector<int> > m_CorrKeywordToFiles;
	
        std::vector<int> m_SelectedFaces;
        
        CreateDBState m_State;
        ProcessingState m_ProcessingState;
            
        virtual void DeleteModelInSubwindow(DatabaseVisWindow * window, int subwindowID);
        virtual void DeleteModel(int modelID);
        virtual void UndoDelete();
    
        VKStringList m_sDeletedModelNames;
        
		VKString m_sQueryKeyword;
        int m_iNumToDownload;
        CreateDBInputDB m_InputDB;
        
        bool m_bRedownloadGoogleData;
        bool m_bSelectCCShowHelp;
    
        VKStringList m_ProcessingStateNames;
        
        int m_iDrawingFlags;
    
        void UpdateTriangleCanDraw();
        Model3D * m_TriangsForModelCanDraw;
        bool * m_aIsTriangleInCC;
        int m_iNTriangles;
        bool m_bRightMouseButton;
};
	

#endif

