/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "AppPartSelection.h"
#include "CollectionDiffuseFullMtx.h"
#include "ModelValueLabel.h"
#include "MRFOptimization.h"
#include "DistanceEuclidean.h"
#include "Sortable.h"
#include "PointDescriptor.h"
#include "SinglePairRigidAlign.h"
#include "CollectionIdentityPointsets.h"

AppPartSelection::AppPartSelection(AnalysisPipeline & pipeline, const VKString & moduleName)
: PipelineStepApplication(pipeline, moduleName)
{	
    m_fCurrBrushSize = 0.05;
    m_CurrBrushColor = RNRgb(0, 0, 1);

	m_bApplyTransformToBoth = false;
	m_bDrawMeshes = true;
	
	m_iSemanticGlueAlignmentFlags = GLUE_SCALE_MODEL_DIAMETER | GLUE_ROTATION_FROM_ALIGNMENT;
	//m_iSemanticGlueAlignmentFlags = GLUE_ROTATION_FROM_ALIGNMENT;
	//m_iSemanticGlueAlignmentFlags = GLUE_SCALE_FROM_ALIGNMENT | GLUE_ROTATION_FROM_ALIGNMENT;
	
	m_aTempArrayPnts1 = NULL;
	m_aTempArrayPnts2 = NULL;
	m_aTempArrayWeights = NULL;
	m_iTempArraySize = 0;
	
	m_InteractiveSelectionFlags = 0;
	//m_InteractiveSelectionFlags = Model3DRenderingParams::SELECT_CHECK_NORMAL;
	m_iPointDescriptorID = 0;
	m_fSmoothnessScale = 1.;
	m_iCurrLabelID = -1;
	m_bWritingNewLabel = false;
	//m_iNPoints = -1;
	m_bNeedsUpdate = false;
	m_fShowPartThreshold = .5;
	m_iShowNumParts = pipeline.GetDB()->NumModels();
	m_afCumulativeSimilarities = NULL;
	m_bRightMouseButton = false;
    ChangeInteractionMode(APP_PART_SELECTION);
    //ChangeInteractionMode(APP_INTERACTIVE_LABELING);
    //ChangeInteractionMode(APP_DRAW_DECAL);
	int NModels = m_Pipeline.GetMainDB()->NumModels();
	for (int i=0; i<NModels; i++)
	{
		m_Pipeline.GetMainDB()->GetModelByID(i)->FillConnectedComponentsFaces(true);
		m_bModelPendingUpdate.push_back(false);
	}
	
	m_sGlueLabelName = "";
	m_SemanticGlueState = GLUE_NONE;
	m_pDBWithToBeGluedModel = NULL;
	m_pModelToBeGlued = NULL;
	
	
}

VKString AppPartSelection::GetAppType()
{
	return "AppPartSelection";
}

void AppPartSelection::Execute()
{
	// execute
}

void AppPartSelection::ChangeInteractionMode(InteractionMode newInteractionMode)
{
    m_InteractionMode = newInteractionMode;
    VKString modeName="Unknown";
    if (m_InteractionMode==APP_PART_SELECTION)
    {
        m_fShowPartThreshold = .5;
        modeName = "Part Selection";
    }
    else if (m_InteractionMode==APP_INTERACTIVE_LABELING)
    {
        m_fShowPartThreshold = 1;
        modeName = "Interactive Labeling";
    }
    else if (m_InteractionMode==APP_DRAW_DECAL)
    {
        m_fShowPartThreshold = .5;
        modeName = "Decal Transfer";
        InitializeFineDetailState();
    }
    std::cout<<"Interaction Mode = "<<modeName.c_str()<<" thesh="<<m_fShowPartThreshold<<std::endl;
}

void AppPartSelection::InteractAppEnabled(DatabaseVisWindow * window)
{
	// remember rendering flags
	
	// change rendering flags for VisWindow
	
}

void AppPartSelection::InteractAppDisabled(DatabaseVisWindow * window)
{
	// change to old rendering flags
	
}

void AppPartSelection::SetInteractionMode(InteractionMode mode)
{
	m_InteractionMode = mode;
}

void AppPartSelection::FillPtrsFromParams()
{
	m_sDefaultModelValue = m_Pipeline.m_Params.GetStrValue(m_sModuleName,   
                                                           "MEMORY_InteractModelValue");
	VKString inputColl = m_Pipeline.m_Params.GetStrValue(m_sModuleName, 
                                                         "INPUT_Collection");
	VKString corrValue = m_Pipeline.m_Params.GetStrValue(m_sModuleName, 
                                                         "INPUT_CollectionCorrValue");

	FillPtrs(inputColl, corrValue, m_sDefaultModelValue);
}

void AppPartSelection::FillPtrs(const VKString & collection, const VKString & corrValue,
								const VKString & outputModelValue)
{
	m_sDefaultModelValue = outputModelValue;
	m_pInputCollection = m_Pipeline.GetMainDB()->GetCorrCollection(collection);
	assert(m_pInputCollection!=NULL);
	m_pInputCorrValue = m_pInputCollection->GetCorrValueByName(corrValue);
	assert(m_pInputCorrValue!=NULL);
	
}

bool AppPartSelection::CaptureKeyPressed(DatabaseVisWindow * window, 
										 const char key, int mouseX, int mouseY)
{
	if (m_bWritingNewLabel && ((key>='A' && key <= 'z') || (key >= '0' && key <= '9')))
	{
		char str[2];
		str[0] = key;
		str[1] = 0;
		m_sCurrLabel += VKString(str);
		window->GetRenderer()->UpdateScreen();
		return true;
	}

	bool keyCaptured = true;
	switch (key)
	{
		case '`':
			std::cout<<"Part Selection App:"<<std::endl;
			std::cout<<"\tRIGHT MOUSE MOVE\t - Hold right mouse button and move to select a regoin"<<std::endl;
			std::cout<<"\t>/<\t - Increase/Decrease threshold for a part to be shown"<<std::endl;
			std::cout<<"\t/\t - Reset selected region"<<std::endl;
			std::cout<<"\t?\t - Search for the most similar parts"<<std::endl;
			std::cout<<"\t\\\t - Simulate right mouse button"<<std::endl;
			std::cout<<"\t;\t - Select/Add label, terminate with '#' ";
			std::cout<<"to accept label, ';' to disregard"<<std::endl;
			std::cout<<"\tL\t - Generate final labelings "<<std::endl;;
			std::cout<<"\tw/r\t - read/write current labels "<<std::endl;;
			std::cout<<"\tc\t - clear final labels only "<<std::endl;
			std::cout<<"\tC\t - clear all labels (intermediate and final) "<<std::endl;
			std::cout<<"\tS/s\t - Adjust smoothness weight (for MRF labeling) "<<std::endl;
			std::cout<<"\t!\t - Choose next point descriptor "<<std::endl;
			std::cout<<"\t't'\t - (general) test normals in region selection"<<std::endl;
			std::cout<<"\t't'\t - (semantic glue alignment stage) transform models together"<<std::endl;
			std::cout<<"\t'T'\t - test distance in region selection (not implemented)"<<std::endl;
			std::cout<<"\t'i'\t - next step in gluing pipeline"<<std::endl;
			std::cout<<"\t'I'\t - clear gluing"<<std::endl;
            std::cout<<"\t'g'\t - semantically glue decal"<<std::endl;
			std::cout<<"\t'm'\t - draw/don't draw meshes"<<std::endl;
			std::cout<<"\t'k/K'\t - Faceted browsing: add +1/-1 facet"<<std::endl;
			std::cout<<"\t'y/Y'\t - Clear last / all facets"<<std::endl;
			keyCaptured = false;
			break;
		case '>':
		case '<':
		{
			double dx = 0;
			double maxVal = 0;
			double minVal = 0;
			if (m_InteractionMode==APP_PART_SELECTION)
			{
				minVal = 0;
				maxVal = 1;
				dx = .01;
			}
			else if (m_InteractionMode==APP_INTERACTIVE_LABELING)
			{
				minVal = 1;
				maxVal = 100;
				dx = 1;
			}
			if (key=='>')
				m_fShowPartThreshold += dx;
			else
				m_fShowPartThreshold -= dx;
			if (m_fShowPartThreshold>maxVal)
				m_fShowPartThreshold = maxVal;
			else if (m_fShowPartThreshold < minVal)
				m_fShowPartThreshold = minVal;
			std::cout<<"ShowPartThresh = "<<m_fShowPartThreshold<<std::endl;
			break;
		}
		case '/':
			SelectCurrentModel(-1);
			GetRefSelectedPnts().clear();
            m_BrushStrokes.clear();
			ClearPaint();
			if (m_afCumulativeSimilarities!=NULL)
			{
				for (int i=0; i<m_Pipeline.GetDB()->NGlobalPnts(); i++)
					m_afCumulativeSimilarities[i] = 0;
			}
			std::cout<<"Cleared All"<<std::endl;
			break;
		case '?':
			m_bNeedsUpdate = true;
			window->m_NewModelOrder = m_aiSimilarModels;
			window->m_iCurrModelTraversalID = 0;
			window->FillMostSimilarModels(window->m_iCurrModelTraversalID, 
										  window->m_NewModelOrder);
			break;
		case 'j':
		case 'J':
		{
			if (key=='j')
			{
				m_aiSimilarModels.clear();
				GlueOtherToCurrent();
				FillCumulativeSimilarities();
				FillModelValuesMostSimilarParts(m_aiSimilarModels, m_iShowNumParts);
				window->m_NewModelOrder = m_aiSimilarModels;
				window->m_iCurrModelTraversalID = 0;
				window->FillMostSimilarModels(window->m_iCurrModelTraversalID, 
											  window->m_NewModelOrder);
			}
			else
			{
				m_GlueDataModelPairToTransform.clear();
			}
			break;
		}
		case 'l':
			SaveCurrRegionLabel();
			break;
		case 'L':
			std::cout<<"Labeling all..."<<std::endl;
			CreateFinalLabelsGlobalMRFOptimization();
			//CreateFinalLabelsMRFOptimization(false);
			//CreateFinalLabelsAssignMax();
			break;
		case 'm':
			m_bDrawMeshes = !m_bDrawMeshes;
			break;
        case 'g':
            if (m_InteractionMode==APP_DRAW_DECAL)
                AdvanceFineDetailState();
            else
                std::cout<<"Can only transfer decal draw_decal mode. Press '2'"<<std::endl;
            break;
		case 'I':
			m_SemanticGlueState=GLUE_NONE;
			m_pDBWithToBeGluedModel = NULL;
			m_pModelToBeGlued = NULL;
			m_GlueDataModelPairToTransform.clear();
			break;
		case 'k':
		case 'K':
		case 'Y':
		case 'y':
			if (key=='k'
				|| key=='K')
			{
				AddFacet((key=='k' ? 1. : -1.));
				window->m_NewModelOrder = m_aiSimilarModels;
				window->m_iCurrModelTraversalID = 0;
				window->FillMostSimilarModels(window->m_iCurrModelTraversalID, 
											  window->m_NewModelOrder);
			}
			else
				ClearFacets(key=='Y');
			break;
		case 'i':
			if (m_SemanticGlueState==GLUE_NONE)		// first type label name (to align to)
			{
				if (m_Pipeline.GetMainDB()==m_Pipeline.GetDB())
				{
					std::cout<<"[WARNING] to be glued model cannot be taken from main DB."<<std::endl;;
					break;
				}

				m_pDBWithToBeGluedModel = m_Pipeline.GetDB();
				m_pModelToBeGlued = m_pDBWithToBeGluedModel->GetModelInSubwindow(0);
				m_bWritingNewLabel = true;
				m_sCurrLabel = "";
				m_SemanticGlueState = GLUE_SELECT_LABEL;
				m_Pipeline.SwitchToMainDB();
				std::cout<<"[INFO] Type in glue label name"<<std::endl;
			}
			else if (m_SemanticGlueState==GLUE_SELECT_LABEL)
			{
				m_sCurrLabel = "";
				std::cout<<"[INFO] Cleared glue label name"<<std::endl;
			}
			else if (m_SemanticGlueState==GLUE_INTERACTIVE_ALIGNMENT)
			{
				SemanticGlueAutoAttachToAll();
				m_SemanticGlueState = GLUE_PROPAGATED_TRANSFORMS;
			}
			else if (m_SemanticGlueState == GLUE_PROPAGATED_TRANSFORMS)
			{
				m_SemanticGlueState = GLUE_HIDE_SECOND_MODEL;
				std::cout<<"[INFO] Hiding glued results"<<std::endl;
			}
			else if (m_SemanticGlueState == GLUE_HIDE_SECOND_MODEL)
			{
				m_SemanticGlueState = GLUE_PROPAGATED_TRANSFORMS;
				std::cout<<"[INFO] Showing glued results"<<std::endl;
			}
			else
				assert(false);
			break;
		case ';':			// add new label
//			if (m_InteractionMode!=APP_INTERACTIVE_LABELING)
//				CaptureKeyPressed(window, '1', mouseX, mouseY);
//			m_bWritingNewLabel = !m_bWritingNewLabel;
//			std::cout<<"Labeling mode: "<<m_bWritingNewLabel<<std::endl;
//			m_sCurrLabel = "";
//			m_iCurrLabelID = -1;
			std::cout<<"[WARNING] temporarily disabled"<<std::endl;
			break;
		case '#':	// wrote label - accepted
			if (m_SemanticGlueState==GLUE_SELECT_LABEL)
			{
				if (!m_CreatedLabels.contains(m_sCurrLabel))
				{
					std::cout<<"[WARNING] Label "<<m_sCurrLabel.c_str()<<" does not exist"<<std::endl;
					std::cout<<"Valid Labels: "<<std::endl;
					for (int i=0; i<m_CreatedLabels.count(); i++)
						std::cout<<"\t\t"<<m_CreatedLabels[i].c_str()<<std::endl;
                    if (m_CreatedLabels.count()==0)
                        CaptureKeyPressed(window, 'I', mouseX, mouseY);
					break;
				}
				assert(m_bWritingNewLabel);
				m_bWritingNewLabel = false;
				m_SemanticGlueState = GLUE_INTERACTIVE_ALIGNMENT;
				
				m_sGlueLabelName = m_sCurrLabel;
				m_sCurrLabel = "";
				Model3D * m1 = m_Pipeline.GetMainDB()->GetModelInSubwindow(0);
				Model3D * m2 = m_pModelToBeGlued;
				ModelPoint p1, p2;
				for (int i=0; i<m1->PointCloud()->NumPoints(); i++)
					if (PaintedOn(m1, m1->PointCloud()->GetPoint(i).NearestVertex()))
					{
						p1 = m1->PointCloud()->GetPoint(i);
						break;
					}
				for (int i=0; i<m2->PointCloud()->NumPoints(); i++)
					if (PaintedOn(m2, m2->PointCloud()->GetPoint(i).NearestVertex()))
					{
						p2 = m2->PointCloud()->GetPoint(i);
						break;
					}
				assert(!p1.Invalid() && !p2.Invalid());
				R3Affine aff = R4identity_matrix;
				R3Point c1 = m1->GetMesh()->Centroid();
				R3Point c2 = m2->GetMesh()->Centroid();
				double s = m1->GetMesh()->AverageRadius(&c1) / m2->GetMesh()->AverageRadius(&c2);
				//aff.Translate(p1.GetPosition() - s * p2.GetPosition());
				//aff.Scale(s);
				m_GlueDataModelPairToTransform[m2][m1] = aff;
				R3Triad triad(R3Vector(1, 0, 0), R3Vector(0, 1, 0), R3Vector(0, 0, 1));
				m_MainPairCoords = R3CoordSystem(R3Point(0,0,0), triad);
				//m_MainPairCoords.Translate(p1.GetPosition() - s * p2.GetPosition());
				m_MainPairCoords.Translate(p1.GetPosition() / s - p2.GetPosition());
				m_fMainPairScale = s;
				ScaleAffine(m1, m2, 0);
				//m_MainPair.Scale(s);
				//aff = m_MainPair.Matrix();
				//R3Point p2new = p2.GetPosition();
				//aff.Apply(p2new);
				//std::cout<<"p1=["<<p1.GetPosition().X()<<" "<<p1.GetPosition().Y()<<" "<<p1.GetPosition().Z()<<"] "<<std::endl;
				//std::cout<<"p2=["<<p2new.X()<<" "<<p2new.Y()<<" "<<p2new.Z()<<"] "<<std::endl;
				std::cout<<"[INFO] Glue to label: "<<m_sGlueLabelName.c_str()<<std::endl;
				std::cout<<"Transform second model to align. "<<std::endl;;
				std::cout<<"\tPress 'b'/'B'' to choose which model to transform. "<<std::endl;;
				std::cout<<"\tPress 'i' when done."<<std::endl;
				m_Pipeline.SwitchToMainDB();
			}
			else if (m_bWritingNewLabel)
			{
				m_iCurrLabelID = m_CreatedLabels.containsGetID(m_sCurrLabel);
				if (m_iCurrLabelID==-1)
				{
					m_iCurrLabelID = m_CreatedLabels.count();
					std::cout<<"Added new label: id="<<m_iCurrLabelID;
					std::cout<<" Name="<<m_sCurrLabel.c_str()<<std::endl;
					m_CreatedLabels.push_back(m_sCurrLabel);
				}
				else 
				{
					std::cout<<"Chose existing label: id="<<m_iCurrLabelID;
					std::cout<<" Name="<<m_CreatedLabels[m_iCurrLabelID].c_str()<<std::endl;
				}
				m_sCurrLabel = "";
				m_bWritingNewLabel = false;
			}
			break;
		case '0':
		case '1':
		case '2':
		{
			m_InteractionMode = (InteractionMode)((int)(key-'0'));
            ChangeInteractionMode(m_InteractionMode);
			break;
		}
		case '\\':
			m_bRightMouseButton = !m_bRightMouseButton;
			break;
		case 'w':
			std::cout<<"Write interaction results"<<std::endl;
			WriteInteractionResults();
			break;
		case 'r':
			ReadInteractionResults();
			break;
		case 'c':
			std::cout<<"Clear final labels"<<std::endl;
			ClearFinalLabels();
			break;
		case 'C':
			std::cout<<"Clear all labels"<<std::endl;
			ClearAllCreatedModelValues();
			ClearFinalLabels();
			break;
		case 't':
		{
			if (m_SemanticGlueState==GLUE_INTERACTIVE_ALIGNMENT)
				m_bApplyTransformToBoth = !m_bApplyTransformToBoth;
			else
			{
				std::cout<<"Interactive Selection:";
				if ((m_InteractiveSelectionFlags & Model3DRenderingParams::SELECT_CHECK_NORMAL)!=0)
				{
					m_InteractiveSelectionFlags &= (~Model3DRenderingParams::SELECT_CHECK_NORMAL);
					std::cout<<" not testing normal"<<std::endl;
				}
				else
				{
					m_InteractiveSelectionFlags |= Model3DRenderingParams::SELECT_CHECK_NORMAL;
					std::cout<<" testing normal"<<std::endl;
				}
			}
			break;
		}
		case 'S':
		case 's':
			m_fSmoothnessScale += (key=='s') ? .1 : -.1;
			if (m_fSmoothnessScale < 0)
				m_fSmoothnessScale = 0;
			std::cout<<"Smoothness: "<<m_fSmoothnessScale<<std::endl;
			break;
		case '!':
		{
			std::cout<<"Writing Alternative Conservation value"<<std::endl;
			double * consVal = new double[m_Pipeline.GetDB()->NGlobalPnts()];
			CreateAlternativeConservation(consVal);
			CreateModelValue("AlternativeConservation", consVal);
			delete [] consVal;
			break;
		}
//		case '!':
//		{	
//			m_iPointDescriptorID++;
//			if (m_iPointDescriptorID >= m_Pipeline.GetDB()->m_PointDescriptorNames.count())
//				m_iPointDescriptorID = -1;
//			std::cout<<"Pnt Descriptor: ";
//			if (m_iPointDescriptorID>=0)
//				std::cout<<m_Pipeline.GetDB()->m_PointDescriptorNames[m_iPointDescriptorID].c_str();
//			else
//				std::cout<<"none";
//			std::cout<<std::endl;
//			break;
//		}
		case 'b':
		case 'B':
			m_Pipeline.ChangeDB(key=='b');
			break;
		case 'o':	// orient. Uses 'top' label and 'front' label
			OrientAllModels(ORIENT_TOP_AND_UP);
			break;
		case 'O':	// orient. Aligns points with 'align' label (or all points)
			OrientAllModels(ORIENT_ALIGN_AREA);
			break;
		default:
			keyCaptured = false;

	};
	if (keyCaptured)
		window->GetRenderer()->UpdateScreen();
	return keyCaptured;
}

bool AppPartSelection::CaptureMouseMotionEvent(DatabaseVisWindow * window, 
                                               int x, int y, int dx, int dy)
{
	bool retVal = false;
//	static TimeProfiler timer;
//	static int nUpdates = 0;
//	nUpdates++;
//	timer.startedProcess("CaptureMouse");
	// Paint on surfaces
	if (window->IsMouseButtonDown(VisWindow::MOUSE_RIGHT)
		|| (m_bRightMouseButton && window->IsMouseButtonDown(VisWindow::MOUSE_LEFT)))
	{
		window->MouseButtonEvent(x, y, VisWindow::MOUSE_LEFT, VisWindow::MOUSE_BT_DOWN);
		
		int selectedWndID = window->SelectedWindowID();
		ModelDatabaseRenderingParams * renderer = m_Pipeline.GetDB()->GetRenderParams();
		assert(selectedWndID>=0 && 
			   selectedWndID < renderer->m_mWindowToModelID.size());
		int modelID = renderer->m_mWindowToModelID[selectedWndID];
		Model3D * model = m_Pipeline.GetDB()->GetModelByID(modelID);
		assert(model!=NULL);
        
        if (m_InteractionMode==APP_DRAW_DECAL
            && m_DecalTransferState == DECAL_TRANS_DRAW_DECAL)
        {
            FineDetailBrushStrokes brush; 
            ModelPoint p = model->RenderingParams()->QuickIntersection(x, y, 0, 0);
            
            if (!p.Invalid())
            {
                bool prevTooClose = false;
//                if (m_BrushStrokes.size() >= 1)
//                {
//                    DistanceEuclidean * distMetric = ((Model3D*)p.GetModel())->DefaultEuclideanDistance();
//                    assert(distMetric!=NULL);
//                    double dist = distMetric->Distance(m_BrushStrokes[m_BrushStrokes.size()-1].m_Pnt, p);
//                    prevTooClose = dist < m_fCurrBrushSize/2.;
//                }
                
                if (!prevTooClose)
                {
                    brush.m_Pnt = p;
                    brush.m_fBrushSize = m_fCurrBrushSize;
                    brush.m_BrushColor = m_CurrBrushColor;
                
                    m_BrushStrokes.push_back(brush);
                }
            }
        }
        else
        {
			ModelPoint p = model->RenderingParams()->QuickIntersection(x, y, 0, 0);
			
			if (!p.Invalid())
			{
				ModelPerVertexValues * paintedPerVtxVals = GetPaintedOnRendering(model);
				
				if (modelID!=GetSelectedModel())
				{
					SelectCurrentModel(modelID);
					GetRefSelectedPnts().clear();
					paintedPerVtxVals->SetValues("auto", 0);
				}
				
				std::vector<int> & refSel = GetRefSelectedPnts();

				CollectionIdentityPointsets * f2p = model->FacesFromPointsInterpolator();
				std::vector<ModelPoint> corrPnts;
				f2p->GetCorrs(p, &corrPnts);
				assert(corrPnts.size()==1);
				int pntID = corrPnts[0].PntCloudID();
				assert(pntID>=0);
				refSel.push_back(pntID);
				
				std::sort(refSel.begin(), refSel.end());
				for (int i=0; i<(int)refSel.size()-1; i++)
					if (refSel[i]==refSel[i+1])
						refSel.erase(refSel.begin() + (i--) );

				FillPaintedOnRendering(model, refSel);
			}
        }				
		//iter->second->Interpolate(ModelValuesInterpolatorDescriptor());
		
		retVal = true;
	}
	else if (window->IsMouseButtonDown(VisWindow::MOUSE_LEFT)
			 && m_SemanticGlueState==GLUE_INTERACTIVE_ALIGNMENT
			 && !m_bApplyTransformToBoth)
	{
		Model3D * model = m_Pipeline.GetMainDB()->GetModelInSubwindow(0);
		assert(model!=NULL);

		double fdx = (double)dx / (double)window->Width();
		double fdy = (double)dy / (double)window->Height();
		if (!window->Ctrl() && !window->Alt() && !window->Shift())	// rotate
			RotateAffine(model, m_pModelToBeGlued, fdx, fdy);
		else if (window->Ctrl())
			ScaleAffine(model, m_pModelToBeGlued, fdy);
		else if (window->Shift())
			TranslateAffine(model, m_pModelToBeGlued, fdx, fdy);
		retVal = true;
	}
//	else if (window->IsMouseButtonDown(VisWindow::MOUSE_LEFT)
//			 && m_SemanticGlueState==GLUE_INTERACTIVE_ALIGNMENT
//			 && m_bApplyTransformToBoth)
//	{
//
//		
//		bool inWnd;
//		double xWnd, yWnd;
//		VKString subwindowName = window->InSubWindow(x, y, &inWnd, &xWnd, &yWnd);
//		
//		if (!inWnd)
//		{
//			assert(m_pModelToBeGlued!=NULL);
//			// camera events
//			Model3D * model = m_Pipeline.GetMainDB()->GetModelInSubwindow(0);
//			assert(model!=NULL);
//			
//			if (!window->Ctrl() && !window->Alt() && !window->Shift())	// rotate
//			{
//				m_pModelToBeGlued->RenderingParams()->Rotate(0, x, y, dx, dy);
//				model->RenderingParams()->Rotate(0, x, y, dx, dy);
//			}
//			else if (window->Ctrl())
//			{
//				m_pModelToBeGlued->RenderingParams()->Scale(0, x, y, dx, dy);
//				model->RenderingParams()->Scale(0, x, y, dx, dy);
//			}
//			else if (window->Shift())
//			{
//				m_pModelToBeGlued->RenderingParams()->Translate(0, x, y, dx, dy);
//				model->RenderingParams()->Translate(0, x, y, dx, dy);
//			}
//			retVal = true;
//		}
//	}
//	timer.finishedProcess("CaptureMouse", nUpdates%100==0);
	return retVal;
}

void AppPartSelection::FillPaintedOnRendering(Model3D* model, std::vector<int> & selectedPntCld)
{
	ModelPerVertexValues * perVertexVals = GetPaintedOnRendering(model);
	for (int i=0; i<(int)selectedPntCld.size(); i++)
		perVertexVals->SetVertexValue(model->PointCloud()->GetPoint(selectedPntCld[i]), 1.);
	
	perVertexVals->Interpolate();
}

ModelPerVertexValues * AppPartSelection::GetPaintedOnRendering(Model3D * model)
{
	std::map<Model3D *, ModelPerVertexValues *>::iterator iter;
	iter = m_pModelIDToPaintedOn.find(model);

	if (iter==m_pModelIDToPaintedOn.end())
	{
		ModelPerVertexValues * retVals = new ModelPerVertexValues(model, false, false);;
		m_pModelIDToPaintedOn[model] = retVals;
		return retVals;
	}
	return iter->second;
}

void AppPartSelection::OrientAllModels(ModelOrientation orient)
{
	TimeProfiler timer;
	if (orient==ORIENT_TOP_AND_UP)
	{
		std::cout<<"Re-orienting all models: based on 'top' and 'up' vectors"<<std::endl;
		for (int modelID=0; modelID<m_Pipeline.GetMainDB()->NumModels(); modelID++)
		{
			timer.WriteProgress("prog", modelID, m_Pipeline.GetMainDB()->NumModels());
			Model3D * model = m_Pipeline.GetMainDB()->GetModelByID(modelID);
			
			PointSet * pnts = model->PointCloud();
			R3Vector upVector(0, 0, 1.);
			R3Vector towardsVector(-0.57735, 0.57735, 0);
			
			// check if labels 'top' and 'front' exist.
			for (int currLabelID=0; currLabelID<m_CreatedLabels.count(); currLabelID++)
			{
				
				if (m_CreatedLabels[currLabelID]=="top"
					|| m_CreatedLabels[currLabelID]=="front")
				{
					VKString modelValueName = m_sDefaultModelValue + "_" + m_CreatedLabels[currLabelID];
					ModelValue * values = model->GetModelValue(modelValueName);
					assert(values!=NULL);
					
					double totalWeight = 0;
					R3Vector aveVec(0,0,0);
					for (int pntID = 0; pntID < pnts->NumPoints(); pntID++)
					{
						double val = values->Value(pnts->GetPoint(pntID));
						totalWeight += val;
						aveVec = aveVec + val * pnts->GetPoint(pntID).Normal();
					}
					
					//aveVec = aveVec / totalWeight;
					if (totalWeight>0)
					{
						aveVec.Normalize();
						
						R3Vector reorientedAve(0,0,0);
						for (int pntID = 0; pntID < pnts->NumPoints(); pntID++)
						{
							double val = values->Value(pnts->GetPoint(pntID));
							R3Vector norm = pnts->GetPoint(pntID).Normal();
							if (aveVec.Dot(norm) < 0)
								reorientedAve = reorientedAve - val * norm;
							else
								reorientedAve = reorientedAve + val * norm;
						}
						reorientedAve.Normalize();
					
						if (m_CreatedLabels[currLabelID]=="top")
							upVector = reorientedAve;
						else if (m_CreatedLabels[currLabelID]=="front")
							towardsVector = reorientedAve;
						else
							assert(false);
					}
				}
			}
			
			// ensure that towards and up are orthonormal
			R3Vector leftVector(upVector);
			leftVector.Cross(towardsVector);
			leftVector.Normalize();
			towardsVector = leftVector;
			towardsVector.Cross(upVector);
			towardsVector.Normalize();
			model->RenderingParams()->UpdateCamera(upVector, -towardsVector);
		}
	}
	else if (orient==ORIENT_ALIGN_AREA)
	{ 
		std::cout<<"Re-orienting all models: based on area labeled 'align'"<<std::endl;
		
		// TODO: align selected?
		
		Model3D * mainModel = m_Pipeline.GetDB()->GetModelInSubwindow(0);
		VKString mainModelName = mainModel->ModelName();
		
		for (int modelID=0; modelID<m_Pipeline.GetDB()->NumModels(); modelID++)
		{
			timer.WriteProgress("prog", modelID, m_Pipeline.GetMainDB()->NumModels());
			
			Model3D * model = m_Pipeline.GetDB()->GetModelByID(modelID);
			if (mainModel==model)
				continue;
			
			VKString modelName = model->ModelName();
			ParamParser & par = m_Pipeline.m_Params;
			VKString inputColl = par.GetStrValue(m_sModuleName, "INPUT_Collection");
			VKString corrValue = par.GetStrValue(m_sModuleName, "INPUT_CollectionCorrValue");
			
			// model names: from, to
			std::vector<PointWithID> corrsP1;
			std::vector<PointWithID> corrsP2;
			std::vector<double> weights;
			
			// find rigid transform that aligns main model to a model in the DB
			SinglePairRigidAlign::FindCorrs(m_Pipeline.GetDB(), mainModelName, modelName, 
											inputColl, corrValue, &corrsP1, &corrsP2, 
                                            &weights, true);
			
			R4Matrix xForm = SinglePairRigidAlign::BestRigidAlignment(corrsP1, corrsP2, 
                                                                      weights, 
																	  true, true, true,
																	  NULL);
			
			// transform main model's camera to new model
			R3Affine alignFromMainToModel(xForm);
						
			R3Vector mainUp = mainModel->RenderingParams()->GetCamUp(0);
			R3Vector mainTow = mainModel->RenderingParams()->GetCamTowards(0);
						
			alignFromMainToModel.Apply(mainUp);
			alignFromMainToModel.Apply(mainTow);
			
			model->RenderingParams()->UpdateCamera(mainUp, mainTow);
		}
	}
	else
		assert(false);
	std::cout<<std::endl;
}

//int AppPartSelection::GetNPoints()
//{
//	if (m_iNPoints<0)
//		m_iNPoints = CollectionDiffuseFullMtx::NPoints(m_Pipeline.GetMainDB());
//
//	return m_iNPoints;
//}

int AppPartSelection::GetSelectedModel()
{
	std::map<ModelDatabase*, int >::iterator iter;
	iter = m_DBToSelectedModelID.find(m_Pipeline.GetDB());
	
	if (iter==m_DBToSelectedModelID.end())
		return -1;
	
	return iter->second;
}

void AppPartSelection::SelectCurrentModel(int modelID)
{
	m_DBToSelectedModelID[m_Pipeline.GetDB()] = modelID;
}

std::vector<int> & AppPartSelection::GetRefSelectedPnts()
{
	return m_DBToSelectedPntIDs[m_Pipeline.GetDB()];
}

bool AppPartSelection::PaintedOn(Model3D * model, int vertexID)
{
	if (GetSelectedModel()==-1)
		return false;
	
	std::map<Model3D*, ModelPerVertexValues *>::iterator iter = m_pModelIDToPaintedOn.find(model);
	if (iter==m_pModelIDToPaintedOn.end())
		return false;
	
	return iter->second->GetValueAtVertex(vertexID) == 1.;
}

void AppPartSelection::ClearPaint()
{
	// clear paint on all models (in the current DB)
	std::map<Model3D *, ModelPerVertexValues *>::iterator iter;
	for (iter = m_pModelIDToPaintedOn.begin(); iter != m_pModelIDToPaintedOn.end(); iter++)
	{
		if (iter->first->GetDB() == m_Pipeline.GetDB())
			iter->second->SetValues("AllVertices", 0);
	}
	
	// clear transfered values (only if main DB)
	if (m_Pipeline.GetDB()==m_Pipeline.GetMainDB())
	{
		std::map<int, ModelPerVertexValues *>::iterator iter2;	
		for (iter2 = m_mModelIDToPerVertexValues.begin(); 
			 iter2 != m_mModelIDToPerVertexValues.end(); iter2++)
			iter2->second->SetValues("AllVertices", 0);
	}
}

void AppPartSelection::SaveCurrRegionLabel()
{
	assert(m_Pipeline.GetDB()==m_Pipeline.GetMainDB());
	// save current painted label
	if (m_iCurrLabelID!=-1)
	{
		std::map<int, std::vector<int> > & modelMap = m_mModelPointLabel[GetSelectedModel()];
		std::vector<int> & selRef = GetRefSelectedPnts();
		for (int i=0; i<(int)selRef.size(); i++)
			modelMap[selRef[i]].push_back(m_iCurrLabelID);
				
		std::cout<<"Labeled: "<<selRef.size();
		std::cout<<" points as "<<m_CreatedLabels[m_iCurrLabelID].c_str()<<std::endl;
		selRef.clear();
	}
	else
		std::cout<<"[WARNING] Could not save region label - label is not defined"<<std::endl;
}


void AppPartSelection::FillCumulativeSimilarities()
{
	assert(m_Pipeline.GetDB()==m_Pipeline.GetMainDB());
	// SLOW_TAG	
	FillPtrsFromParams();
	
	ModelDatabase * db = m_Pipeline.GetMainDB(); 
	int nGlobalPoints = db->NGlobalPnts();
//	int NPoints = GetNPoints();
//	int NModels = m_Pipeline.GetMainDB()->NumModels();
	
	if (m_afCumulativeSimilarities==NULL)
		m_afCumulativeSimilarities = new double[nGlobalPoints];
	
	for (int i=0; i<nGlobalPoints; i++)
		m_afCumulativeSimilarities[i] = 0;
	
	// accumulate similarities
	Model3D * model = m_Pipeline.GetMainDB()->GetModelByID(GetSelectedModel());
	if (model==NULL)
	{
		std::cout<<"[WARNING] Cum Sim not filled"<<std::endl;
		return;
	}
	assert(model!=NULL);
	PointSet * points = model->PointCloud();
	assert(points!=NULL);

	std::vector<ModelPoint> selectedRegion;
	std::vector<int> & selRef = GetRefSelectedPnts();
	for (int i=0; i<(int)selRef.size(); i++)
	{
		int pntID = selRef[i];
		selectedRegion.push_back(points->GetPoint(pntID));
		
	}
	m_pInputCollection->AutoNormalizeCorrs();
	//m_pInputCollection->SetRegionNormalization(selectedRegion);
	TimeProfiler timer;
	int maxCount = 0;
	for (int i=0; i<(int)selectedRegion.size(); i++)
	{
		ModelPoint pntFrom = selectedRegion[i];
		std::vector<ModelPoint> corrs;
		m_pInputCollection->GetCorrs(pntFrom, &corrs);
		if (i==0)
			maxCount = (int)corrs.size() * (int)selectedRegion.size();

//		for (int j=0; j<db->NumModels(); j++)
//			corrs.push_back(m_pInputCollection->GetBestCorr(pntFrom, db->GetModelByID(j), 
//															"none", NULL));
		std::vector<double> corrVals;
		for (int j=0; j<(int)corrs.size(); j++)
		{
			if (j%100==0)
				timer.WriteProgress("TmpCalcCumSim", i*(int)corrs.size() + j, maxCount);
			
			double v1 = m_pInputCorrValue->Value(pntFrom, corrs[j]);
			//double v2 = m_pInputCorrValue->Value(corrs[j], pntFrom);
			double v2 = 1.;
			corrVals.push_back(pow(v1*v2, 1.));
		}
		double norm = 1.;
//////// BELOW NEW STUFF JAN 9
		double * perPointSums = new double [db->NumModels()];
		// find per model sums (to be used by sorting)
		for (int j=0; j<db->NumModels(); j++)
			perPointSums[j] = 0;
		
		for (int j=0; j<(int)corrs.size(); j++)
		{
			double val = corrVals[j];
			perPointSums[corrs[j].GetModel()->ModelID()] += val;
		}
		
		// find best value (for given 'from' point)
		norm = 0;
		for (int j=0; j<db->NumModels(); j++)
		{
			if (perPointSums[j] > norm)
				norm = perPointSums[j];
		}
		
		//std::cout<<"\nNorm["<<i<<"] = "<<norm<<std::endl;
		//exit(0);
		delete [] perPointSums;
//////// ABOVE NEW STUFF JAN 9	
		
		// fill similarity values
		for (int j=0; j<(int)corrs.size(); j++)
		{
			int mID = corrs[j].GetModel()->ModelID();
			if (mID!=GetSelectedModel())
			{
				//int position = CollectionDiffuse::MatrixPosition(corrs[j], NPoints, NModels);
				int position = db->GlobalID(corrs[j]);
				//double corrVal = corrVals[j]/norm;
				double corrVal = corrVals[j] / norm;
				
				assert(position < nGlobalPoints);
				double currVal = m_afCumulativeSimilarities[position];
				m_afCumulativeSimilarities[position] = vkMax(corrVal, currVal);
				//m_afCumulativeSimilarities[position] += corrVal;
			}
		}
		
		//int position = CollectionDiffuse::MatrixPosition(pntFrom, NPoints, NModels);
		int position = db->GlobalID(pntFrom);
		m_afCumulativeSimilarities[position] = (double)PaintedOn((Model3D*)pntFrom.GetModel(),
																 pntFrom.NearestVertex());
	}
	std::cout<<std::endl;

	m_pInputCollection->AutoNormalizeCorrs();
	
	m_PerModelShapeDescriptorScaling.clear();
	if (m_GlueDataModelPairToTransform.size()>0)
	{
		
		std::cout<<"NOTE: cumulative values w/ shape descriptors!"<<std::endl;
		ShapeDescriptorSimilarityTerm(selRef, 
									  GetSelectedModel(), 
									  m_PerModelShapeDescriptorScaling);
	}
	
	// normalize max-to-1
	//for (int i=0; i<NPoints * NModels; i++)
	//	m_afCumulativeSimilarities[i] /= maxVal;
}

void AppPartSelection::FillModelValuesMostSimilarParts(std::vector<int> & modelIDs, 
													   int NBestModels)
{
	modelIDs.clear();
	assert(m_Pipeline.GetMainDB()==m_Pipeline.GetDB());
	m_fGlobalMax = 0;
	m_ModelValueOrder.clear();
	m_PerModelValue.clear();
	m_PerModelMax.clear();
	
	ModelDatabase * db = m_Pipeline.GetMainDB();
	int nGlobalPoints = db->NGlobalPnts();
	
	//int NPoints = GetNPoints();
	//int NModels = m_Pipeline.GetMainDB()->NumModels();

	std::vector<Sortable> sortablePerModelVals;
	// init all model values to 0
	for (int i=0; i<db->NumModels(); i++)
	{
		sortablePerModelVals.push_back(Sortable(0, NULL, i));
		m_PerModelMax.push_back(0);
		m_bModelPendingUpdate[i] = true;
	}
	
	// decide on model scores (sum values)
	for (int i=0; i<nGlobalPoints; i++)
	{
		double currVal = m_afCumulativeSimilarities[i];
		//int modelID = i / NPoints;
		ModelPoint p = db->GetPoint(i);
		int modelID = p.GetModel()->ModelID();
		if (GetSelectedModel()!=modelID)
		{
			sortablePerModelVals[modelID].value -= currVal;
			m_PerModelMax[modelID] = vkMax(m_PerModelMax[modelID], 
										   m_afCumulativeSimilarities[i]);
			m_fGlobalMax = vkMax(m_fGlobalMax, m_afCumulativeSimilarities[i]);
		}
	}
	
	if (m_PerModelShapeDescriptorScaling.size()==sortablePerModelVals.size())
	{
		for (int i=0; i<(int)sortablePerModelVals.size(); i++)
			sortablePerModelVals[i].value = -m_PerModelShapeDescriptorScaling[i];
//			sortablePerModelVals[i].value /= m_PerModelShapeDescriptorScaling[i];
		
	}
	
	std::sort(sortablePerModelVals.begin(), sortablePerModelVals.end());
	for (int i=0; i<(int)sortablePerModelVals.size(); i++)
	{
		int modelID = sortablePerModelVals[i].id;
		m_ModelValueOrder.push_back(modelID);
		m_PerModelValue.push_back(-sortablePerModelVals[i].value);
	}
	
	// add model values to NBestModels
	for (int i=0; i<vkMin((int)sortablePerModelVals.size(), NBestModels); i++)
	{
		int modelID = sortablePerModelVals[i].id;
		modelIDs.push_back(modelID);
		FillModelValuesForModel(modelID);
	}
	
//	std::map<int, ModelPerVertexValues *>::iterator iter;
//	for (iter = m_mModelIDToPerVertexValues.begin(); 
//		 iter != m_mModelIDToPerVertexValues.end(); iter++)
//		iter->second->Interpolate(ModelValuesInterpolatorDescriptor());
}

void AppPartSelection::FillModelValuesForModel(int modelID)
{
	m_bModelPendingUpdate[modelID] = false;
	ModelDatabase * db = m_Pipeline.GetMainDB();
	
//	int NPoints = GetNPoints();
//	int NModels = m_Pipeline.GetMainDB()->NumModels();
	
	Model3D * model = m_Pipeline.GetMainDB()->GetModelByID(modelID);
	assert(model!=NULL);
	PointSet * modelPnts = model->PointCloud();
	ModelValue * modelValue = model->GetModelValue(m_sDefaultModelValue);
	if (modelValue==NULL)
	{
		modelValue = new ModelValue(model, m_sDefaultModelValue, false, false);
		model->AddModelValue(m_sDefaultModelValue, modelValue);
	}

//	double modelNorm = m_PerModelMax[modelID];
	double modelNorm = m_fGlobalMax;
	for (int j=0; j<modelPnts->NumPoints(); j++)	
	{
		ModelPoint pnt = modelPnts->GetPoint(j);
		//int pntGlobalID = CollectionDiffuse::MatrixPosition(pnt, NPoints, NModels);
		int pntGlobalID = db->GlobalID(pnt);
		modelValue->SetValue(pnt, m_afCumulativeSimilarities[pntGlobalID]/modelNorm);
	}
	
	std::map<int, ModelPerVertexValues *>::iterator iter;
	iter = m_mModelIDToPerVertexValues.find(modelID);
	if (iter==m_mModelIDToPerVertexValues.end())
	{
		m_mModelIDToPerVertexValues[modelID] = modelValue->CreateCopyOfPerVertexValues();
		//m_mModelIDToPerVertexValues[modelID]->Interpolate(ModelValuesInterpolatorDescriptor());
		m_mModelIDToPerVertexValues[modelID]->Interpolate();
	}
	else
	{
		modelValue->CopyPerVertexValuesTo(iter->second);
		//iter->second->Interpolate(ModelValuesInterpolatorDescriptor());
		iter->second->Interpolate();
	}
	
	// if label is selected - draw it into model value
	if (m_InteractionMode==APP_INTERACTIVE_LABELING && m_iCurrLabelID!=-1)
	{
		// create model value for new label if doesn't exist
		VKString modelValueName = m_sDefaultModelValue + "_" + m_CreatedLabels[m_iCurrLabelID];
		ModelValue * modelValueSingleLabel = model->GetModelValue(modelValueName);
		if (modelValueSingleLabel==NULL)
		{
			modelValueSingleLabel = modelValue->CreateCopyOfModelValue(modelValueName);
			assert(modelValueSingleLabel!=NULL);
			model->AddModelValue(modelValueName, modelValueSingleLabel);
		}
		else
		{
			modelValueSingleLabel->TakeMaxValues(modelValue);
		}
		if (!m_Pipeline.GetDB()->m_ModelValueNames.contains(modelValueName))
			m_Pipeline.GetDB()->m_ModelValueNames.push_back(modelValueName);
	}
}

void AppPartSelection::CreateFinalLabelsAssignMax()
{
	VKString newModelLabels = m_sDefaultModelValue+"_Labels";
	for (int modelID = 0; modelID < m_Pipeline.GetDB()->NumModels(); modelID++)
	{
		Model3D * model = m_Pipeline.GetDB()->GetModelByID(modelID);
		if (model->GetModelValue(newModelLabels)!=NULL)
			model->RemoveModelValue(newModelLabels);
		ModelValueLabel * newLabels = new ModelValueLabel(model, newModelLabels);

		std::vector<ModelValue*> modelValues;
		for (int labelID=0; labelID<m_CreatedLabels.count(); labelID++)	
		{
			VKString modelValueName = m_sDefaultModelValue + "_" + m_CreatedLabels[labelID];
			ModelValue * value = model->GetModelValue(modelValueName);
			modelValues.push_back(value);
		}
		
		PointSet * pntSet = model->PointCloud();
		for (int pntID=0; pntID<pntSet->NumPoints(); pntID++)
		{
			ModelPoint pnt = pntSet->GetPoint(pntID);
			int bestLabelID = -1;
			double bestLabelVal = 0;
			for (int labelID=0; labelID<m_CreatedLabels.count(); labelID++)
			{
				if (modelValues[labelID]!=NULL)
				{
					double val = modelValues[labelID]->Value(pnt);
					if (bestLabelID==-1 || bestLabelVal < val)
					{
						bestLabelVal = val;
						bestLabelID = labelID;
					}
				}
			}
			newLabels->AssignSegmentID(pnt, bestLabelID);
		}
		
		for (int labelID=0; labelID<m_CreatedLabels.count(); labelID++)
			newLabels->AssignLabel(labelID, m_CreatedLabels[labelID]);
		
		//newLabels->AssignFacesFromPointSet("auto", "auto");
		newLabels->AssignFaceLabels();
		model->AddModelValue(newModelLabels, newLabels);
	}

	if (!m_Pipeline.GetDB()->m_ModelValueNames.contains(newModelLabels))
		m_Pipeline.GetDB()->m_ModelValueNames.push_back(newModelLabels);
}

void AppPartSelection::CreateFinalLabelsMRFOptimization(bool refineFaces)
{	
#ifdef INCLUDE_MRFLIB	
	VKString newModelLabels = m_sDefaultModelValue+"_Labels";
//	int modelID = 3;
	for (int modelID = 0; modelID < m_Pipeline.GetDB()->NumModels(); modelID++)
	{
		std::vector<int> pointLabels;		
		Model3D * model = m_Pipeline.GetDB()->GetModelByID(modelID);
		PointSet * pntSet = model->PointCloud();

		// find model values
		std::vector<ModelValue*> modelValues;
		for (int labelID=0; labelID<m_CreatedLabels.count(); labelID++)	
		{
			VKString modelValueName = m_sDefaultModelValue + "_" + m_CreatedLabels[labelID];
			ModelValue * value = model->GetModelValue(modelValueName);
			modelValues.push_back(value);
		}
		
		DistanceEuclidean * distance = model->DefaultEuclideanDistance();
		assert(distance!=NULL);
		
		// Create MRF - construct graph
		MRFGeneralGraph * mrfGraph = new MRFGeneralGraph(pntSet->NumPoints(), 
														 m_CreatedLabels.count());
		MRFOptimization * mrfSolver = new MRFOptimization(mrfGraph);

		// go over all labels and assign node probabilities
		for (int pntID=0; pntID<pntSet->NumPoints(); pntID++)
		{
			int bestLabelID = -1;
			double bestLabelValue = 0;
			for (int labelID=0; labelID<m_CreatedLabels.count(); labelID++)
			{
				if (modelValues[labelID]!=NULL)
				{
					// for each point - set node and edge values
					ModelPoint pnt = pntSet->GetPoint(pntID);
					
					// fill node label
					double val = modelValues[labelID]->Value(pnt);
					if (val > bestLabelValue)
					{
						bestLabelValue = val;
						bestLabelID = labelID;
					}
					assert(val>=0 && val <=1.);
					mrfGraph->SetNodeCost(pntID, labelID, 1.-val);					
				}
			}
			if (bestLabelID==-1)
				bestLabelID = 0;
			pointLabels.push_back(bestLabelID);
		}
		
		PointDescriptorsContainer * pntDescr = NULL;
		if (m_iPointDescriptorID>=0
			&& m_iPointDescriptorID < m_Pipeline.GetDB()->m_PointDescriptorNames.count())
		{
			VKString descrName = m_Pipeline.GetDB()->m_PointDescriptorNames[m_iPointDescriptorID];
			pntDescr = pntSet->GetSetDescriptor(descrName);
			assert(pntDescr!=NULL);
		}

		// go over all edges and find neighbors
		for (int pntID=0; pntID<pntSet->NumPoints(); pntID++)
		{
			ModelPoint pnt = pntSet->GetPoint(pntID);
			R3Vector norm1 = pnt.Normal();			
			
			// fill edge similarity
			std::vector<ModelPoint> nhd;
			double maxDist = .2;
			//double sigma2 = pow(maxDist*.5, 2.);
			//double sigma2 = pow(maxDist*.5, 2.);
			distance->GetNeighborhood(pnt, maxDist, &nhd);
			double scale = (1. / (double)nhd.size()) * m_fSmoothnessScale;
			//double scale = m_fSmoothnessScale;
			for (int n=0; n<(int)nhd.size(); n++)
			{
				if (pntID < nhd[n].PntCloudID())		// only add once
				{
					double geomSim = 0;
					if (pntDescr==NULL)
					{
						geomSim = norm1.Dot(nhd[n].Normal());
					
						if (geomSim < 0)
							geomSim *= -1.;
						
						geomSim += .1;
					}
					else
						geomSim = pntDescr->PointSimilarity(pntID, nhd[n].PntCloudID());
					
					if (geomSim!=0)
					{
						//double w = exp(-distance->Distance(pntSet->GetPoint(pntID), nhd[n]) / sigma2);
						double w = 1.;
						mrfSolver->AddNeighbors(pntID, nhd[n].PntCloudID());
						mrfGraph->SetEdgeValue(pntID, nhd[n].PntCloudID(), w * scale * geomSim);
					}
				}
			}
		}
		
		mrfSolver->Cluster(&pointLabels);
		
		// assign labels to points
		ModelValueLabel * newLabels = new ModelValueLabel(model, newModelLabels);
		if (model->GetModelValue(newModelLabels)!=NULL)
			model->RemoveModelValue(newModelLabels);
		for (int pntID=0; pntID<(int)pointLabels.size(); pntID++)
		{
			ModelPoint pnt = pntSet->GetPoint(pntID);
			newLabels->AssignSegmentID(pnt, pointLabels[pntID]);
		}

		// optimize for face labels
		if (refineFaces)	// again perform full MRF optimization at face levels
		{
			//// Currently NOT functional. MRF library crushes.
			////		the reason is probably number of disconnected components.
			////		not clear, though...
			//// Potential TODO: see 1. CCs in graph, 2. connecte CCs, 3. use different MRF lib
			assert(false);
			std::vector<int> faceLabels;
			
			MRFGeneralGraph * mrfGraphFineFaces = new MRFGeneralGraph(model->GetMesh()->NFaces(), 
																	  m_CreatedLabels.count());
			MRFOptimization * mrfSolverFineFaces = new MRFOptimization(mrfGraphFineFaces);

			// fill node probabilities (for each label)
			for (int fID=0; fID < model->GetMesh()->NFaces(); fID++)
			{
				R3MeshFace * face = ((R3Mesh*)model->GetMesh())->Face(fID);
				R3Point fp = ((R3Mesh*)model->GetMesh())->FaceCentroid(face);
				
				ModelPoint pnt = distance->GetClosestPoint(fp); 
				int pntID = pnt.PntCloudID();
				assert(pntID>=0 && pntID<(int)pointLabels.size());
				faceLabels.push_back(pointLabels[pntID]);
				for (int labelID=0; labelID < m_CreatedLabels.count(); labelID++)
				{
					if (modelValues[labelID]!=NULL)
					{
						double val = modelValues[labelID]->Value(pnt);
						assert(val>=0 && val <=1.);
						mrfGraphFineFaces->SetNodeCost(fID, labelID, 1.-val);
					}
				}
			}
			
			// fill edge probabilities
			//model->CreateSingleCCForFaces();
			//assert(false);	// change NHD: what if large face + small faces
			model->CreateFaceNhd("FaceNHD", .05);
			
			//double scale = m_fSmoothnessScale / 3.;
			double scale = m_fSmoothnessScale;
			for (int fID=0; fID < model->GetMesh()->NFaces(); fID++)
			{
				R3MeshFace * face = ((R3Mesh*)model->GetMesh())->Face(fID);	
				R3Vector norm1 = ((R3Mesh*)model->GetMesh())->FaceNormal(face);
				std::vector<int> *nhd = NULL;
				std::vector<double> *weights = NULL;
				
				//model->SingleCCFaceNeighbors(fID, &nhd);
				model->FaceNhd(fID, &nhd, &weights);
				assert(nhd!=NULL && weights!=NULL);
				for (int nfid=0; nfid<(int)nhd->size(); nfid++)
				{
					if ((*nhd)[nfid] > fID)
					{
						R3MeshFace * nf = ((R3Mesh*)model->GetMesh())->Face((*nhd)[nfid]);
						R3Vector norm2 = ((R3Mesh*)model->GetMesh())->FaceNormal(nf);
						double geomSim = norm1.Dot(norm2);
						if (geomSim < 0)
							geomSim *= -1.;
						double nhdW = (*weights)[nfid];
						//double nhdW = 1.;
						//if (geomSim!=0)
						{
							mrfSolverFineFaces->AddNeighbors(fID, (*nhd)[nfid]);
							mrfGraphFineFaces->SetEdgeValue(fID, (*nhd)[nfid], scale * nhdW * geomSim);
						}
					}
				}
				
				
//				for (int nfid=0; nfid<3; nfid++)
//				{
//					R3MeshFace * nf = ((R3Mesh*)model->GetMesh())->FaceOnFace(face,nfid);
//					model->SingleCCFaceNeighbors();
//					if (nf!=NULL)
//					{
//						int neighborFaceID = ((R3Mesh*)model->GetMesh())->FaceID(nf);
//						assert(neighborFaceID>=0 
//							   && neighborFaceID < ((R3Mesh*)model->GetMesh())->NFaces());
//						if (neighborFaceID > fID)	// add only once
//						{
//							R3Vector norm2 = ((R3Mesh*)model->GetMesh())->FaceNormal(nf);
//							double geomSim = norm1.Dot(norm2);
//							if (geomSim < 0)
//								geomSim *= -1.;
//							if (geomSim!=0)
//							{
//								//if (fID<690)	//685 - works, 686 - doesn't (modelID = 2)
//								//{
//									//std::cout<<"\tAdding neighbr: "<<fID<<" -> "<<neighborFaceID<<" val="<<geomSim<<std::endl;
//									mrfSolverFineFaces->AddNeighbors(fID, neighborFaceID);
//									mrfGraphFineFaces->SetEdgeValue(fID, neighborFaceID, scale * geomSim);
//								//}
//							}							
//						}
//					}
//				}
			}
			
			// optimize for the best face labels
			//faceLabels.clear();
			mrfSolverFineFaces->Cluster(&faceLabels);
			
			for (int i=0; i<(int)faceLabels.size(); i++)
				newLabels->AssignSegmentIDToFace(i, faceLabels[i]);
			
			delete mrfSolverFineFaces;
			delete mrfGraphFineFaces;
		}
		else				// greedily assign face labels (based on the nearest pnt)
			//newLabels->AssignFacesFromPointSet("auto", "auto");
			newLabels->AssignFaceLabels();
		
		
		// label names for segments
		for (int labelID=0; labelID<m_CreatedLabels.count(); labelID++)
			newLabels->AssignLabel(labelID, m_CreatedLabels[labelID]);

		// add final labels object to the model
		model->AddModelValue(newModelLabels, newLabels);
		
		// clear solvers
		delete mrfSolver;
		delete mrfGraph;
		mrfSolver = NULL;
		mrfGraph = NULL;
	}
	
	if (!m_Pipeline.GetDB()->m_ModelValueNames.contains(newModelLabels))
		m_Pipeline.GetDB()->m_ModelValueNames.push_back(newModelLabels);
#else
	std::cout<<"[ERROR] MRF optimization library is not included"<<std::endl;
	assert(false);
#endif
}

void AppPartSelection::CreateFinalLabelsGlobalMRFOptimization()
{
#ifdef INCLUDE_MRFLIB
	std::cout<<"Global MRF Optimization"<<std::endl;
	FillPtrsFromParams();
	// construct global MRF graph
//	int NPoints = CollectionDiffuse::NPoints(m_Pipeline.GetMainDB());
//	int NModels = m_Pipeline.GetMainDB()->NumModels();
	ModelDatabase * db = m_Pipeline.GetMainDB();
	int nGlobalPoints = db->NGlobalPnts();
	MRFGeneralGraph * mrfGraph = new MRFGeneralGraph(nGlobalPoints, m_CreatedLabels.count(), 0);
	MRFOptimization * mrfSolver = new MRFOptimization(mrfGraph);
	double nodeMislabelPenalty = 1000.;	// something big - since it was labeled interactively
	
	// assign node values with 'infinite weight' based on original painting
	// NOTE : no penalty for most of the nodes
	std::vector<ModelPoint> regionForNorm;
	int nLabeled=0;
	std::map<int, std::map<int, std::vector<int> > >::iterator modelIter;
	for (modelIter = m_mModelPointLabel.begin();
		 modelIter != m_mModelPointLabel.end(); modelIter++)
	{
		Model3D * model = m_Pipeline.GetMainDB()->GetModelByID(modelIter->first);
		PointSet * pnts = model->PointCloud();
		//std::cout<<"Model = "<<model->ModelName().c_str()<<std::endl;
		for (std::map<int, std::vector<int> >::iterator pntIter = modelIter->second.begin(); 
			 pntIter != modelIter->second.end(); pntIter++)
		{
			ModelPoint p = pnts->GetPoint(pntIter->first);
			regionForNorm.push_back(p);
//			int nodeID = CollectionDiffuse::MatrixPosition(p, NPoints, NModels);
			int nodeID = db->GlobalID(p);
			
			// set all point cost to the penalty value
			for (int labelID = 0; labelID<(int)m_CreatedLabels.count(); labelID++)
				mrfGraph->SetNodeCost(nodeID, labelID, nodeMislabelPenalty);
			
			// for available labels assign penalty to 0
			for (int l = 0; l < (int)pntIter->second.size(); l++)
				mrfGraph->SetNodeCost(nodeID, pntIter->second[l], 0);
			nLabeled++;
		}
	}
	std::cout<<"\tFilled node potentials: non-zero nodes="<<nLabeled<<std::endl;
	//std::cout<<"LabeledPoints = "<<nLabeled<<std::endl;
	
	// normalize diffusion distances	
	//m_pInputCollection->SetRegionNormalization(regionForNorm);
	
	// TODO: potentially unite local geometry / corr edges w/ identity mapping
	
	// assign edge costs
	TimeProfiler edgePotentials;
	double maxDist = .2;
	for (int modelID=0; modelID<db->NumModels(); modelID++)
	{
		edgePotentials.WriteProgress("edgePot", modelID, db->NumModels());
		// prepare: find model / point set / distance
		Model3D * model = m_Pipeline.GetMainDB()->GetModelByID(modelID);
		ModelDistance * distance = model->DefaultEuclideanDistance();
		assert(distance!=NULL);
		PointSet * pnts = model->PointCloud();
		
		// check if local similarity term is available
		PointDescriptorsContainer * pntDescr = NULL;
		if (m_iPointDescriptorID>=0
			&& m_iPointDescriptorID < m_Pipeline.GetDB()->m_PointDescriptorNames.count())
		{
			VKString descrName = m_Pipeline.GetDB()->m_PointDescriptorNames[m_iPointDescriptorID];
			pntDescr = pnts->GetSetDescriptor(descrName);
			assert(pntDescr!=NULL);
		}
		
		// go through all points
		for (int pntID=0; pntID<pnts->NumPoints(); pntID++)
		{
			ModelPoint pnt = pnts->GetPoint(pntID);
			std::vector<ModelPoint> nhd;
			//int nodeID1 = CollectionDiffuse::MatrixPosition(pnt, NPoints, NModels);
			int nodeID1 = db->GlobalID(pnt);
			
			// assign edges based on local geometry	
			distance->GetNeighborhood(pnt, maxDist, &nhd);
			double scale = (1. / (double)nhd.size()) * m_fSmoothnessScale;
			for (int n=0; n<(int)nhd.size(); n++)
			{
				if (pntID < nhd[n].PntCloudID())		// only add once
				{
					double geomSim = 1.;
					if (pntDescr!=NULL)
						geomSim = pntDescr->PointSimilarity(pntID, nhd[n].PntCloudID());
					
					if (geomSim!=0 && scale!=0)
					{
						//int nodeID2 = CollectionDiffuse::MatrixPosition(nhd[n], NPoints, NModels);
						int nodeID2 = db->GlobalID(nhd[n]);
						mrfSolver->AddNeighbors(nodeID1, nodeID2);
						mrfGraph->SetEdgeValue(nodeID1, nodeID2, scale * geomSim);
					}
				}
			}
			
			// assign edges based on correspondence values
			nhd.clear();
			m_pInputCollection->GetCorrs(pnt, &nhd);
			scale = 1. / (double)nhd.size();
			for (int n=0; n<(int)nhd.size(); n++)
			{
				//int nodeID2 = CollectionDiffuse::MatrixPosition(nhd[n], NPoints, NModels);
				int nodeID2 = db->GlobalID(nhd[n]);
				if (nodeID1 < nodeID2 && nhd[n].GetModel() != model)		// only add once, don't add local
				{
					double corrSim = m_pInputCorrValue->Value(pnt, nhd[n]);
					mrfSolver->AddNeighbors(nodeID1, nodeID2);
					mrfGraph->SetEdgeValue(nodeID1, nodeID2, scale * corrSim);
				}
			}
		}
	}

	std::cout<<"\tMRF Optimization"<<std::endl;
	mrfGraph->CheckGraph();
	
	// optimize
	std::vector<int> pointLabels;
	mrfSolver->Cluster(&pointLabels);
	
	// create labels
	VKString newModelLabels = m_sDefaultModelValue+"_Labels";
	for (int modelID=0; modelID<db->NumModels(); modelID++)
	{
		Model3D * model = m_Pipeline.GetMainDB()->GetModelByID(modelID);
		PointSet * pnts = model->PointCloud();
		
		ModelValueLabel * newLabels = new ModelValueLabel(model, newModelLabels);
		
		if (model->GetModelValue(newModelLabels)!=NULL)
			model->RemoveModelValue(newModelLabels);
				
		for (int pntID=0; pntID<pnts->NumPoints(); pntID++)
		{
			ModelPoint pnt = pnts->GetPoint(pntID);
			//int globalID = CollectionDiffuse::MatrixPosition(pnt, NPoints, NModels);
			int globalID = db->GlobalID(pnt);
			assert(globalID < (int)pointLabels.size());
			newLabels->AssignSegmentID(pnt, pointLabels[globalID]);
		}
		
		newLabels->AssignFaceLabels();
		
		// label names for segments
		for (int labelID=0; labelID<m_CreatedLabels.count(); labelID++)
			newLabels->AssignLabel(labelID, m_CreatedLabels[labelID]);
		
		// add final labels object to the model
		model->AddModelValue(newModelLabels, newLabels);
	}

	// clear solvers
	delete mrfSolver;
	delete mrfGraph;
	
	if (!m_Pipeline.GetDB()->m_ModelValueNames.contains(newModelLabels))
		m_Pipeline.GetDB()->m_ModelValueNames.push_back(newModelLabels);
	
	m_pInputCollection->AutoNormalizeCorrs();	
#else
	std::cout<<"[ERROR] MRF optimization library is not included"<<std::endl;
	assert(false);
#endif
}

void AppPartSelection::SemanticGlueAutoAttachToAll()
{
	// find closest points for reference pair
		// models
	Model3D * model1 = m_Pipeline.GetMainDB()->GetModelInSubwindow(0);
	Model3D * model2 = m_pModelToBeGlued;
	PointSet * pnts1 = model1->PointCloud();
	PointSet * pnts2 = model2->PointCloud();
	std::vector<int> closestPairs;
	FillGluedCorrs(model1, model2, &closestPairs);
	
	assert(m_iTempArraySize >= (int)closestPairs.size()/2);
	
	for (int i=0; i<(int)closestPairs.size()/2; i++)
		m_aTempArrayWeights[i] = 1.;
	
	// find and store alignment
	m_fM1ToM2DiameterScale = -1;
	FindRigidAlignment(model1, model2, closestPairs, m_aTempArrayWeights);
	
	// for remaing models - align
	FillPtrsFromParams();
	
	std::map<const Model3D *, std::vector<int> > newCorrPairs;
	std::map<const Model3D *, std::vector<double> > newCorrPairsWeights;
	
	for (int i=0; i<(int)closestPairs.size()/2; i++)
	{
		ModelPoint oldP1 = pnts1->GetPoint(closestPairs[2*i+0]);
		std::vector<ModelPoint> corrs;
		m_pInputCollection->GetCorrs(oldP1, &corrs);
		
		for (int j=0; j<(int)corrs.size(); j++)
		{
			double val = m_pInputCorrValue->Value(oldP1, corrs[j]);
			newCorrPairs[corrs[j].GetModel()].push_back(corrs[j].PntCloudID());
			newCorrPairs[corrs[j].GetModel()].push_back(closestPairs[2*i+1]);
			newCorrPairsWeights[corrs[j].GetModel()].push_back(val);
		}
	}
	
	for (int modelID = 0; modelID < m_Pipeline.GetMainDB()->NumModels(); modelID++)
	{
		Model3D * newModel1 = m_Pipeline.GetMainDB()->GetModelByID(modelID);
		if (newModel1!=model1)
		{
			std::vector<int> & currModelPairs = newCorrPairs[newModel1];
			std::vector<double> & currModelWeights = newCorrPairsWeights[newModel1];
			assert(currModelPairs.size()/2 == currModelWeights.size());
			double * weightsPtr = m_aTempArrayWeights;
			for (int i=0; i<(int)currModelWeights.size(); i++)
			{
				weightsPtr[i] = currModelWeights[i];
			}
			FindRigidAlignment(newModel1, model2, currModelPairs, weightsPtr);
		}
	}
}

void AppPartSelection::FillCorrespondenceValues(Model3D * m1, std::vector<int> & pointsOnM1,
                                                std::map<Model3D*, std::vector<int> > 
                                                                                * savePairs,
                                                std::map<Model3D*, std::vector<double> >
                                                                                * saveWeights)
{
	PointSet * pnts1 = m1->PointCloud();
    for (int i=0; i<(int)pointsOnM1.size(); i++)
	{
		ModelPoint p1 = pnts1->GetPoint(pointsOnM1[i]);
		std::vector<ModelPoint> corrs;
		m_pInputCollection->GetCorrs(p1, &corrs);
		
		for (int j=0; j<(int)corrs.size(); j++)
		{
			double val = m_pInputCorrValue->Value(p1, corrs[j]);
			(*savePairs)[(Model3D*)corrs[j].GetModel()].push_back(pointsOnM1[i]);
			(*savePairs)[(Model3D*)corrs[j].GetModel()].push_back(corrs[j].PntCloudID());
			(*saveWeights)[(Model3D*)corrs[j].GetModel()].push_back(val);
		}
	}
}

void AppPartSelection::FillGluedCorrs(Model3D * model1, Model3D * model2, 
									  std::vector<int> * savePairs, 
									  R3Affine * saveT1, R3Affine * saveT2,
									  PointWithID ** savePntsArray1, int * saveNSelected1,
									  PointWithID ** savePntsArray2, int * saveNSelected2)
{
	// camera transformations that align models
//	model1->RenderingParams()->LoadMyCamera(0);
//	R4Matrix M1 = model1->RenderingParams()->CurrCameraMatrix();
//	R4Matrix M1inv = model1->RenderingParams()->CurrCameraInverseMatrix();
//	R3Affine T1(M1inv);
//	//R3Affine T1(R4identity_matrix);
//	model1->RenderingParams()->RestoreCameraAfterLoad(0);
//	
//	model2->RenderingParams()->LoadMyCamera(0);
//	R4Matrix M2 = model2->RenderingParams()->CurrCameraMatrix();
//	R4Matrix M2inv = model2->RenderingParams()->CurrCameraInverseMatrix();
//	R3Affine T2(M2inv);
//	//R3Affine T2(M1 * M2inv);
//	model2->RenderingParams()->RestoreCameraAfterLoad(0);
	R3Affine T1(R4identity_matrix);
	R3Affine & T2 = m_GlueDataModelPairToTransform[model2][model1];
	
	if (saveT1!=NULL)
		*saveT1 = T1;
	if (saveT2!=NULL)
		*saveT2 = T2;
	
	PointSet * pnts1 = model1->PointCloud();
	PointSet * pnts2 = model2->PointCloud();
	
	// fill temp arrays
	if (m_aTempArrayPnts1==NULL)
	{
		assert(m_aTempArrayPnts2==NULL);
		m_iTempArraySize = vkMin(100, m_Pipeline.GetMainDB()->NumModels()) * pnts2->NumPoints();
		m_aTempArrayPnts1 = new PointWithID[m_iTempArraySize];
		m_aTempArrayPnts2 = new PointWithID[m_iTempArraySize];
		m_aTempArrayWeights = new double[m_iTempArraySize];
	}
	
	// fill transformed pnts
	int nSelected1 = 0;
	int nSelected2 = 0;
	for (int pID1 = 0; pID1 < pnts1->NumPoints(); pID1++)
	{
		ModelPoint p1 = pnts1->GetPoint(pID1);
		if (PaintedOn(model1, p1.NearestVertex()))
		{
			assert(m_iTempArraySize > nSelected1);
			m_aTempArrayPnts1[nSelected1].pnt = p1.GetPosition();
			T1.Apply(m_aTempArrayPnts1[nSelected1].pnt);
			m_aTempArrayPnts1[nSelected1].id = pID1;
			R3Point tmp = m_aTempArrayPnts1[nSelected1].pnt;
			nSelected1++;
		}
	}
	
	for (int pID2 = 0; pID2 < pnts2->NumPoints(); pID2++)
	{
		ModelPoint p2 = pnts2->GetPoint(pID2);
		if (PaintedOn(model2, p2.NearestVertex()))
		{
			assert(m_iTempArraySize > nSelected2);
			m_aTempArrayPnts2[nSelected2].pnt = p2.GetPosition();
			T2.Apply(m_aTempArrayPnts2[nSelected2].pnt);
			m_aTempArrayPnts2[nSelected2].id = pID2;
			nSelected2++;
		}
	}
	
//	std::cout<<"Selected: "<<nSelected1<<" "<<nSelected2<<std::endl;

	if (savePntsArray1!=NULL)
		{*savePntsArray1 = m_aTempArrayPnts1;}
	if (savePntsArray2!=NULL)
		{*savePntsArray2 = m_aTempArrayPnts2;}
	if (saveNSelected1!=NULL)
		{*saveNSelected1 = nSelected1;}
	if (saveNSelected2!=NULL)
		{*saveNSelected2 = nSelected2;}
	
	// fill best distances
	int * set1To2 = new int[nSelected1];
	int * set2To1 = new int[nSelected2];
	double * set2To1Dists = new double[nSelected2];
	
	for (int i=0; i<nSelected2; i++)
	{
		set2To1[i] = -1;
		set2To1Dists[i] = FLT_MAX;
	}
	
	for (int pID1 = 0; pID1 < nSelected1; pID1++)
	{
		int bestID = -1;
		double bestDist = FLT_MAX;
		
		for (int pID2 = 0; pID2 < nSelected2; pID2++)
		{
			double dist = R3Distance(m_aTempArrayPnts1[pID1].pnt, 
									 m_aTempArrayPnts2[pID2].pnt);
			if (dist < bestDist)
			{
				//std::cout<<"dist = "<<dist<<" best="<<pID1<<std::endl;
				bestDist = dist;
				bestID = pID2;
			}
			if (dist < set2To1Dists[pID2])
			{
				set2To1Dists[pID2] = dist;
				set2To1[pID2] = pID1;
			}
		}
		assert(bestID>=0);
		set1To2[pID1] = bestID;
	}

	// fill MCN
	assert(savePairs!=NULL);
	for (int i = 0; i<nSelected1; i++)
	{
		//std::cout<<"Blue->Green: "<<i<<" -> "<<set1To2[i]<<std::endl;
		//std::cout<<"Green->Blue: "<<set1To2[i]<<" -> "<<set1To2[i]<<std::endl;
		if (i==set2To1[set1To2[i]])
		{
			savePairs->push_back(m_aTempArrayPnts1[i].id);
			savePairs->push_back(m_aTempArrayPnts2[set1To2[i]].id);
			//std::cout<<"Corr: "<<m_aTempArrayPnts1[i].id<<" -> ";
			//std::cout<<m_aTempArrayPnts2[set1To2[i]].id<<std::endl;
		}
	}
	
	delete [] set1To2;
	delete [] set2To1;
	delete [] set2To1Dists;
	
}

void AppPartSelection::FindRigidAlignment(Model3D * m1, Model3D * m2, 
										  std::vector<int> & corrPairs,
										  double * weights)
{
	// rigidly align reference pair (ICP?)
	assert(corrPairs.size()%2==0);
//	assert(corrPairs.size()/2==weights.size());
	PointSet * pnts1 = m1->PointCloud();
	PointSet * pnts2 = m2->PointCloud();
		
	RNArray<R3Point*> rnArrayFrom;
	RNArray<R3Point*> rnArrayTo;
	
//	std::cout<<"Models: "<<m1->ModelID()<<" -> "<<m2->ModelID()<<std::endl;
//	std::cout<<"\tarray = "<<m_iTempArraySize<<" pairs: "<<corrPairs.size()<<std::endl;
	assert(m_iTempArraySize >= (int)corrPairs.size()/2);
	
	bool preScaled = (m_iSemanticGlueAlignmentFlags & GLUE_SCALE_MODEL_DIAMETER) != 0;
	double preScaleValue = 1;
	if (preScaled)
	{
		if (m_fM1ToM2DiameterScale>0)
		{
			R3Point c1 = m1->GetMesh()->Centroid();
			R3Point c2 = m2->GetMesh()->Centroid();
			double rad1= m1->GetMesh()->AverageRadius(&c1);
			double rad2= m2->GetMesh()->AverageRadius(&c2);
			preScaleValue = rad1 / rad2;
			//preScaleValue *= m_fM1ToM2DiameterScale;
		}
		else
		{
			m_fM1ToM2DiameterScale = m_fMainPairScale;
			preScaleValue = m_fM1ToM2DiameterScale;
		}
	}
	else if (m_fM1ToM2DiameterScale <= 0)
	{
		m_fM1ToM2DiameterScale = m_fMainPairScale;
		preScaleValue = m_fM1ToM2DiameterScale;
	}
	else
		preScaleValue = m_fM1ToM2DiameterScale;
	
	for (int i=0; i<(int)corrPairs.size()/2; i++)
	{
		m_aTempArrayPnts1[i].pnt = pnts1->GetPoint(corrPairs[2*i+0]).GetPosition();
		m_aTempArrayPnts2[i].pnt = preScaleValue * pnts2->GetPoint(corrPairs[2*i+1]).GetPosition();
		
//		std::cout<<"corr: "<<corrPairs[2*i+0]<<" -> "<<corrPairs[2*i + 1]<<std::endl;
//		std::cout<<"\tpts1=["<<m_aTempArrayPnts1[i].pnt.X()<<" "<<m_aTempArrayPnts1[i].pnt.Y()<<" ";
//		std::cout<<m_aTempArrayPnts1[i].pnt.Z()<<std::endl;;
//		std::cout<<"\tpts2=["<<m_aTempArrayPnts2[i].pnt.X()<<" "<<m_aTempArrayPnts2[i].pnt.Y()<<" ";
//		std::cout<<m_aTempArrayPnts2[i].pnt.Z()<<std::endl;;
		rnArrayFrom.Insert(&(m_aTempArrayPnts2[i].pnt));
		rnArrayTo.Insert(&(m_aTempArrayPnts1[i].pnt));
	}
	
	bool scaleFromAlign = (m_iSemanticGlueAlignmentFlags & GLUE_SCALE_FROM_ALIGNMENT)!=0;
//	std::cout<<"scale = "<<(!preScaled)<<" scaleVal="<<preScaleValue<<std::endl;
//	std::cout<<"scaleFromAlign="<<scaleFromAlign<<std::endl;
	R4Matrix xFormMtx = R3AlignPoints(rnArrayTo, rnArrayFrom, weights, 
									  true, true, scaleFromAlign);
//	R4Matrix xFormMtx = SinglePairRigidAlign::R3AlignPointsInXYPlane(
//												rnArrayTo, rnArrayFrom, weights, 
//												true, true, scaleFromAlign);
	
	
//	std::cout<<"xform="<<std::endl;
//	for (int i=0; i<4; i++)
//	{
//		for (int j=0; j<4; j++)
//			std::cout<<xFormMtx[i][j]<<" ";
//		std::cout<<std::endl;
//	}

	// ICP?
	
	// store matrix
	R4Matrix preScaleMtx = R4identity_matrix;
	preScaleMtx.Scale(preScaleValue);
	m_GlueDataModelPairToTransform[m2][m1] = xFormMtx * preScaleMtx;
//	R3Affine inv = R3Affine(m_GlueDataModelPairToTransform[m2][m1].InverseMatrix());
//	m_GlueDataModelPairToTransform[m1][m2] = inv; 
	
}

void AppPartSelection::ClearFinalLabels()
{
	ClearCreatedModelValue("Labels");
}

void AppPartSelection::ClearAllCreatedModelValues()
{
	m_mModelPointLabel.clear();
	
	while (m_CreatedLabels.count()>0)
		ClearCreatedModelValue(m_CreatedLabels[0]);
}

void AppPartSelection::ClearCreatedModelValue(const VKString & modelValueName)
{
	VKString realName = m_sDefaultModelValue + "_" + modelValueName;
	for (int modelID = 0; modelID < m_Pipeline.GetDB()->NumModels(); modelID++)
	{
		Model3D * model = m_Pipeline.GetDB()->GetModelByID(modelID);
		if (model->GetModelValue(realName)!=NULL)
			model->RemoveModelValue(realName);
	}
	if (m_Pipeline.GetDB()->m_ModelValueNames.contains(realName))
		m_Pipeline.GetDB()->m_ModelValueNames.remove(realName);
	if (m_CreatedLabels.contains(modelValueName))
		m_CreatedLabels.remove(modelValueName);
}


void AppPartSelection::Draw(DatabaseVisWindow * window)
{
    // prepare intersection buffers
    //if (m_InteractionMode==APP_DRAW_DECAL)
    {
        Model3D * model = m_Pipeline.GetDB()->GetModelInSubwindow(0);
        assert(model!=NULL);
        assert(m_EnabledQuickIntersectionBuffers.size() < 2);
        if (m_EnabledQuickIntersectionBuffers.size()==1 
            && m_EnabledQuickIntersectionBuffers[0] != model)
        {
            m_EnabledQuickIntersectionBuffers[0]->FastIntersectionBuffer(false);
            m_EnabledQuickIntersectionBuffers.clear();
        }
        
        if (m_EnabledQuickIntersectionBuffers.size()==0)
        {
            m_EnabledQuickIntersectionBuffers.push_back(model);
            m_EnabledQuickIntersectionBuffers[0]->FastIntersectionBuffer(true);
        }
    }
    
    // check if is in process of gluing - interupt if necessary
    if (DrawGluingApp(window))
        return;
	
	if (m_bWritingNewLabel)
		window->GetRenderer()->RenderText(R2Point(0, 0), VKString("Label: ")+m_sCurrLabel,
										  1., 0., 0., VisWindowRenderer::TEXT_SIZE_LARGE);
	
	// Render selected region
	int modelID = m_Pipeline.GetDB()->GetRenderParams()->m_mWindowToModelID[0];
	Model3D * model = m_Pipeline.GetDB()->GetModelByID(modelID); 
	DrawMesh(window, model, 0);
	
    // Draw decal details
    if (m_InteractionMode==APP_DRAW_DECAL)
        DrawFineDetailPainting(window, m_Pipeline.GetDB()->GetModelInSubwindow(0), 0);
	
	// Render corresponding parts
	if (GetSelectedModel()==-1 || GetRefSelectedPnts().size()==0)
		return;
	
	int nWindows = (int)m_Pipeline.GetDB()->GetRenderParams()->m_mWindowToModelID.size();

    UpdateSimilaritiesToPainted(window);
	
	// Render parts
	for (int subWindowID = 1; subWindowID < nWindows; subWindowID++)
	{
		modelID = m_Pipeline.GetDB()->GetRenderParams()->m_mWindowToModelID[subWindowID];
		model = m_Pipeline.GetDB()->GetModelByID(modelID);
		DrawMesh(window, model, subWindowID);
	}    
}

void AppPartSelection::DrawFineDetailPainting(DatabaseVisWindow * window, 
                                             Model3D * model, int subwindowID)
{
    //TODO: leave only code in DrawGluingApp
    model->RenderingParams()->LoadMyCamera(subwindowID);
    
    // draw points with given brush size
    double r = model->RenderingParams()->GetStandardRadius();
    for (int i=0; i<(int)m_BrushStrokes.size(); i++)
    {
        R3Point p = m_BrushStrokes[i].m_Pnt.GetPosition();
        double brushSize = m_BrushStrokes[i].m_fBrushSize * r;
        RNRgb & brushColor = m_BrushStrokes[i].m_BrushColor;
        glEnable(GL_LIGHTING);
        glColor3d(brushColor.R(), brushColor.G(), brushColor.B());
        brushSize *= 10;
        R3Sphere(p, brushSize).Draw();
    }
    
    model->RenderingParams()->RestoreCameraAfterLoad(subwindowID);
}

bool AppPartSelection::DrawGluingApp(DatabaseVisWindow * window)
{
    bool retVal = false;
    ModelDatabase * db = m_Pipeline.GetMainDB();
    ModelDatabaseRenderingParams * dbRender = db->GetRenderParams();
    int nWindows = (int)dbRender->m_mWindowToModelID.size();
    if (m_InteractionMode == APP_DRAW_DECAL)
    {
        for (int subWindowID = 0; subWindowID < nWindows; subWindowID++)
        {
            if ((m_iDecalTransferFlags & DECAL_DRAW_TRANSFERED) == 0
                && subWindowID>0)
                break;
            Model3D * model = m_Pipeline.GetMainDB()->GetModelInSubwindow(subWindowID);
            
            model->RenderingParams()->LoadMyCamera(subWindowID);
            if ((m_iDecalTransferFlags & DECAL_DRAW_MESH) != 0
                || (m_iDecalTransferFlags & DECAL_DRAW_MESH_REGION) != 0)
                DrawMesh(window, model, subWindowID, NULL, false);

            if ((m_iDecalTransferFlags & DECAL_DRAW_DECAL) != 0
                && m_BrushStrokes.size() > 0)
            {
                // draw points with given brush size
                double r = model->RenderingParams()->GetStandardRadius();
                R3Affine xForm = R4identity_matrix;
                Model3D * mainModel = (Model3D*)m_BrushStrokes[0].m_Pnt.GetModel();
                if (mainModel!=model)
                    xForm = m_GlueDataModelPairToTransform[mainModel][model];

                for (int i=0; i<(int)m_BrushStrokes.size(); i++)
                {
                    R3Point p = m_BrushStrokes[i].m_Pnt.GetPosition();

                    double brushSize = m_BrushStrokes[i].m_fBrushSize * r;
                    RNRgb & brushColor = m_BrushStrokes[i].m_BrushColor;
                    glEnable(GL_LIGHTING);
                    //glColor3d(brushColor.R(), brushColor.G(), brushColor.B());
					static GLfloat material[4];
					material[0] = brushColor.R();
					material[1] = brushColor.G();
					material[2] = brushColor.B();
					material[3] = 1.;
					glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, material); 
                    brushSize *= 10;
                    xForm.Apply(p);
                    R3Sphere(p, brushSize).Draw();
                }
            }

            model->RenderingParams()->RestoreCameraAfterLoad(subWindowID);
        }

        retVal = true;
    }
    else
    {
        if (m_SemanticGlueState==GLUE_INTERACTIVE_ALIGNMENT)
        {
            assert(m_pDBWithToBeGluedModel!=NULL && m_pModelToBeGlued!=NULL);
            //DrawMesh(window, m_Pipeline.GetMainDB()->GetModelInSubwindow(0), 0);
            //DrawMesh(window, m_pModelToBeGlued, 0);
            Model3D * model = m_Pipeline.GetMainDB()->GetModelInSubwindow(0);
            R3Affine & aff = m_GlueDataModelPairToTransform[m_pModelToBeGlued][model];
            model->RenderingParams()->LoadMyCamera(0);
            DrawMesh(window, model, 0, NULL, false);
            DrawMesh(window, m_pModelToBeGlued, 0, &aff, false);
            DrawAlignmentEdges(window);
            model->RenderingParams()->RestoreCameraAfterLoad(0);	
            retVal = true;
        }
        else if (m_SemanticGlueState==GLUE_PROPAGATED_TRANSFORMS)
        {
            //std::cout<<"Attempting to draw xforms"<<std::endl;
            assert(m_pDBWithToBeGluedModel!=NULL && m_pModelToBeGlued!=NULL);
            std::map<Model3D *, std::map<Model3D *, R3Affine> >::iterator iter;
            iter = m_GlueDataModelPairToTransform.find(m_pModelToBeGlued);
            assert(iter != m_GlueDataModelPairToTransform.end());	// at least one
            int nWindows = (int)dbRender->m_mWindowToModelID.size();
            for (int subWindowID = 0; subWindowID < nWindows; subWindowID++)
            {
                int modelID = m_Pipeline.GetMainDB()->GetRenderParams()->m_mWindowToModelID[subWindowID];
                Model3D * model = m_Pipeline.GetMainDB()->GetModelByID(modelID);
                
                std::map<Model3D *, R3Affine>::iterator iter2 = iter->second.find(model);
                
                if (iter2==iter->second.end())
                    continue;
                
                model->RenderingParams()->LoadMyCamera(subWindowID);
                DrawMesh(window, model, subWindowID, NULL, false);
                DrawMesh(window, m_pModelToBeGlued, subWindowID, &(iter2->second), false);
                model->RenderingParams()->RestoreCameraAfterLoad(subWindowID);
                retVal = true;
            }
        }
		else if (m_GlueDataModelPairToTransform.size() > 0)
		{
//			std::cout<<"Rendering..."<<std::endl;
            Model3D * model = m_Pipeline.GetMainDB()->GetModelInSubwindow(0);
			model->RenderingParams()->LoadMyCamera(0);
			
			DrawMesh(window, model, 0, NULL, false);
			
			//int nIter = vkMin((int)m_aiSimilarModels.size(), 3);
			//for (int mID=0; mID<m_Pipeline.GetMainDB()->NumModels(); mID++)
			//for (int i=0; i<nIter; i++)
			ModelDatabase * db = m_Pipeline.GetMainDB();
			int subwindows = (int)db->GetRenderParams()->m_mWindowToModelID.size();
			for (int mID = 0; mID < subwindows; mID++)
			{
				//int mID = m_aiSimilarModels[i];
				//Model3D * m2 = m_Pipeline.GetMainDB()->GetModelByID(mID);
				Model3D * m2 = m_Pipeline.GetMainDB()->GetModelInSubwindow(mID);
				if (m2==model)
					continue;
				std::map<Model3D *, std::map<Model3D *, R3Affine> >::iterator iter1;
				iter1 = m_GlueDataModelPairToTransform.find(m2);
				if (iter1 != m_GlueDataModelPairToTransform.end())
				{
					std::map<Model3D*, R3Affine>::iterator iter2 = iter1->second.find(model);
					if (iter2!=iter1->second.end())
					{
						//std::cout<<"Drawing: "<<m2->ModelName().c_str()<<std::endl;
						DrawMesh(window, m2, 0, &(iter2->second), false);
					}
				}
			}
			model->RenderingParams()->RestoreCameraAfterLoad(0);	
			retVal = false;
		}
    }
    return retVal;
}


void AppPartSelection::DrawAlignmentEdges(DatabaseVisWindow * window)
{
	assert(m_pDBWithToBeGluedModel!=NULL && m_pModelToBeGlued!=NULL);
	Model3D * m1 = m_Pipeline.GetMainDB()->GetModelInSubwindow(0);
	Model3D * m2 = m_pModelToBeGlued;
	PointSet * pnts1 = m1->PointCloud();
	PointSet * pnts2 = m2->PointCloud();
	std::vector<int> savePairs;
	R3Affine T1;
	R3Affine T2;
	
	PointWithID * transformed1;
	int nTransformed1 = 0;
	PointWithID * transformed2;
	int nTransformed2 = 0;
	
	FillGluedCorrs(m1, m2, &savePairs, &T1, &T2, 
				   &transformed1, &nTransformed1, &transformed2, &nTransformed2);

	glDisable(GL_LIGHTING);
	for (int corrID = 0; corrID < savePairs.size()/2; corrID++)
	{
		int p1ID = savePairs[2*corrID+0];
		int p2ID = savePairs[2*corrID+1];
		
		R3Point pos1 = pnts1->GetPoint(p1ID).GetPosition();
		R3Point pos2 = pnts2->GetPoint(p2ID).GetPosition();
		T1.Apply(pos1);
		T2.Apply(pos2);

		glColor3d(.5, .5, 0);
		glBegin(GL_LINES);
		glVertex3d(pos1.X(), pos1.Y(), pos1.Z());
		glVertex3d(pos2.X(), pos2.Y(), pos2.Z());
		glEnd();
	}
	
	glPointSize(4.);	
	glBegin(GL_POINTS);
	glColor3d(0, 0, 1);
	for (int i=0; i<nTransformed1; i++)
	{
		//std::cout<<"Drawing1 "<<transformed1[i].id<<std::endl;
		R3Point pos = pnts1->GetPoint(transformed1[i].id).GetPosition();
		T1.Apply(pos);
		glVertex3d(pos.X(), pos.Y(), pos.Z());
		//glVertex3d(transformed1[i].pnt.X(), transformed1[i].pnt.Y(), transformed1[i].pnt.Z());
	}
	glColor3d(0, 1, 0);
	for (int i=0; i<nTransformed2; i++)
	{
		R3Point pos = pnts2->GetPoint(transformed2[i].id).GetPosition();
		T2.Apply(pos);
		glVertex3d(pos.X(), pos.Y(), pos.Z());
		//std::cout<<"Drawing2 "<<transformed2[i].id<<std::endl;
		//glVertex3d(transformed2[i].pnt.X(), transformed2[i].pnt.Y(), transformed2[i].pnt.Z());
	}
	glEnd();
}

void AppPartSelection::DrawMesh(DatabaseVisWindow * window, Model3D * model, 
								int subwindowID, R3Affine * xform, bool autoHandleCameras)
{
	if (!m_bDrawMeshes)
		return;
	
	bool additionalModel = model->GetDB()!=m_Pipeline.GetDB();
	
	int modelID = model->ModelID();
	
	ModelPerVertexValues * perVertexVals = NULL;
	
	// TODO: check if subwindowID = 0 mesh changed
	if (!additionalModel)
	{
		if (subwindowID!=0 && m_aiSimilarModels.size()==0)
			return;
		
		if (subwindowID!=0 && m_Pipeline.GetDB()!=m_Pipeline.GetMainDB())
			return;

		// Fill values for new model
		
		if (m_bModelPendingUpdate[modelID])
			FillModelValuesForModel(modelID);
		
		if (subwindowID!=0 && 
			m_mModelIDToPerVertexValues.find(modelID)==m_mModelIDToPerVertexValues.end())
			return;
	}
		
	if (autoHandleCameras)
		model->RenderingParams()->LoadMyCamera(subwindowID);
	
	if (!additionalModel && subwindowID>0)
		perVertexVals = m_mModelIDToPerVertexValues[modelID];

	glEnable(GL_LIGHTING);
    glEnable(GL_POLYGON_SMOOTH);
    glShadeModel(GL_SMOOTH);
	glDisable(GL_CULL_FACE);
	glColor3d(1.,1., 1.);
	glDisable(GL_TEXTURE);
	const R3Mesh * mesh = model->GetMesh();
	
	static GLfloat material1[4];
	static GLfloat material2[4];

	//window->StandardSurfaceColor(material1);
	material1[0] = 1;
	material1[1] = 1;
	material1[2] = 1;
	material1[3] = 1;
	if (m_iCurrLabelID==-1)
	{
//		material2[0] = 1;
//		material2[1] = 0;
//		material2[2] = 0;
//		material2[3] = material1[3];
		material2[0] = .1;
		material2[1] = .19;
		material2[2] = .55;
		material2[3] = material1[3];
	}
	else
	{
		//window->MapIntToColor(m_iCurrLabelID, surfaceMaterial2);
		double r, g, b;
		window->MapIntToRadicalColor(m_iCurrLabelID, r, g, b);
		material2[0] = r;
		material2[1] = g;
		material2[2] = b;
		material2[3] = material1[3];
	}
    
	if (m_GlueDataModelPairToTransform.size()>0)
	{
		glBegin(GL_TRIANGLES);
		for (int i=0; i<mesh->NFaces(); i++)
		{
			const R3MeshFace * face = mesh->Face(i);
			const R3MeshVertex * v1 = mesh->VertexOnFace(face, 0);
			const R3MeshVertex * v2 = mesh->VertexOnFace(face, 1);
			const R3MeshVertex * v3 = mesh->VertexOnFace(face, 2);
			assert(v1!=NULL && v2!=NULL && v3!=NULL);
			DrawMeshVertex(subwindowID, perVertexVals, material1, material2, model, v1, xform);
			DrawMeshVertex(subwindowID, perVertexVals, material1, material2, model, v2, xform);
			DrawMeshVertex(subwindowID, perVertexVals, material1, material2, model, v3, xform);		
		}
		glEnd();
	}
	else
	{
		PointSet * pts = model->PointCloud();
		double r = model->RenderingParams()->GetStandardRadius()*2;
		std::vector<int> & selpts = GetRefSelectedPnts();		
		if (modelID==GetSelectedModel())
		{
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, material2); 
			for (int i=0; i<(int)selpts.size(); i++)
				R3Sphere(pts->GetPoint(selpts[i]).GetPosition(), r).Draw();
		}
		else
		{
			if (m_afCumulativeSimilarities!=NULL)
			{
				//int nNorm = (int)selpts.size();
				for (int i=0; i<(int)pts->NumPoints(); i++)
				{
					int gID = m_Pipeline.GetDB()->GlobalID(pts->GetPoint(i));
					double val = m_afCumulativeSimilarities[gID] / m_fGlobalMax;
					if (val>0)
					{
						double rv = vkMin(1., val);
						material2[0] = 1.;
						material2[1] = 1. - rv;
						material2[2] = 1. - rv;
						glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, material2);
						R3Sphere(pts->GetPoint(i).GetPosition(), (r/1.9 + r * val)).Draw();
					}
				}
			}
		}
	}
	
	if (autoHandleCameras)
		model->RenderingParams()->RestoreCameraAfterLoad(subwindowID);
}

void AppPartSelection::DrawMeshVertex(int subwindow, ModelPerVertexValues * perVertexVals,
									  GLfloat surfaceMaterial1[4], GLfloat surfaceMaterial2[4], 
									  Model3D * model, const R3MeshVertex * vtx, R3Affine * xform)
{
	static GLfloat surfaceMaterialI[4];
	const R3Mesh * mesh = model->GetMesh();
	int vID = mesh->VertexID(vtx);
	R3Point vPos = mesh->VertexPosition(vtx);
	R3Vector vNorm = mesh->VertexNormal(vtx);

	// draw faces
	if (subwindow==0 || perVertexVals==NULL)
	{
		if (PaintedOn(model, vID))
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial2); 
		else
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial1); 
	}
	else 
	{
		assert(perVertexVals!=NULL);
		double vVal = perVertexVals->GetValueAtVertex(vID);
		if (m_InteractionMode==APP_PART_SELECTION)
		{
			if (perVertexVals->GetValueAtVertex(vID) > m_fShowPartThreshold)
				glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial2);
			else
				glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial1);
		}
		else if (m_InteractionMode==APP_INTERACTIVE_LABELING)
		{
			vVal = pow(vVal, m_fShowPartThreshold);
			
			for (int i=0; i<4; i++)
				surfaceMaterialI[i] = vVal * surfaceMaterial2[i] + (1-vVal) * surfaceMaterial1[i];
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterialI); 
		}
		else
			assert(false);
	}
	if (xform!=NULL)
	{
		xform->Apply(vPos);
		xform->Apply(vNorm);
	}
	glNormal3d(vNorm.X(), vNorm.Y(), vNorm.Z());
	glVertex3d(vPos.X(), vPos.Y(), vPos.Z());
}

void AppPartSelection::UpdateSimilaritiesToPainted(DatabaseVisWindow * window)
{
	if (m_bNeedsUpdate)
	{
		m_bNeedsUpdate = false;
		if (m_Pipeline.GetMainDB()!=m_Pipeline.GetDB())
		{
			std::cout<<"[WARNING] This DB cannot be updated (no correspondences)"<<std::endl;
		}
		else if (m_FacetsPerModelValues.size() > 0)
		{
			window->m_NewModelOrder = m_aiSimilarModels;
			window->m_iCurrModelTraversalID = 0;
			window->FillMostSimilarModels(window->m_iCurrModelTraversalID, window->m_NewModelOrder);
		}
		else
		{
			m_aiSimilarModels.clear();
			
			// fill similarities
			FillCumulativeSimilarities();
            
			// find models / parts from similarities (fill ModelValue as well)
			if (m_InteractionMode==APP_INTERACTIVE_LABELING)
				FillModelValuesMostSimilarParts(m_aiSimilarModels, m_Pipeline.GetDB()->NumModels());
			else
				FillModelValuesMostSimilarParts(m_aiSimilarModels, m_iShowNumParts);
			
			assert(m_PerModelValue.size()==m_PerModelMax.size()
				   && m_aiSimilarModels.size()==m_PerModelMax.size());
			std::cout<<"Model : sum \t max"<<std::endl;
			for (int i=0; i<(int)m_aiSimilarModels.size(); i++)
			{
				int sID = ((int)m_aiSimilarModels.size())-1 - i;
				int mID = m_aiSimilarModels[sID];
				std::cout<<"\t"<<m_Pipeline.GetDB()->GetModelByID(mID)->ModelName().c_str();
				std::cout<<" : "<<m_PerModelValue[sID]<<" "<<m_PerModelMax[mID]<<std::endl;
			}
			
			window->m_NewModelOrder = m_aiSimilarModels;
			window->m_iCurrModelTraversalID = 0;
			window->FillMostSimilarModels(window->m_iCurrModelTraversalID, window->m_NewModelOrder);
			
			// re-render
			window->GetRenderer()->UpdateScreen();
		}
	}
}

void AppPartSelection::WriteResultsChild(std::ofstream & textStream)
{

}

bool AppPartSelection::ReadResultsChild(std::ifstream & textStream)
{
	ReadInteractionResults();
}

void AppPartSelection::ScaleAffine(Model3D * m1, Model3D * m2, double dy)
{
	R3Affine & aff = m_GlueDataModelPairToTransform[m2][m1];
	//R3Point p(0,0,0);		aff.Apply(p);		R3Vector pv(p.X(), p.Y(), p.Z());
	//aff.Translate(-pv);
	//aff.Scale(1. + dy);
	//aff.Translate(pv);
	//
	m_fMainPairScale *= (1. + dy);
	aff = R4identity_matrix;
	aff.Scale(m_fMainPairScale);
	R3Affine another = m_MainPairCoords.Matrix();
	another.Apply(aff);
}

void AppPartSelection::RotateAffine(Model3D * m1, Model3D * m2, double dx, double dy)
{
//	R3Affine & aff = m_GlueDataModelPairToTransform[m2][m1];
//	R3Point p=m2->GetMesh()->Centroid();		aff.Apply(p);		R3Vector pv(p.X(), p.Y(), p.Z());
	
	m1->RenderingParams()->LoadMyCamera(0);
	R3Viewer * v1 = m1->RenderingParams()->GetViewer(0);
	R3Vector camUp = v1->Camera().Up();
	R3Vector camRight = v1->Camera().Right();
	R3Vector camTow = v1->Camera().Towards();
	m1->RenderingParams()->RestoreCameraAfterLoad(0);
	RNAngle theta = -4.0 * (fabs(dx) + fabs(dy));
    R3Vector vector = (camRight * dx) + (camUp * dy);
    R3Vector axis = camTow % vector;
    axis.Normalize();
	
	m_MainPairCoords.Rotate(axis, theta);	
	ScaleAffine(m1, m2, 0);
	
//	aff.Translate(-pv);
//	aff.Rotate(axis, theta);
//	aff.Translate(pv);
	
	// aff.Rotate(const R3Vector& vector, RNAngle radians);
}

void AppPartSelection::TranslateAffine(Model3D * m1, Model3D * m2, double dx, double dy)
{
	R3Affine & aff = m_GlueDataModelPairToTransform[m2][m1];
	m1->RenderingParams()->LoadMyCamera(0);
	R3Viewer * v1 = m1->RenderingParams()->GetViewer(0);
	R3Vector camUp = v1->Camera().Up();
	R3Vector camRight = v1->Camera().Right();
	m1->RenderingParams()->RestoreCameraAfterLoad(0);

	R3Vector vector = 2 * (camRight * dx) + (camUp * dy);
	
	R3Point ctr = m1->GetMesh()->Centroid();
	double scale1 = m1->GetMesh()->AverageRadius(&ctr);

	ctr = m2->GetMesh()->Centroid();
	double scale2 = m2->GetMesh()->AverageRadius(&ctr);
	
	m_MainPairCoords.Translate(vector * scale2);	
	ScaleAffine(m1, m2, 0);

	//aff.Translate(vector);
}

void AppPartSelection::InitializeFineDetailState()
{
    m_iDEFAULT_DECALDRAW_EVERYTHING  = DECAL_DRAW_DECAL | DECAL_DRAW_MESH;
    m_iDEFAULT_DECALDRAW_EVERYTHING |= DECAL_DRAW_MESH_REGION | DECAL_DRAW_TRANSFERED;
    
    m_iDEFAULT_DECALDRAW_ONLY_DECAL  = DECAL_DRAW_DECAL | DECAL_DRAW_TRANSFERED;

    m_iDEFAULT_DECALDRAW_CLEAN_DECAL  = DECAL_DRAW_DECAL | DECAL_DRAW_TRANSFERED;
    m_iDEFAULT_DECALDRAW_CLEAN_DECAL |= DECAL_DRAW_MESH;

    m_DecalTransferState = DECAL_TRANS_SELECT_REGION;
    m_iDecalTransferFlags = m_iDEFAULT_DECALDRAW_EVERYTHING;
}

void AppPartSelection::AdvanceFineDetailState()
{
    if (m_DecalTransferState == DECAL_TRANS_SELECT_REGION)
    {
        // estimate transforms if a region was selected
        if (GetRefSelectedPnts().size()>0)
            TransformPaintedToOther();
        
        m_DecalTransferState = DECAL_TRANS_DRAW_DECAL;
    }
    else if (m_DecalTransferState == DECAL_TRANS_DRAW_DECAL)
    {
            // IF NO TRANSFORMS: estimate transforms / transfer decal
        if (GetRefSelectedPnts().size()==0)
        {
            FillCoarsePaintingFromFinePainting();
            TransformPaintedToOther();
        }
        m_DecalTransferState = DECAL_TRANS_DECAL_TRANSFERED;
    }
    else if (m_DecalTransferState == DECAL_TRANS_DECAL_TRANSFERED)
    {
        m_DecalTransferState = DECAL_TRANS_SELECT_REGION;
    }
}

void AppPartSelection::TransformPaintedToOther()
{
    std::cout<<"transform painted to other"<<std::endl;
    m_iSemanticGlueAlignmentFlags = GLUE_SCALE_FROM_ALIGNMENT | GLUE_ROTATION_FROM_ALIGNMENT;
    
    std::map<ModelDatabase*, int >::iterator iter;
    iter = m_DBToSelectedModelID.find(m_Pipeline.GetDB());
    assert(iter!=m_DBToSelectedModelID.end());
    Model3D * model = m_Pipeline.GetDB()->GetModelByID(iter->second);
    assert(model!=NULL);
    std::map<Model3D*, std::vector<int> > pairs;
    std::map<Model3D*, std::vector<double> > weights;
    FillCorrespondenceValues(model, GetRefSelectedPnts(), &pairs, &weights);

    std::cout<<"transform painted to other - filled corrs"<<std::endl;
    
    int maxNWeights = 0;
    for (std::map<Model3D*, std::vector<double> >::iterator weightsIter = weights.begin();
         weightsIter != weights.end(); weightsIter++)
        maxNWeights = vkMax(maxNWeights, (int)weightsIter->second.size());
    assert(maxNWeights>0);
    
    if (m_iTempArraySize < maxNWeights)
    {
        if (m_iTempArraySize!=0)
        {
            delete [] m_aTempArrayPnts1;
            delete [] m_aTempArrayPnts2;
            delete [] m_aTempArrayWeights;
        }
        
        m_iTempArraySize = maxNWeights;
        m_aTempArrayPnts1 = new PointWithID[m_iTempArraySize];
        m_aTempArrayPnts2 = new PointWithID[m_iTempArraySize];
        m_aTempArrayWeights = new double[m_iTempArraySize];
    }

    for (std::map<Model3D*, std::vector<int> >::iterator pairsIter = pairs.begin();
         pairsIter != pairs.end(); pairsIter++)
    {
        std::map<Model3D*, std::vector<double> >::iterator weightsIter;
        weightsIter = weights.find(pairsIter->first);
        assert(weightsIter != weights.end());
        assert(weightsIter->second.size() == pairsIter->second.size()/2);
        assert(pairsIter->second.size()%2 == 0);
        
        std::vector<int> currPairs;
        for (int i=0; i<(int)weightsIter->second.size(); i++)
        {
            currPairs.push_back(pairsIter->second[2*i+1]);
            currPairs.push_back(pairsIter->second[2*i+0]);
            m_aTempArrayWeights[i] = weightsIter->second[i];
        }
        //FindRigidAlignment(model, pairsIter->first, currPairs, weightsArray);

        //std::cout<<"transform painted to other - preRigid: 1"<<std::endl;
        FindRigidAlignment(pairsIter->first, model, currPairs, m_aTempArrayWeights);
        //std::cout<<"transform painted to other - preRigid: 2"<<std::endl;
    }
    std::cout<<"transform painted to other - DONE!"<<std::endl;    
    
}

void AppPartSelection::FillCoarsePaintingFromFinePainting()
{
    // estimate region that should have been selected
    assert(false);
}

void AppPartSelection::WriteInteractionResults()
{
	FillPtrsFromParams();
	// app directory
	VKString appDir = m_Pipeline.GetMainDB()->GetAppDir(m_sAppName);
	
	// write created labels (names)
	std::ofstream textStreamInteractive((appDir+"InteractionResults.txt").c_str());
	assert(textStreamInteractive.is_open());
	textStreamInteractive<<"InteractionMode "<<(int)m_InteractionMode<<"\n";
	textStreamInteractive<<"CreatedLabels "<<m_CreatedLabels.count()<<"\n";
	for (int labelID=0; labelID<(int)m_CreatedLabels.count(); labelID++)
		textStreamInteractive<<m_CreatedLabels[labelID].c_str()<<" ";
	Model3D * model = m_Pipeline.GetMainDB()->GetModelByID(0);
	VKString finLabels = m_sDefaultModelValue + "_Labels";		
	ModelValue * finalLabels = model->GetModelValue(finLabels);
	if (finalLabels!=NULL)
		textStreamInteractive<<"FinalLabels"<<"\n";
	else
		textStreamInteractive<<"NoFinalLabels"<<"\n";
	
	// write created label values (including final labels)
	for (int modelID = 0; modelID < m_Pipeline.GetMainDB()->NumModels(); modelID++)
	{
		Model3D * model = m_Pipeline.GetMainDB()->GetModelByID(modelID);
		for (int labelID=0; labelID<(int)m_CreatedLabels.count(); labelID++)
		{
			VKString modelValueName = m_sDefaultModelValue + "_" + m_CreatedLabels[labelID];
			ModelValue * modelValue = model->GetModelValue(modelValueName);
			if (modelValue!=NULL)
				modelValue->WriteModelValue(m_Pipeline.GetMainDB());
		}
	
		VKString modelValueName = m_sDefaultModelValue + "_Labels";		
		ModelValue * finalLabels = model->GetModelValue(modelValueName);
		if (finalLabels!=NULL)
			finalLabels->WriteModelValue(m_Pipeline.GetMainDB());			
	}
	
	textStreamInteractive<<"OriginalPaint "<<m_mModelPointLabel.size()<<"\n";
	for (std::map<int, std::map<int, std::vector<int> > >::iterator modelIter = m_mModelPointLabel.begin(); 
		 modelIter != m_mModelPointLabel.end(); modelIter++)
	{
		textStreamInteractive<<"Model "<<modelIter->first<<" "<<modelIter->second.size()<<"\n";
		for (std::map<int, std::vector<int> >::iterator pointIter = modelIter->second.begin(); 
			 pointIter != modelIter->second.end(); pointIter++)
		{
			textStreamInteractive<<pointIter->first<<" "<<pointIter->second.size()<<"\n";
			for (int l=0; l<pointIter->second.size(); l++)
				textStreamInteractive<<pointIter->second[l]<<" ";
			textStreamInteractive<<"\n";
		}
	}	
	
	// painted stuff
	textStreamInteractive<<"PaintedModels "<<m_DBToSelectedModelID.size()<<"\n";
	for (std::map<ModelDatabase*, int >::iterator iter = m_DBToSelectedModelID.begin(); 
		 iter != m_DBToSelectedModelID.end(); iter++)
	{
		textStreamInteractive<<"ModelDescr "<<iter->first->DatabaseName().c_str()<<" "<<iter->second<<"\n";
		assert(m_DBToSelectedPntIDs.find(iter->first)!=m_DBToSelectedPntIDs.end());
		std::vector<int> & selPnts = m_DBToSelectedPntIDs[iter->first];
		textStreamInteractive<<"NSelected "<<selPnts.size()<<"\n";
		for (int i=0; i<(int)selPnts.size(); i++)
			textStreamInteractive<<selPnts[i]<<" ";
		textStreamInteractive<<"\n";
	}
	
	
	// write glue interactions
	std::ofstream textStreamGlueInteraction((appDir+"GlueInteractions.txt").c_str());
	if (m_sGlueLabelName=="")
		m_sGlueLabelName = "none";
	textStreamGlueInteraction<<"GlueState "<<m_SemanticGlueState<<"\n";
	textStreamGlueInteraction<<"Label "<<m_sGlueLabelName.c_str()<<"\n";
	textStreamGlueInteraction<<"ModelSelected "<<(m_pDBWithToBeGluedModel!=NULL)<<"\n";
	if (m_pDBWithToBeGluedModel!=NULL)
	{
		assert(m_pModelToBeGlued!=NULL);
		textStreamGlueInteraction<<"AddDB "<<m_pDBWithToBeGluedModel->DatabaseName().c_str()<<"\n";
		textStreamGlueInteraction<<"AddModelID "<<m_pModelToBeGlued->ModelID()<<"\n";
	}

	// write glue transforms
	std::map<Model3D *, std::map<Model3D *, R3Affine> >::iterator iter;
	int nTransforms = 0;
	for (iter = m_GlueDataModelPairToTransform.begin(); 
         iter != m_GlueDataModelPairToTransform.end(); iter++)
		nTransforms += iter->second.size();
	textStreamGlueInteraction<<"NTransforms "<<nTransforms<<"\n";
	
	for (iter = m_GlueDataModelPairToTransform.begin(); 
         iter != m_GlueDataModelPairToTransform.end(); iter++)
	{
		textStreamGlueInteraction<<iter->first->ModelID()<<" ";
		textStreamGlueInteraction<<iter->first->GetDB()->DatabaseName().c_str()<<" ";
		for (std::map<Model3D *, R3Affine>::iterator iter2 = iter->second.begin(); 
			 iter2 != iter->second.end(); iter2++)
		{
			textStreamGlueInteraction<<iter2->first->ModelID()<<" ";
			textStreamGlueInteraction<<iter2->first->GetDB()->DatabaseName().c_str()<<" ";
			
			const R4Matrix& mtx = iter2->second.Matrix();
			
			for (int i=0; i<4; i++)
			for (int j=0; j<4; j++)
				textStreamGlueInteraction<<mtx[i][j]<<" ";
			textStreamGlueInteraction<<"\n";
		}
	}
	const R3Point & o = m_MainPairCoords.Origin();
	const R3Triad & t = m_MainPairCoords.Axes();
	textStreamGlueInteraction<<"MainPairXform "<<m_fMainPairScale<<" ";
	textStreamGlueInteraction<<o.X()<<" "<<o.Y()<<" "<<o.Z()<<" ";
	textStreamGlueInteraction<<t[0].X()<<" "<<t[0].Y()<<" "<<t[0].Z()<<" ";
	textStreamGlueInteraction<<t[1].X()<<" "<<t[1].Y()<<" "<<t[1].Z()<<" ";
	textStreamGlueInteraction<<t[2].X()<<" "<<t[2].Y()<<" "<<t[2].Z()<<" ";
	
	// write all camera transformations
	std::ofstream textStreamAllCameraTransforms((appDir+"CameraTransformations.txt").c_str());
	textStreamAllCameraTransforms<<"NDBs "<<(int)(m_Pipeline.NumAdditionalDBs()+1)<<"\n";
	for (int dbID=-1; dbID<m_Pipeline.NumAdditionalDBs(); dbID++)
	{	
		ModelDatabase * db = NULL;
		if (dbID==-1)
			db = m_Pipeline.GetMainDB();
		else
			db = m_Pipeline.GetAdditionalDB(dbID);
		assert(db!=NULL);
		
		textStreamAllCameraTransforms<<"DB "<<db->DatabaseName().c_str()<<" "<<db->NumModels()<<"\n";
		for (int modelID = 0; modelID < db->NumModels(); modelID++)
			db->GetModelByID(modelID)->RenderingParams()->WriteCameras(textStreamAllCameraTransforms);
	}	
}

void AppPartSelection::ReadInteractionResults()
{
	FillPtrsFromParams();
	VKString appDir = m_Pipeline.GetMainDB()->GetAppDir(m_sAppName);
	VKString tempStr;
	
	// read created labels (names / values)
	std::ifstream textStreamInteractive((appDir+"InteractionResults.txt").c_str());
	if (textStreamInteractive.is_open())
	{
		tempStr.readToken(textStreamInteractive);
		assert(tempStr=="InteractionMode");
		int interactMode;
		textStreamInteractive>>interactMode;
		m_InteractionMode = (InteractionMode)interactMode;
		
		tempStr.readToken(textStreamInteractive);
		assert(tempStr=="CreatedLabels");
		int nLabels;
		textStreamInteractive>>nLabels;
		for (int i=0; i<nLabels; i++)
		{
			tempStr.readToken(textStreamInteractive);
			m_CreatedLabels.push_back(tempStr);
		}
		tempStr.readToken(textStreamInteractive);
		bool loadFinal;
		if (tempStr=="FinalLabels")
			loadFinal = true;
		else if (tempStr=="NoFinalLabels")
			loadFinal = false;
		else
			assert(false);

		
		for (int modelID = 0; modelID < m_Pipeline.GetMainDB()->NumModels(); modelID++)
		{
			Model3D * model = m_Pipeline.GetMainDB()->GetModelByID(modelID);
			for (int labelID=0; labelID<(int)m_CreatedLabels.count(); labelID++)
			{
				VKString modelValueName = m_sDefaultModelValue+"_"+m_CreatedLabels[labelID];
				if (model->GetModelValue(modelValueName)!=NULL)
					model->RemoveModelValue(modelValueName);
					
				bool couldRead = ModelValue::ReadModelValue(model->ModelName(), 
															modelValueName, m_Pipeline.GetMainDB());
				if (!couldRead)
					m_CreatedLabels.remove(modelValueName);
				else 
				{
					if (!m_Pipeline.GetDB()->m_ModelValueNames.contains(modelValueName))
						m_Pipeline.GetDB()->m_ModelValueNames.push_back(modelValueName);
				}

			}
			if (loadFinal)
			{
				VKString finLabels = m_sDefaultModelValue + "_Labels";		
				if (model->GetModelValue(finLabels)==NULL)
					model->RemoveModelValue(finLabels);
				
				bool couldRead = ModelValue::ReadModelValue(model->ModelName(), 
															finLabels, m_Pipeline.GetMainDB());
				if (couldRead)
				{
					if (!m_Pipeline.GetDB()->m_ModelValueNames.contains(finLabels))
						m_Pipeline.GetDB()->m_ModelValueNames.push_back(finLabels);
				}
			}
		}
		
		tempStr.readToken(textStreamInteractive);
		assert(tempStr=="OriginalPaint");
		int nModels;
		textStreamInteractive>>nModels;
		
		for (int m=0; m<nModels; m++)
		{
			tempStr.readToken(textStreamInteractive);
			assert(tempStr=="Model");
			int modelID;
			int nPnts;
			textStreamInteractive>>modelID>>nPnts;
			for (int p=0; p<nPnts; p++)
			{
				int pntID; 
				int nLabels;
				textStreamInteractive>>pntID>>nLabels;
				for (int l=0; l<nLabels; l++)
				{
					int labelID;
					textStreamInteractive>>labelID;
					m_mModelPointLabel[modelID][pntID].push_back(labelID);
				}
			}
		}
		
		// painted stuff
		tempStr.readToken(textStreamInteractive);		assert(tempStr=="PaintedModels");
		int nSelectedModels;
		textStreamInteractive>>nSelectedModels;
		m_DBToSelectedModelID.clear();
		m_DBToSelectedPntIDs.clear();
		
		for (std::map<Model3D*, ModelPerVertexValues *>::iterator iter = m_pModelIDToPaintedOn.begin(); 
			 iter != m_pModelIDToPaintedOn.end(); iter++)
			iter->second->SetValues("auto", 0);
		
		for (int i=0; i<nSelectedModels; i++)
		{
			tempStr.readToken(textStreamInteractive);	assert(tempStr=="ModelDescr");
			tempStr.readToken(textStreamInteractive);
			int modelID;
			textStreamInteractive>>modelID;
			
			ModelDatabase * db = m_Pipeline.GetAdditionalDBByName(tempStr);
			assert(db!=NULL);			
			m_DBToSelectedModelID[db] = modelID;
			
			tempStr.readToken(textStreamInteractive);	assert(tempStr=="NSelected");
			int nSelected;
			textStreamInteractive>>nSelected;
			
			for (int sID = 0; sID < nSelected; sID++)
			{
				int pntID;
				textStreamInteractive>>pntID;
				m_DBToSelectedPntIDs[db].push_back(pntID);
			}
			FillPaintedOnRendering(db->GetModelByID(modelID), m_DBToSelectedPntIDs[db]);
		}		
	}
	
	// read glue interactions
	std::ifstream textStreamGlueInteraction((appDir+"GlueInteractions.txt").c_str());
	if (textStreamGlueInteraction.is_open())
	{
		tempStr.readToken(textStreamGlueInteraction);		assert(tempStr=="GlueState");
		int state;		textStreamGlueInteraction>>state;
		m_SemanticGlueState = (SemanticGlueState) state;
		
		tempStr.readToken(textStreamGlueInteraction);		assert(tempStr=="Label");
		m_sGlueLabelName.readToken(textStreamGlueInteraction);
		
		tempStr.readToken(textStreamGlueInteraction);		assert(tempStr=="ModelSelected");
		bool modelSelected;
		textStreamGlueInteraction>>modelSelected;
		if (modelSelected)
		{
			tempStr.readToken(textStreamGlueInteraction);		assert(tempStr=="AddDB");
			tempStr.readToken(textStreamGlueInteraction);
			m_pDBWithToBeGluedModel = m_Pipeline.GetAdditionalDBByName(tempStr);
			tempStr.readToken(textStreamGlueInteraction);		assert(tempStr=="AddModelID");
			int modelID = 0;
			textStreamGlueInteraction>>modelID;
			m_pModelToBeGlued = m_pDBWithToBeGluedModel->GetModelByID(modelID);
		}
		
		// read transforms
		tempStr.readToken(textStreamGlueInteraction);		assert(tempStr=="NTransforms");
		int NTransforms;
		textStreamGlueInteraction>>NTransforms;
		for (int i=0; i<NTransforms; i++)
		{
			int model1ID, model2ID;
			textStreamGlueInteraction>>model1ID;
			tempStr.readToken(textStreamGlueInteraction);
			ModelDatabase * db1 = m_Pipeline.GetAdditionalDBByName(tempStr);
			textStreamGlueInteraction>>model2ID;
			tempStr.readToken(textStreamGlueInteraction);
			ModelDatabase * db2 = m_Pipeline.GetAdditionalDBByName(tempStr);
			
			Model3D * m1 = db1->GetModelByID(model1ID);
			Model3D * m2 = db2->GetModelByID(model2ID);
			
			// NOTE: m1 - m2 - is actually correct order here.
			R4Matrix mtx;
			for (int i=0; i<4; i++)
			for (int j=0; j<4; j++)
				textStreamGlueInteraction>>mtx[i][j];
			m_GlueDataModelPairToTransform[m1][m2] = R3Affine(mtx);
		}
		
		
		tempStr.readToken(textStreamGlueInteraction);		assert(tempStr=="MainPairXform");
		textStreamGlueInteraction>>m_fMainPairScale;
		double x, y, z;		textStreamGlueInteraction>>x>>y>>z;
		R3Point o(x, y, z);
		textStreamGlueInteraction>>x>>y>>z;
		R3Vector v1(x, y, z);
		textStreamGlueInteraction>>x>>y>>z;
		R3Vector v2(x, y, z);
		textStreamGlueInteraction>>x>>y>>z;
		R3Vector v3(x, y, z);
		m_MainPairCoords = R3CoordSystem(o, R3Triad(v1, v2, v3));
	}
	
	std::ifstream textStreamAllCameraTransforms((appDir+"CameraTransformations.txt").c_str());
	
	bool noVis = false;
	if (textStreamAllCameraTransforms.is_open())
	{
		tempStr.readToken(textStreamAllCameraTransforms);		assert(tempStr=="NDBs");
		int nDBs = 0;
		textStreamAllCameraTransforms>>nDBs;
		assert(nDBs==m_Pipeline.NumAdditionalDBs() + 1);
		for (int dbID=-1; dbID<m_Pipeline.NumAdditionalDBs(); dbID++)
		{	
			ModelDatabase * db = NULL;
			if (dbID==-1)
				db = m_Pipeline.GetMainDB();
			else
				db = m_Pipeline.GetAdditionalDB(dbID);
			assert(db!=NULL);
						
			tempStr.readToken(textStreamAllCameraTransforms);		assert(tempStr=="DB");
			tempStr.readToken(textStreamAllCameraTransforms);		assert(tempStr==db->DatabaseName());
			int nModels;
			textStreamAllCameraTransforms>>nModels;
			assert(nModels==db->NumModels());
						
			for (int modelID = 0; modelID < db->NumModels(); modelID++)
			{
				if (db->GetModelByID(modelID)->RenderingParams()==NULL)
				{
					noVis = true;
					break;
				}
				db->GetModelByID(modelID)->RenderingParams()->ReadCameras(textStreamAllCameraTransforms);
			}
			if (noVis)
				break;
		}
	}
	if (noVis)
		std::cout<<"[WARNING] Cameras were not loaded (vis. window does not exist)"<<std::endl;
}

void AppPartSelection::CreateAlternativeConservation(double * values)
{			
	for (int i=0; i<m_Pipeline.GetMainDB()->NGlobalPnts(); i++)
		values[i] = -1;
	
	int modelID = m_Pipeline.GetDB()->GetRenderParams()->m_mWindowToModelID[0];
	Model3D * model = m_Pipeline.GetDB()->GetModelByID(modelID); 
	assert(model!=NULL);
	PointSet * pts = model->PointCloud();

	for (int i=0; i<pts->NumPoints(); i++)
	{
		ModelPoint p = pts->GetPoint(i);
		int gID = m_Pipeline.GetMainDB()->GlobalID(p);
		
		std::vector<ModelPoint> corrs;
		m_pInputCollection->GetCorrs(p, &corrs);
		
		std::vector<double> vals;
		double mean= 0;
		double * perModelCorrs = new double[m_Pipeline.GetDB()->NumModels()];
		for (int j=0; j<m_Pipeline.GetDB()->NumModels(); j++)
			perModelCorrs[j] = 0;
		
		for (int j=0; j< (int)corrs.size(); j++)
		{
			double corrVal = m_pInputCorrValue->Value(p, corrs[j]);
			double currVal = perModelCorrs[corrs[j].GetModel()->ModelID()];
			perModelCorrs[corrs[j].GetModel()->ModelID()] = vkMax(currVal, corrVal); 
		}
		for (int j=0; j<m_Pipeline.GetDB()->NumModels(); j++)
		{
			vals.push_back(perModelCorrs[j]);
			mean += perModelCorrs[j];
		}
		mean /= (double)m_Pipeline.GetDB()->NumModels();
		
		double stdDev = 0;
		for (int j=0; j<(int)vals.size(); j++)
			stdDev += pow(vals[j] - mean, 2.);
		
		delete [] perModelCorrs;
		
		stdDev = sqrt(stdDev);
		//values[gID] = stdDev * mean;
		values[gID] = mean;
	}
}

void AppPartSelection::CreateModelValue(const VKString & modelValueName, double * values)
{
//	double visFracMax = .8;
//	double visFracMin = .1;
	ModelDatabase * db = m_Pipeline.GetDB();
	int nPoints = db->NGlobalPnts();
	
	for (int modelID=0; modelID < m_Pipeline.GetDB()->NumModels(); modelID++)
	{
		Model3D * model =db->GetModelByID(modelID);
		assert(model!=NULL);
		PointSet * modelPnts = model->PointCloud();
		ModelValue * modelValue = model->GetModelValue(modelValueName);
		double firstVal = values[db->GlobalID(modelPnts->GetPoint(0))];
		if (firstVal < 0 && modelValue!=NULL)
			continue;
		if (modelValue!=NULL)
			model->RemoveModelValue(modelValueName);
		modelValue = new ModelValue(model, modelValueName, false, false);
		model->AddModelValue(modelValueName, modelValue);
		
//		std::vector<double> modelVals;
//		for (int j=0; j<modelPnts->NumPoints(); j++)
//		{
//			double val = consData[db->GlobalID(modelPnts->GetPoint(j))];
//			modelVals.push_back(val);
//		}
//		std::sort(modelVals.begin(), modelVals.end());
//		int maxValID = (int)(visFracMax * (int)modelVals.size());
//		assert(maxValID < modelVals.size());
//		int minValID = (int)(visFracMin * (int)modelVals.size());
//		assert(maxValID < modelVals.size());
//		double modelMax = modelVals[maxValID];
//		double modelMin = modelVals[minValID];
		//std::cout<<"Vals: ["<<modelMin<<" "<<modelMax<<"]"<<std::endl;
		for (int j=0; j<modelPnts->NumPoints(); j++)	
		{
			ModelPoint pnt = modelPnts->GetPoint(j);
			double val = values[db->GlobalID(pnt)];
//			val = val > modelMax ? modelMax : val;
//			val = val < modelMin ? modelMin : val;
//			modelValue->SetValue(pnt, -log(val));
			modelValue->SetValue(pnt, val);
		}
	}
	
	if (!m_Pipeline.GetDB()->m_ModelValueNames.contains(modelValueName))
		m_Pipeline.GetDB()->m_ModelValueNames.push_back(modelValueName);
}

void AppPartSelection::AddFacet(double weight)
{
	if (GetRefSelectedPnts().size()==0)
		return;
	
	m_FacetsNPoints.push_back((int)GetRefSelectedPnts().size());
	FillCumulativeSimilarities();
	
	int nGlobalPnts = m_Pipeline.GetDB()->NGlobalPnts();
	int nModels = m_Pipeline.GetDB()->NumModels();
	
	int lastID = (int)m_FacetsPerModelValues.size();
	m_FacetsPerModelValues.push_back(std::vector<double>());
	
	for (int i=0; i<nModels; i++)
		m_FacetsPerModelValues[lastID].push_back(0);
	m_FacetsPerPointValues.push_back(new double[nGlobalPnts]);
	for (int i=0; i<nGlobalPnts; i++)
		m_FacetsPerPointValues[lastID][i] = 0;

	std::vector<Sortable> sortedModels;
	if (m_PerModelShapeDescriptorScaling.size()==nModels)
		// fuzzy corrs based descriptor
	{
		std::cout<<"Shape descriptor facet "<<lastID<<" weight="<<weight;
		std::cout<<" nPts="<<GetRefSelectedPnts().size()<<std::endl;
		m_FacetsPerModelValues[lastID] = m_PerModelShapeDescriptorScaling;
		for (int i=0; i<(int)m_FacetsPerModelValues[lastID].size(); i++)
			m_FacetsPerModelValues[lastID][i] *= weight;
			
		for (int i=0; i<nModels; i++)
			sortedModels.push_back(Sortable(0., NULL, i));
		
//		bool negVals = false;
		for (int i=0; i<(int)m_FacetsPerModelValues.size(); i++)
		{
			for (int j=0; j<(int)m_FacetsPerModelValues[i].size(); j++)
			{
//				if (m_FacetsPerModelValues[i][j] < -.0001)
//					negVals = true;
				
				//sortedModels[j].value *= m_FacetsPerModelValues[i][j];
				sortedModels[j].value += m_FacetsPerModelValues[i][j];			
			}				
		}
//		if (negVals)
//		{	
//			for (int j=0; j<(int)sortedModels.size(); j++)
//				sortedModels[j].value = -vkAbs(sortedModels[j].value);
//		}

	}
	else	// just fuzzy corrs for sorting
	{
		std::cout<<"Shape descriptor fuzzy corrs "<<lastID<<" weight="<<weight;
		std::cout<<" nPts="<<GetRefSelectedPnts().size()<<std::endl;

		for (int i=0; i<nModels; i++)
			sortedModels.push_back(Sortable(1., NULL, i));

		for (int i=0; i<nGlobalPnts; i++)
		{
			int mID = m_Pipeline.GetDB()->GetPoint(i).GetModel()->ModelID();
			m_FacetsPerModelValues[lastID][mID] += m_afCumulativeSimilarities[i];
			m_FacetsPerPointValues[lastID][i] = m_afCumulativeSimilarities[i];
			m_afCumulativeSimilarities[i] = 1.;
		}
		
		std::vector<Sortable> sortedModels;
		for (int i=0; i<nModels; i++)
			sortedModels.push_back(Sortable(1., NULL, i));
			
		for (int i=0; i<(int)m_FacetsPerModelValues.size(); i++)
		{
			//int norm = m_FacetsNPoints[i];
			double norm = 1.;
			for (int j=0; j<(int)m_FacetsPerModelValues[i].size(); j++)
				sortedModels[j].value *= (m_FacetsPerModelValues[i][j] / norm);
			
			for (int j=0; j<nGlobalPnts; j++)
				m_afCumulativeSimilarities[j] *= m_FacetsPerPointValues[i][j];
		}
		m_fGlobalMax = 0;
		for (int j=0; j<nGlobalPnts; j++)
			m_fGlobalMax = vkMax(m_fGlobalMax, m_afCumulativeSimilarities[j]);
	}
	std::sort(sortedModels.begin(), sortedModels.end());
	
	m_aiSimilarModels.clear();
	for (int i=(int)sortedModels.size()-1; i>=0; i--)
		m_aiSimilarModels.push_back(sortedModels[i].id);

	for (int i=0; i<(int)sortedModels.size(); i++)
	{
		Model3D * m = m_Pipeline.GetDB()->GetModelByID(sortedModels[i].id);
		std::cout<<m->ModelName().c_str()<<" : "<<sortedModels[i].value<<std::endl;
	}
	
	m_bNeedsUpdate = true;
}

void AppPartSelection::ClearFacets(bool clearAll)
{
	if (!clearAll)
	{
		int nF = (int)m_FacetsPerModelValues.size() - 1;
		if (nF <0)
		{
			std::cout<<"No facets"<<std::endl;
			return;
		}

		m_FacetsPerModelValues.erase(m_FacetsPerModelValues.begin() + nF);
		m_FacetsNPoints.erase(m_FacetsNPoints.begin() + nF);
		
		delete [] m_FacetsPerPointValues[nF];
		m_FacetsPerPointValues.erase(m_FacetsPerPointValues.begin() + nF);
		std::cout<<"Cleared last facet: "<<m_FacetsPerModelValues.size()<<std::endl;
	}
	else 
	{
		m_FacetsPerModelValues.clear();
		m_FacetsNPoints.clear();
		for (int i=0; i<(int)m_FacetsPerPointValues.size(); i++)
			delete [] m_FacetsPerPointValues[i];
		m_FacetsPerPointValues.clear();
		std::cout<<"Cleared All Facets"<<std::endl;
	}
}

void AppPartSelection::ShapeDescriptorSimilarityTerm(std::vector<int> & pnts, 
													 int modelFrom, 
													 std::vector<double> & similarities)
{
	bool expensive = true;
	bool toRedOnly = false;
	bool fromRedError = false;
	// store all pairwise similarities	
	m_PerModelShapeDescriptorScaling.clear();
	
	assert(m_afCumulativeSimilarities!=NULL);
	
	ModelDatabase * db = m_Pipeline.GetDB();
	Model3D * m = db->GetModelByID(modelFrom);
	assert(m!=NULL);
	
	R3MeshSearchTree * searchTree = m->GetSearchTree();
	assert(searchTree!=NULL);
	R3Kdtree<const PointInSet *> * kdTree = m->PointCloud()->GetKdTree();
	assert(kdTree!=NULL);
	
	DistanceEuclidean * eucl = m->DefaultEuclideanDistance();
	assert(eucl!=NULL);
	double distNorm = eucl->GetStandardNorm();
	double sigma2 = pow(distNorm / 10., 2);
	
	if (m_GlueDataModelPairToTransform.size()==0)
	{
		std::cout<<"[WARNING] No transformation prescribed - ";
		std::cout<<"not using shape descriptor"<<std::endl;
		return;
	}
	// cumulative values must have been filled
	for (int i=0; i<db->NumModels(); i++)
	{
		Model3D * mto = db->GetModelByID(i);
		assert(mto!=NULL);
		
		R3Affine xForm(R4identity_matrix);
		
		if (mto!=m)
		{
			std::map<Model3D *, std::map<Model3D *, R3Affine> >::iterator iter1;
			iter1 = m_GlueDataModelPairToTransform.find(mto);
			assert(iter1!=m_GlueDataModelPairToTransform.end());
			std::map<Model3D *, R3Affine>::iterator iter2 = iter1->second.find(m);
			assert(iter2!=iter1->second.end());
			xForm = iter2->second;
		}
		
		// for each non-zero 'cumulative similarities'
		PointSet * ptsTo = mto->PointCloud();
		double normCumVal = 0;
		double totalScore = 0;
		double totalScoreFrom = 0;
		double rmse = 0;
		double rmseFrom = 0;
		assert(!fromRedError);
//		if (fromRedError)
//		{
//			R3Affine xFormInv(xForm.InverseMatrix());
//			R3MeshSearchTree * searchTree2 = mto->GetSearchTree();			
//			
//			double normCumVal = 0;
//			for (int pFromID=0; pFromID<(int)pnts.size(); pFromID++)
//			{
//				R3MeshIntersection closest;
//				R3Point pFrom = m->PointCloud()->GetPoint(pnts[pFromID]).GetPosition();
//				xFormInv.Apply(pFrom);
//				searchTree2->FindClosest(pFrom, closest);
//				assert (closest.type != R3_MESH_NULL_TYPE);
//				R3Point transformedTo = closest.point;
//				
//				double cumVal = 1.;
//				double minDist = FLT_MAX;
//				
//				for (int j=0; j<ptsTo->NumPoints(); j++)
//				{			
//					const ModelPoint & p = ptsTo->GetPoint(j);
//					double tempDist = R3Distance(p.GetPosition(), transformedTo);
//					if (tempDist < minDist)
//					{
//						minDist = tempDist;
//						int gID = db->GlobalID(p);
//						cumVal = m_afCumulativeSimilarities[gID];
//					}
//				}
//				xForm.Apply(transformedTo);
//				double currDist = R3Distance(pFrom, transformedTo);
//				rmseFrom += currDist * cumVal;
//				normCumVal += cumVal;
//				
//				totalScoreFrom += cumVal * (exp(-currDist*currDist / sigma2));
//			}
//			rmseFrom/=normCumVal;
//			//totalScoreFrom /= normCumVal;
//		}
		for (int j=0; j<ptsTo->NumPoints(); j++)
		{
			const ModelPoint & p = ptsTo->GetPoint(j);
			int gID = db->GlobalID(p);
			double cumVal = m_afCumulativeSimilarities[gID];
			if (cumVal<=0)
				continue;
			
			R3Point transformedTo = p.GetPosition();
			xForm.Apply(transformedTo);
			
			double dist = -1;
			assert(!toRedOnly);
//			if (toRedOnly)
//			{
//				for (int pFromID=0; pFromID<(int)pnts.size(); pFromID++)
//				{
//					R3Point nearestPFrom = m->PointCloud()->GetPoint(pnts[pFromID]).GetPosition();
//					double currDist = R3Distance(nearestPFrom, transformedTo);
//					if (dist < 0 || currDist < dist)
//						dist = currDist;
//				}
//			}
//			else
			{
				R3Point nearestPfrom;
				assert(expensive);
				if (expensive)
				{
					R3MeshIntersection closest;
					searchTree->FindClosest(transformedTo, closest);
					assert (closest.type != R3_MESH_NULL_TYPE);
					nearestPfrom = closest.point;
				}
//				else
//				{
//					const PointInSet * kdTreePnt = kdTree->FindClosest(transformedTo);
//					nearestPfrom = kdTreePnt->GetPosition();
//				}
				dist = R3Distance(nearestPfrom, transformedTo);
			}
						
			normCumVal += cumVal;
			//rmse += cumVal * pow(dist/distNorm, 2.);
			rmse += cumVal * dist / distNorm;
			
			//totalScore += cumVal * (exp(-dist*dist / sigma2) - .1);
			totalScore += cumVal * (exp(-dist*dist / sigma2));
			//totalScore += exp(-dist*dist / sigma2);
			//std::cout<<"\tScore="<<cumVal<<" * "<<(exp(-dist*dist / sigma2) - 0.1)<<std::endl;
		}
		
		
		// for every point 
		//rmse = sqrt(rmse / normCumVal);
		rmse = rmse / normCumVal;
		
		// check how far is it from m
		//m_PerModelShapeDescriptorScaling.push_back(1. / (.01 + rmse));
		assert(!fromRedError);
//		if (fromRedError)
//			m_PerModelShapeDescriptorScaling.push_back(-rmse - rmseFrom);
//		else
			m_PerModelShapeDescriptorScaling.push_back(-rmse);
//		if (fromRedError)
//			m_PerModelShapeDescriptorScaling.push_back(totalScoreFrom);
//		else
//			m_PerModelShapeDescriptorScaling.push_back(totalScore);
		
	}
}

void AppPartSelection::GlueOtherToCurrent()
{
//	Model3D * m = m_Pipeline.GetMainDB()->GetModelInSubwindow(0);
//	Model3D * m2 = m_Pipeline.GetMainDB()->GetModelInSubwindow(1);
//	m_pModelToBeGlued = m2;
//	R3Affine affid(R4identity_matrix);
//	m_GlueDataModelPairToTransform[m_pModelToBeGlued][m] = affid;
//	SemanticGlueAutoAttachToAll();
//	FillCumulativeSimilarities();
	
	// for remaing models - align
	FillPtrsFromParams();
	
	if (m_aTempArrayPnts1==NULL)
	{
		assert(m_aTempArrayPnts2==NULL);
		m_iTempArraySize = m_Pipeline.GetDB()->NGlobalPnts();
		m_aTempArrayPnts1 = new PointWithID[m_iTempArraySize];
		m_aTempArrayPnts2 = new PointWithID[m_iTempArraySize];
		m_aTempArrayWeights = new double[m_iTempArraySize];
	}

	std::vector<int> & selPts = GetRefSelectedPnts();
	Model3D * m1 = m_Pipeline.GetMainDB()->GetModelInSubwindow(0);
	PointSet * pts1 = m1->PointCloud();
	
	std::map<const Model3D *, std::vector<int> > newCorrPairs;
	std::map<const Model3D *, std::vector<double> > newCorrPairsWeights;
	
	for (int p1ID=0; p1ID<pts1->NumPoints(); p1ID++)
//	for (int i=0; i<(int)selPts.size(); i++)
	{
		bool included = false;
		
		for (int i=0; i<(int)selPts.size() && !included; i++)
			if (selPts[i]==p1ID)
				included = true;
		double dampingFactor = 1.;
		if (!included)
			dampingFactor = 1. / (4*pts1->NumPoints());
			 
		//int p1ID = selPts[i];
		ModelPoint p = pts1->GetPoint(p1ID);
		std::vector<ModelPoint> corrs;
		m_pInputCollection->GetCorrs(p, &corrs);
		
//		std::cout<<"p["<<p1ID<<"] -> "<<corrs.size()<<std::endl;
		for (int j=0; j<(int)corrs.size(); j++)
		{
			double val = m_pInputCorrValue->Value(p, corrs[j]);
			Model3D * m2 = (Model3D*)corrs[j].GetModel();
			if (dampingFactor!=0)
			{
				newCorrPairs[m2].push_back(p1ID);
				newCorrPairs[m2].push_back(corrs[j].PntCloudID());
				newCorrPairsWeights[m2].push_back(val * dampingFactor);
			}
//			if (m2==m_Pipeline.GetMainDB()->GetModelInSubwindow(1))
//				std::cout<<"\t"<<p1ID<<" -> "<<corrs[j].PntCloudID()<<" "<<val<<std::endl;
		}
	}
	
	//int m2ID = m_Pipeline.GetMainDB()->GetModelInSubwindow(1)->ModelID();
	for (int m2ID = 0; m2ID < m_Pipeline.GetMainDB()->NumModels(); m2ID++)
	{
		Model3D * m2 = m_Pipeline.GetMainDB()->GetModelByID(m2ID);
		if (m2!=m1)
		{
			std::vector<int> & currModelPairs = newCorrPairs[m2];
//			std::cout<<"pairs: "<<currModelPairs.size()<<std::endl;
//			std::cout<<"p0="<<selPts.size()<<std::endl;
			std::vector<double> & currModelWeights = newCorrPairsWeights[m2];
			assert(currModelPairs.size()/2 == currModelWeights.size());
			double * weightsPtr = m_aTempArrayWeights;
			for (int i=0; i<(int)currModelWeights.size(); i++)
				weightsPtr[i] = currModelWeights[i];
			m_fMainPairScale = 1.;
			m_fM1ToM2DiameterScale = 1.;
			FindRigidAlignment(m1, m2, currModelPairs, weightsPtr);
			//FindRigidAlignment(m2, m1, currModelPairs, weightsPtr);
		}
	}
}




