/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/


#ifndef __CollectionAggregatePairwise_H
#define __CollectionAggregatePairwise_H

//#define INCLUDE_LS_FUZZY

#ifdef INCLUDE_LS_FUZZY
	#include "LSFuzzy.h"
#endif

#include "CollectionDiffuse.h"

class CollectionAggregatePairwise : public CollectionDiffuse
{
	public:
		enum AggregationMode
		{
			AGG_UNDEFINED,
			AGG_DIFFUSION_MAP,
			AGG_LINEAR_SYSTEM
		};
	
		CollectionAggregatePairwise(const VKString & collectionName, ModelDatabase * db);
		~CollectionAggregatePairwise();
		
		virtual void InitDiffMapAggregation(const CorrMtxParams & corrMtxParams,
											const DiffusionParams & diffusionParams,
											int nEigs,
											int * globalIDToClusterID,
											int * globalIDToInClusterID,
											int ** inClusterToGlobalID,
											int * inClusterToGlobalIDSize,
											int numClusters,
											int clusterID);
	
		virtual void InitLinSysMapAggregation(const CorrMtxParams & corrMtxParams,
											  int * globalIDToClusterID,
											  int * globalIDToInClusterID,
											  int ** inClusterToGlobalID,
											  int * inClusterToGlobalIDSize,
											  int numClusters,
											  int clusterID);
		virtual void ClearPerPointCounter();
		virtual void AddEntry(int p1, int p2, double val);
		int * m_aPerPointCounter;
		
		virtual double DiffusedCorrValue(const ModelPoint & pntFrom, 
									   const ModelPoint & pntTo);
		virtual double GetAggValue(int p1, int p2);
	
		virtual void ClearPairwiseEntries();
		
		virtual void InteractiveChangeDiffusionTime(double dt);
		virtual void InteractiveChangeDiffusionSigma(double ds);
	
		virtual void Embed(const VKString & dirPath, 
						   double * timeInSec);
	
		virtual double IntegratedFuzzy (int model1, int model2,
										double ** savedVals);
	
		virtual void FillValsForPair(int modelFrom, int modelTo,
									 std::map<int, double> * saveVals,
									 int nRowsInSaveVals);	
		virtual void GetCorrs(const ModelPoint & pnt, 
							  std::vector<ModelPoint> * corrPnts,
							  std::vector<double> * vals);
		virtual void GetCorrs(const ModelPoint & pnt, 
							  std::vector<ModelPoint> * corrPnts);
		virtual ModelPoint GetBestCorr(const ModelPoint & pnt, 
								   Model3D * toModel, 
								   const VKString & modelValue, 
								   double * value);
	
		virtual VKString GetCollectionType();
	
		virtual void Write(const VKString & directory, const VKString & prefix);
		virtual void Read(const VKString & directory, const VKString & prefix);
		
		virtual AggregationMode GetAggMode() const;

		virtual int InClusterID(const ModelPoint & pnt);
		virtual int InClusterID(int globalID);
		virtual ModelPoint GetPoint(int inClusterID);
		
	
	// draw stuff
		virtual void DrawAllEmbeddedPoints(DatabaseVisWindow * window,
										   int basedOnSingleModel);
		virtual void SelectEmbeddedPoint(DatabaseVisWindow * window,
										 double x, double y,
										 int basedOnSingleModel);
		virtual void GetEmbeddedBounds(double &minX, double &minY, 
									   double &maxX, double &maxY, 
									   int basedOnSingleModel);

		virtual void FillModelGrouping(std::vector<int> & grouping);
	
		virtual DiffusionMap * GetDiffusionMap();
	
	protected:
		DiffusionMap * m_pDiffMap;	
#ifdef INCLUDE_LS_FUZZY
		LSFuzzy * m_pLSFuzzyCorrs;
#endif	
		AggregationMode m_AggMode;

			// group id for each model (in diffusion map)
		int * m_aModelToInClusterGroupID;

			// global clustering info:
		int m_iClusterID;		// cluster for this aggregation object
		int m_iNumClusters;		// total number of clusters
	
		int * m_GlobalIDToClusterID;		// map every point (in the DB) to cluster ID
		int * m_GlobalIDToInClusterID;		// map every point (in the DB) to ID within cluster
		int ** m_InClusterToGlobalID;		// for each cluster map point to global ID
		int * m_InClusterToGlobalIDSize;	// 

};


#endif




