/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/


#ifndef __CollectionDiffuseHierarchical_H
#define __CollectionDiffuseHierarchical_H

#include "CollectionDiffuse.h"
#include "DiffusionMap.h"

/**
 * A single hierarchy level
 */
class LevelOfDiffusionHierarchy
{
	public:
		struct DiffuseMapID
		{
			int m_iMatrixID;
			int m_iEntryID;
		};
	
		enum GroupingType
		{
			GROUP_MODELS,
			GROUP_PNTS
		};
	
		LevelOfDiffusionHierarchy(ModelDatabase * db, 
								  const CollectionDiffuse::DiffusionParams & diffusionParams, 
								  GroupingType gt, int levelID);
		
		void SetElementsToGroupMapping(std::vector<int> & elementToGroupMapping);
		void SetElementsToGroupMapping(GroupingType elementsType);
		void SetClustering(std::vector<int> & groupClustering);
		void InitDiffusionMaps();
		void AddCorrValue(int groupID1, int groupID2, double value);
		void EmbedAll();
		void Cluster(int nElementsInCluster, std::vector<int> & mGroupIDToClusterID);
		void ExpandGroupIntoElements(std::vector<int> & mGroupIDToClusterID, 
									 std::vector<int> & mElementsIDToClusterID);
		double EmbeddedSimilarity(int groupID1, int groupID2);
			
		GroupingType m_GroupingType;
		
		std::vector<DiffusionMap *> m_apDiffusionMaps;
		std::vector<DiffuseMapID> m_GroupClustering;
		std::vector<std::vector<int> > m_ClusterAndMatrixPositionToGroupID;
		std::vector<int> m_mElementToGroupID;
		std::vector<std::vector<int> > m_mGroupIDToElements;
		ModelDatabase * m_pDB;
		int m_iLevelID;
	
		int m_iNPoints;
		int m_iNModels;
		CollectionDiffuse::DiffusionParams m_DiffusionParams;
	
		virtual bool ReadCollectionChild(const VKString & collectionDir);
		virtual void WriteCollectionChild(const VKString & collectionDir);

};

/**
 * All hierarchies for diffusion.
 * Currently only most similar models are compared.
 */
class CollectionDiffuseHierarchical : public CollectionDiffuse
{
	public:
		CollectionDiffuseHierarchical(const VKString & collectionName, 
									  ModelDatabase * db);
		
		CollectionDiffuseHierarchical(const VKString & collectionName, 
									  ModelDatabase * db, 
									  const CorrMtxParams & corrMtxFill, 
									  const DiffusionParams & diffusionParams);
		void FillValues();
		virtual void AddEntry(int p1, int p2, double val);
		virtual void GetCorrs(const ModelPoint & pnt,	
							  std::vector<ModelPoint> * corrPnts);		
		virtual double DiffusedCorrValue(const ModelPoint & pntFrom, 
										 const ModelPoint & pntTo);
		virtual void SetRegionNormalization(std::vector<ModelPoint> & pnts);
	
		virtual void CreateHierarchies();			// prior knowledge (e.g. points group into models)
		virtual void FillHierarchies();				// fill values (integrate from points to models)
		virtual void FillClustering(int levelID);	// create clusters / diffusion hierarchies
	
		virtual VKString GetCollectionType();
		
		virtual bool ReadCollectionChild(std::ifstream & textStream);
		virtual void WriteCollectionChild(std::ofstream & textStream);
		
	protected:
		int m_iNPoints;
		int m_iNModels;
		std::vector<LevelOfDiffusionHierarchy*> m_aHierarchies;
		struct CorrValueID
		{
			int id;
			double val;
		};
	
		std::vector<CorrValueID> ** m_PointToPointGroupedByModel;
};


#endif


