/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "ModelDatabase.h"
#include "CorrCollection.h"
#include "CollectionPairwise.h"
//#include "CollectionDiffuseFullMtx.h"
#include "SinglePairSmooth.h"
#include "SinglePairPnt2Pnt.h"
#include "SinglePairRigidAlign.h"
#include "DatabaseVisWindow.h"
#include "CollectionGroundTruth.h"
#include "CollectionDiffuseInterpolator.h"
#include "CollectionDiffuseDynamicGraph.h"
#include "CorrValue.h"

CorrCollection::CorrCollection(const VKString & collectionName,
							   ModelDatabase * database, int corrID)
{
	m_sCollectionName = collectionName;
	m_pDB = database;
	m_iCorrID = corrID;
	m_bAutoRegionNormalize = true;
}

CorrCollection::~CorrCollection()
{
}

void CorrCollection::RemoveModel(int )
{
	assert(false);
}

CollectionPairwise * CorrCollection::SafeCastToPairwise()
{
	return NULL;
}


void CorrCollection::GetRegionCorrVals(std::vector<ModelPoint> & queryRegion,
									   const VKString & corrValue,
									   std::vector<ModelPoint> * corrPnts, 
									   std::vector<double> * corrVals)
{
	assert(false);	
	// use part selection alg: 
}

bool CorrCollection::ReadAndCreateCollection(const VKString & collectionName,
											 ModelDatabase * db, 
											 AnalysisPipeline * pipeline)
{

	VKString collectionFile = db->GetCollectionFile(collectionName);
	std::ifstream textStream(collectionFile.c_str());
	
	if (!textStream.is_open())
		return false;
	
	VKString text;
	text.readToken(textStream);
	assert(text=="CollectionFormat");
	
	CorrCollection * coll = NULL;
	text.readToken(textStream);
	if (text=="CollectionPairwise")
		coll = new CollectionPairwise(collectionName, db, false);
	else if (text=="CollectionDiffuseFullMtx")
		assert(false);
//		coll = new CollectionDiffuseFullMtx(collectionName, db);
	else if (text=="CollectionGroundTruth")
		coll = new CollectionGroundTruth(collectionName, db);
	else if (text=="CollectionDiffuseInterpolator")
		coll = new CollectionDiffuseInterpolator(collectionName, db);
	else if (text=="CollectionDiffuseDynamicGraph")
		coll = new CollectionDiffuseDynamicGraph(collectionName, pipeline, 
												 "DynGraph:ReadModuleName");
	assert(coll!=NULL);
	db->AddCorrCollection(collectionName, coll);
	
	bool couldRead = coll->ReadCollectionChild(textStream);
	assert(couldRead);
	
	text.readToken(textStream);
	assert(text=="CollReadAlign");
	textStream.close();
	
	return couldRead;	
}

bool CorrCollection::ReadAndCreateCollectionSinglePairWithID(const VKString & collectionName,
															 int corrID, ModelDatabase * db)
{
	VKString collectionFile = db->GetCollectionFileWithID(collectionName, corrID);
	std::ifstream textStream(collectionFile.c_str());
		
	if (!textStream.is_open())
		return false;
	
	bool couldRead = ReadAndCreateCollectionSinglePairWithID(collectionName, 
															 corrID, db, textStream);
	textStream.close();
	return couldRead;
}

bool CorrCollection::ReadAndCreateCollectionSinglePairWithID(const VKString & collectionName,  
															 int corrID, ModelDatabase * db,
															 std::ifstream & textStream)
{
	CorrCollection * coll = db->GetCorrCollection(collectionName);
	assert(coll!=NULL);
	CollectionPairwise * collectPairwise = coll->SafeCastToPairwise();
	assert(collectPairwise!=NULL);
	
	VKString text;
	text.readToken(textStream);
	assert(text=="CollectionFormat");
	
	text.readToken(textStream);
	
	CollectionSinglePair * readingCollection = NULL;
	if (text=="SinglePairSmooth")
		readingCollection = new SinglePairSmooth(collectionName, db, corrID);
	else if (text=="SinglePairPnt2Pnt")
		readingCollection = new SinglePairPnt2Pnt(collectionName, db, corrID);
	else if (text=="SinglePairRigidAlign")
		readingCollection = new SinglePairRigidAlign(collectionName, db, corrID);
	
	assert(readingCollection!=NULL);
	bool couldRead = readingCollection->ReadCollectionChild(textStream);
	collectPairwise->AddModelCorr(readingCollection);
	assert(couldRead);
	
	text.readToken(textStream);
	assert(text=="CollReadAlign");
	return couldRead;
}


bool CorrCollection::ReadCollectionChild(std::ifstream & )
{
	// nothing - implement in child
	return true;
}

void CorrCollection::WriteCollection()
{
	assert(m_sCollectionName!="");
	VKString corrsFile = m_pDB->GetCollectionFileWithID(m_sCollectionName, m_iCorrID);
	m_pDB->CreateDirsOnPathToFile(corrsFile);
	std::ofstream textStream(corrsFile.c_str());
	assert(textStream.is_open());
	WriteCollection(textStream);
	textStream.close();
}

void CorrCollection::WriteCollection(std::ofstream & textStream)
{
	textStream<<"CollectionFormat "<<GetCollectionType().c_str()<<"\n";
	WriteCollectionChild(textStream);
	textStream<<"CollReadAlign\n";
}

void CorrCollection::WriteCollectionChild(std::ofstream & )
{
	// nothing - implement in child
}

void CorrCollection::Draw(DatabaseVisWindow * window)
{
	if (window->IsCorrFlagSet(DatabaseVisWindow::RENDER_CORRS_SET))
	{
		if (window->IsCorrFlagSet(DatabaseVisWindow::RENDER_CORRS_EDGES))
			DrawCollectionEdges(window);

		if (window->IsCorrFlagSet(DatabaseVisWindow::RENDER_CORRS_PNTS))
		{
			static bool errMsg = true;
			if (errMsg)
			{
				std::cout<<"[WARNING] Rendering CORR PNTS not implemented"<<std::endl;
				errMsg = false;
			}
		}

		if (window->IsCorrFlagSet(DatabaseVisWindow::RENDER_CORRS_MVALUE))
		{
			static bool errMsg = true;
			if (errMsg)
			{
				std::cout<<"[WARNING] Rendering CORR MVALUE not implemented"<<std::endl;
				errMsg = false;
			}
		}
		
		if (window->IsCorrFlagSet(DatabaseVisWindow::RENDER_CORRS_CVALUE))
			DrawCollectionCorrValue(window);
		
		if (window->IsCorrFlagSet(DatabaseVisWindow::RENDER_CORRS_SPECIAL_1))
			DrawSpecial(window);
	}
}

void CorrCollection::DrawSpecial(DatabaseVisWindow * window)
{

}

void CorrCollection::DrawCorrespondenceEdge(const ModelPoint & pnt)
{
	//std::cout<<"Drawing corr edge"<<std::endl;
	CorrValue * corrValue = NULL;
	bool onlyBest = true;
	if (onlyBest)
	{	
		static GLfloat material[4];
		//Model3DRenderingParams * modelRender = ((Model3D*)pnt.GetModel())->RenderingParams();
		R3Point pFrom = pnt.GetPosition();
		for (int j=0; j < (int)m_pDB->GetRenderParams()->m_mWindowToModelID.size(); j++)
		{
			int mID = m_pDB->GetRenderParams()->m_mWindowToModelID[j];
			Model3D * toMdl = m_pDB->GetModelByID(mID);
			
			if (toMdl==pnt.GetModel())
				continue;
			
			ModelPoint bestPnt = GetBestCorr(pnt, toMdl, "auto", NULL);
			if (bestPnt.Invalid())
				continue;
			R3Point pToOrigCoord = bestPnt.GetPosition();
			
			Model3DRenderingParams * rend = toMdl->RenderingParams();
			assert(rend!=NULL);
			
			R3Point pTo;
			if (j==0)
				pTo = pToOrigCoord;
			else
				pTo = rend->TransformMyPntToCurrCamera(pToOrigCoord, j);
			
			glDisable(GL_LIGHTING);
			glEnable(GL_DEPTH_TEST);
			glLineWidth(2.);
			glBegin(GL_LINES);
			glColor3d(0, 1., 0);
			glVertex3d(pFrom.X(), pFrom.Y(), pFrom.Z());
			glVertex3d(pTo.X(), pTo.Y(), pTo.Z());
			glEnd();	

			rend->m_RenderAdditionalPnts.push_back(pToOrigCoord);
//			glEnable(GL_LIGHTING);
//			double r = rend->GetStandardRadius()*1.4;
//			material[0] = .2;
//			material[1] = .2;
//			material[2] = .8;
//			material[3] = 1.;
//			glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);		
//			R3Sphere(pTo, r).Draw();
		}
		
		//std::cout<<"Drawing corr edge - DONE!"<<std::endl;
		return;
	}
	
	if (onlyBest)
	{
		if (m_CorrValues.size()<1)
		{
			std::cout<<"[WARNING] Cannot render best edge. No corr value in ";
			std::cout<<m_sCollectionName.c_str()<<std::endl;
			return;
		}
		else
			corrValue = m_CorrValues[0];
	}
	

	
	std::vector<ModelPoint> pnts;
	GetCorrs(pnt, &pnts);
	R3Point pFrom = pnt.GetPosition();
	Model3D * fromMdl = (Model3D*)pnt.GetModel();

//	Model3DRenderingParams * fromRend = fromMdl->RenderingParams();
//	static bool err = true;
//	if (m_pDB->GetRenderParams()->m_mWindowToModelID[0]!=fromMdl->ModelID() && err)
//	{
//		err = false;
//		std::cout<<"[WARNING] Can only draw correspondence edges for 0th model"<<std::endl;
//		return;
//	}
//	
//	fromRend->LoadMyCamera(0);
	
	if (onlyBest)
	{
		std::map<int, int> modelIDToWndID;
		std::vector<ModelPoint> modelToBestPnt;
		std::vector<double> modelToBestValue;
		for (int j=0; j < (int)m_pDB->GetRenderParams()->m_mWindowToModelID.size(); j++)
		{
			modelToBestPnt.push_back(ModelPoint());
			modelToBestValue.push_back(-FLT_MAX);
			modelIDToWndID[m_pDB->GetRenderParams()->m_mWindowToModelID[j]] = j;
		}
			
		for (int i=0; i<(int)pnts.size(); i++)
		{
			double value = corrValue->Value(pnt, pnts[i]);
			Model3D * toMdl = (Model3D*)pnts[i].GetModel();
			
			int toMdlID = toMdl->ModelID();
			std::map<int, int>::iterator iter = modelIDToWndID.find(toMdlID);
			if (iter!=modelIDToWndID.end())
			{
				if (modelToBestValue[iter->second] < value)
				{
					modelToBestValue[iter->second] = value;
					modelToBestPnt[iter->second] = pnts[i];
				}
			}
		}
		
		for (int j=0; j < (int)m_pDB->GetRenderParams()->m_mWindowToModelID.size(); j++)
		{
			if (modelToBestPnt[j].Invalid())
				continue;
			R3Point pToOrigCoord = modelToBestPnt[j].GetPosition();
			Model3D * toMdl = (Model3D*)modelToBestPnt[j].GetModel();
			Model3DRenderingParams * rend = toMdl->RenderingParams();
			R3Point pTo;
			if (j==0)
				pTo = pToOrigCoord;
			else
				pTo = rend->TransformMyPntToCurrCamera(pToOrigCoord, j);
			
			glVertex3d(pFrom.X(), pFrom.Y(), pFrom.Z());
			glVertex3d(pTo.X(), pTo.Y(), pTo.Z());
		}
	}
	else
	{
		for (int i=0; i<(int)pnts.size(); i++)
		{
			Model3D * toMdl = (Model3D*)pnts[i].GetModel();

			int toMdlID = toMdl->ModelID();
			R3Point pToOrigCoord = pnts[i].GetPosition();
			
			for (int j=1; j < (int)m_pDB->GetRenderParams()->m_mWindowToModelID.size(); j++)
			{
				if (toMdlID == m_pDB->GetRenderParams()->m_mWindowToModelID[j])
				{
					Model3DRenderingParams * rend = toMdl->RenderingParams();
					R3Point pTo = rend->TransformMyPntToCurrCamera(pToOrigCoord, j);
					glVertex3d(pFrom.X(), pFrom.Y(), pFrom.Z());
					glVertex3d(pTo.X(), pTo.Y(), pTo.Z());
				}
			}
		}
	}
	glEnd();	
//	fromRend->RestoreCameraAfterLoad(0);	
}

void CorrCollection::DrawCollectionEdges(DatabaseVisWindow * window)
{
	int maxEdges = 50;
	int mainModelID = m_pDB->GetRenderParams()->m_mWindowToModelID[0];
	Model3D * mainModel = m_pDB->GetModelByID(mainModelID);
	assert(mainModel!=NULL);

	mainModel->RenderingParams()->Rotate(0, 0, 0, 0, 0);
	mainModel->RenderingParams()->LoadMyCamera(0);
	
	ModelPoint pnt = window->SelectedPoint();
	if (pnt.Invalid())
	{
		VKString pntSetName = mainModel->RenderingParams()->GetPointSetToRender();
		PointSet * pnts = mainModel->GetPointSet(pntSetName);
		if (pnts!=NULL)
		{
			for (int i=0; i<vkMin(pnts->NumPoints(), maxEdges); i++)
				DrawCorrespondenceEdge(pnts->GetPoint(i));
		}
	}
	else
	{
		DrawCorrespondenceEdge(pnt);
	}
	mainModel->RenderingParams()->RestoreCameraAfterLoad(0);
}

/**
 * This function draws correspondence values between a selected point 
 * and other points in a database
 */
void CorrCollection::DrawCollectionCorrValue(DatabaseVisWindow * window)
{
	// if pending update
	if (!window->PendingCorrValueRenderingUpdate())
		return;
	
	// get corr value
	CorrValue * corrValue = GetCorrValueByName(window->CorrValueToRender());
	if (corrValue==NULL)
	{
//		CollectionPairwise * collPairwise = SafeCastToPairwise();
//		if (collPairwise!=NULL)
//		{
//			int fromModelID = m_pDB->GetRenderParams()->m_mWindowToModelID[0];
//			int nModels = collPairwise->NumCorrsFromModel(fromModelID);
//			std::cout<<"Rendering models: "<<nModels<<" fromModel = "<<fromModelID<<std::endl;
//			for (int i=0; i<nModels; i++)
//			{
//				int corrID = collPairwise->GetFromModelCorrGlobalID(fromModelID, i);
//				CollectionSinglePair * coll = collPairwise->GetModelCorr(corrID);
//				for (int j=1; j < (int)m_pDB->GetRenderParams()->m_mWindowToModelID.size(); j++)
//				{
//					if (coll->ModelTo()->ModelID()==m_pDB->GetRenderParams()->m_mWindowToModelID[j])
//					{
//						coll->DrawCollectionCorrValue(window);
//						break;
//					}
//				}
//			}
//		}
//		else
//		{
			if (m_CorrValues.size() > 0)
				corrValue = m_CorrValues[0];
//		}
	}

	if (corrValue!=NULL)
	{
		corrValue->Draw(window, this);
	}
	else
		std::cout<<"[WARNING] Corr value not selected: "<<window->CorrValueToRender().c_str()<<std::endl;
}

CorrValue * CorrCollection::GetCorrValueByName(const VKString & corrValueName)
{
	std::map<VKString, int>::iterator iter = m_NameToCorrValue.find(corrValueName);
	if (iter!=m_NameToCorrValue.end())
	{
		if (iter->second >= (int)m_CorrValues.size())
		{
			std::cout<<"[ERROR] Trying to get: "<<corrValueName.c_str()<<" in ";
			std::cout<<m_sCollectionName.c_str()<<std::endl;
			std::cout<<"ID="<<iter->second<<" existing values: ";
			std::cout<<m_CorrValues.size()<<std::endl;
		}
		assert(iter->second < (int)m_CorrValues.size() );
		return m_CorrValues[iter->second];
	}
	return NULL;
}

void CorrCollection::AddCorrValue(const VKString & corrValueName, CorrValue * corrValue)
{	
	static bool alreadyLoadedErr = true;
	if (m_NameToCorrValue.find(corrValueName)!=m_NameToCorrValue.end())
	{
		if (alreadyLoadedErr)
		{		
			std::cout<<"[WARNING] Corr Value was already loaded: ";
			std::cout<<corrValueName.c_str()<<std::endl;
		}
		alreadyLoadedErr = false;
		return;
	}
	m_NameToCorrValue[corrValueName] = (int)m_CorrValues.size();
	m_CorrValues.push_back(corrValue);
	if (!m_pDB->m_CorrValueNames.contains(corrValueName))
		m_pDB->m_CorrValueNames.push_back(corrValueName);
}

ModelPoint CorrCollection::GetBestCorr(const ModelPoint & pnt, 
									   Model3D * toModel,
									   const VKString & modelValue, 
									   double * value)
{
	assert(false);
}

void CorrCollection::GetCorrsForPairWithVal(Model3D * fromModel,
									Model3D * toModel, 
									std::map<int, double> * savePerPointVals)
{
	assert(false);
}


void CorrCollection::AutoNormalizeCorrs()
{
	m_bAutoRegionNormalize = true;
}

void CorrCollection::SetRegionNormalization(std::vector<ModelPoint> & )
{
	// no normalization for the most
	m_bAutoRegionNormalize = false;
}

bool CorrCollection::ReadCorrValues(const VKString & corrValueName)
{
	assert(false);
	return false;
}
void CorrCollection::WriteCorrValues(const VKString & corrValueName)
{
	// for now only use it to write / read concatenated pairwise corr values
	assert(false);	
}


