/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/


#ifndef __CorrCollection_H
#define __CorrCollection_H

#include "VKString.h"

#include "Model3D.h"
#include "ModelPoint.h"

class CorrValue;
class CollectionPairwise;
class ModelDatabase;
class AnalysisPipeline;

class CorrCollection
{
	public:
		CorrCollection(const VKString & collectionName,
					   ModelDatabase * database, 
					   int corrID);
		virtual ~CorrCollection();
	
		virtual CollectionPairwise * SafeCastToPairwise();
		virtual CorrValue * GetCorrValueByName(const VKString & corrValue);
		virtual void AddCorrValue(const VKString & corrValueName, CorrValue * corrValue);
	
		virtual void AutoNormalizeCorrs();
		virtual void SetRegionNormalization(std::vector<ModelPoint> & pnts);
		virtual void RemoveModel(int modelID);
		
	
	// Implement in a child
		virtual void GetRegionCorrVals(std::vector<ModelPoint> & queryRegion,
									   const VKString & corrValue,
									   std::vector<ModelPoint> * corrPnts, 
									   std::vector<double> * corrVals);
	
		virtual ModelPoint GetBestCorr(const ModelPoint & pnt, 
									   Model3D * toModel, 
									   const VKString & modelValue, 
									   double * value);
		virtual void GetCorrsForPairWithVal(Model3D * fromModel,
											Model3D * toModel, 
											std::map<int, double> * savePerPointVals);
	
		virtual void GetCorrs(const ModelPoint & pnt, 
							  std::vector<ModelPoint> * corrPnts) = 0;
		virtual VKString GetCollectionType() = 0;
		virtual bool ReadCollectionChild(std::ifstream & textStream);
		virtual void WriteCollectionChild(std::ofstream & textStream);
		virtual void DrawSpecial(DatabaseVisWindow * window);
	////////////////
	
		virtual void Draw(DatabaseVisWindow * window);
		virtual void DrawCollectionEdges(DatabaseVisWindow * window);
		virtual void DrawCollectionCorrValue(DatabaseVisWindow * window);
		virtual void DrawCorrespondenceEdge(const ModelPoint & pnt);
	
		static bool ReadAndCreateCollection(const VKString & collectionName,
											ModelDatabase * db, 
											AnalysisPipeline * pipeline);
		static bool ReadAndCreateCollectionSinglePairWithID(const VKString & collectionName,  
															int corrID, ModelDatabase * db,
															std::ifstream & textStream);
		static bool ReadAndCreateCollectionSinglePairWithID(const VKString & collectionName,
															int corrID, ModelDatabase * db);
		virtual void WriteCollection(std::ofstream & textStream);
		virtual void WriteCollection();	
	
		virtual bool ReadCorrValues(const VKString & corrValueName);
		virtual void WriteCorrValues(const VKString & corrValueName);

	protected:
		bool m_bAutoRegionNormalize;
		int m_iCorrID;
		VKString m_sCollectionName;
		ModelDatabase * m_pDB;
		
		std::vector<CorrValue *> m_CorrValues;
		std::map<VKString, int> m_NameToCorrValue;
};



#endif
