/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/


#ifndef __CorrValue_H
#define __CorrValue_H

#include "ModelValue.h"
class CorrValuePairwise;
class ModelDatabase;
class DatabaseVisWindow;
class CorrCollection;

/**
 * CorrValue - correspondence value
 */
class CorrValue 
{
	public:
		enum CorrValueFlags
		{
			CORR_FLAG_NONE = 0,
			CORR_NO_CACHING = 1,
			CORR_CACHE_USING_ARRAY = 2,
			CORR_POINT_CLOUD_ONLY = 4,
			CORR_WRITE_CACHE = 8
		};
	
		CorrValue(ModelDatabase * db, const VKString & corrCollection,
				  int corrID, const VKString & corrValueName, int flags);
		virtual ~CorrValue();
	
		enum IntOperator
		{
			INT_OP_DOUBLE_SUM,
			INT_OP_SUM_OF_MAX,
		};
		virtual void SetIntegratedValue(double value);
		virtual double IntegratedValue(const VKString & domainModel, 
									   const VKString & domainPntSet = "auto",
									   const VKString & rangeModel = "auto", 
									   const VKString & rangeoPntSet = "auto", 
									   IntOperator op = INT_OP_SUM_OF_MAX,
									   bool symmetric = true);
		virtual double Value(const ModelPoint & fromPnt, const ModelPoint & toPnt);

		virtual double GetCachedValue(const ModelPoint & fromPnt, 
									  const ModelPoint & toPnt, bool & valid);
		virtual void SetCacheValue(const ModelPoint & fromPnt, 
								   const ModelPoint & toPnt, double value);
	
		virtual CorrCollection * GetCorrCollection();
		
		virtual CorrValuePairwise * GetSelfAsPairwise();
	
		virtual void Draw(DatabaseVisWindow * window, CorrCollection * collection);
		virtual void FindCorrsForDraw(DatabaseVisWindow * window, CorrCollection * collection, 
									  const ModelPoint & pnt, std::vector<ModelPoint> * pnts);
		virtual void DrawOnFromModel(DatabaseVisWindow * window, CorrCollection * collection);
		virtual void DrawOnTargetModels(const ModelPoint & pnt,
										DatabaseVisWindow * window, CorrCollection * collection);
	
	// implement in a child:	
		virtual VKString CorrValueType();
	
	protected:
		virtual double CalculateValue(const ModelPoint & fromPnt, const ModelPoint & toPnt);
	
		virtual bool ReadCorrValueChild(std::ifstream & textStream);
		virtual void WriteCorrValueChild(std::ofstream & textStream);
	
	/////////////
	
	public:
//		static bool ReadCorrValue(ModelDatabase * db, 
//								  const VKString & corrCollection, int corrID, 
//								  const VKString & corrValueName);
//		virtual void WriteCorrValue();
		static bool ReadCorrValue(std::ifstream & textStream,
								  ModelDatabase * db, 
								  const VKString & corrCollection,
								  int corrID,
								  const VKString & corrValueName);
		virtual void WriteCorrValue(std::ofstream & textStream);
		
		virtual bool ReadCachedValues(std::ifstream & textStream);
		virtual	void WriteCachedValues(std::ofstream & textStream);
		
		virtual bool ReadCachedValuesArbitraryPairs(std::ifstream & textStream);
		virtual	void WriteCachedValuesArbitraryPairs(std::ofstream & textStream);	
		
	
	protected:
		double m_fIntergratedValue;
		virtual double GetCachedValueArbitraryMap(const ModelPoint & fromPnt, const ModelPoint & toPnt, bool & valid);
		virtual void SetCacheValueArbitraryMap(const ModelPoint & fromPnt, const ModelPoint & toPnt, double value);
	
		VKString m_sCorrValueName;
		VKString m_sCorrCollectionName;
		int m_iCorrID;
		ModelDatabase * m_pDB;
		std::map<ModelPoint, std::map<ModelPoint, double> > m_ArbitraryPairValuesMap;
		
		bool IsFlagSet(CorrValueFlags flag);
		int m_iFlags;
	
		bool m_bIntegrateBest;
};

class CorrValueConstant : public CorrValue
{
	public:
		CorrValueConstant(ModelDatabase * db);
		
	protected:
		double CalculateValue(const ModelPoint & fromPnt, const ModelPoint & toPnt);
};

#endif


