/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "ModelDatabase.h"
#include "CorrValueDiffuse.h"
#include "CollectionDiffuse.h"

CorrValueDiffuse::CorrValueDiffuse(ModelDatabase * db, const VKString & diffuseCollection, 
								   const VKString & corrValName)
: CorrValue(db, diffuseCollection, -1, corrValName, CorrValue::CORR_NO_CACHING)
{
	m_pDiffuseCollection = NULL;
}

CorrValueDiffuse::~CorrValueDiffuse()
{
}

VKString CorrValueDiffuse::CorrValueType()
{
	return "CorrValueDiffuse";
}

void CorrValueDiffuse::FillPtrs()
{
	if (m_pDiffuseCollection==NULL)
	{
		CorrCollection * coll = m_pDB->GetCorrCollection(m_sCorrCollectionName);
		assert(coll->GetCollectionType().startsWith("CollectionDiffuse")
			   || coll->GetCollectionType()=="CollectionRadiosityFullMtx");
		m_pDiffuseCollection = (CollectionDiffuse*)coll;
	}	
}


double CorrValueDiffuse::CalculateValue(const ModelPoint & fromPnt, const ModelPoint & toPnt)
{
	FillPtrs();
	return m_pDiffuseCollection->DiffusedCorrValue(fromPnt, toPnt);
}




