/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __CorrValueSmooth_H
#define __CorrValueSmooth_H

#include "CorrValuePairwise.h"
#include "CorrValue.h"

class CorrValueSmooth : public CorrValuePairwise
{
	public:
		CorrValueSmooth(ModelDatabase * db, const VKString & corrCollection, int corrID,
						const VKString & corrValueName, int flags);

		CorrValueSmooth(ModelDatabase * db, const VKString & corrCollectionName, int corrID, 
						const VKString & corrValueName, bool pntCldOnly,
						const VKString & corrCollectionNonSmoothed, int corrIDNonSmoothed,
						const VKString & corrValueNonSmooth, const VKString & distanceMetric, 
						double sigmaDistance, bool negateCorrVal, double sigmaCorrValue);
		virtual ~CorrValueSmooth();
	
	// implement in a child
		virtual VKString CorrValueType();
	
		virtual void AddSmoothedCorrs(const ModelPoint & fromPnt, 
									  std::vector<ModelPoint> & toPnts);
		virtual double GetSmoothedDistance(const ModelPoint & fromPnt, 
										   const ModelPoint & toPnt);
		
	protected:
		virtual bool ReadCorrValueChild(std::ifstream & textStream);
		virtual void WriteCorrValueChild(std::ofstream & textStream);
	/////////
	
	protected:
		std::map<ModelPoint, double > * m_aGaussianWeights;
		int m_iSizeOfWeightsArray;
	
		virtual double CalculateValue(const ModelPoint & fromPnt, const ModelPoint & toPnt);
		virtual void FillPtrs();
	
		Model3D * m_pModelFrom;
		Model3D * m_pModelTo;
		ModelDistance * m_pDistanceOnModelTo;
		CorrValue * m_pCorrValueNonSmooth;
		CollectionSinglePair * m_pCorrs;
		double m_fSigmaDistance2;
		bool m_bNegateCorrVal;	
		double m_fSigmaCorrValue2;
	
		VKString m_sNonSmoothedCollection;
		int m_iNonSmoothedCorrID;
		VKString m_sNonSmoothedCorrValue;
		VKString m_sDistanceMetricName;
};

#endif




