/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/


#ifndef __DatabaseVisWindow_H
#define __DatabaseVisWindow_H

#include <vector>
#include <iostream>
#include <assert.h>

#include "ParamParser.h"
#include "AnalysisPipeline.h"
#include "VisWindow.h"
#include "ModelPoint.h"
#include "ModelPerVertexValues.h"

class Model3DRenderingParams;
class Model3D;

class DatabaseVisWindow : public VisWindow
{
	public:
		enum CorrRenderFlags
		{
			RENDER_CORRS_NONE	= 0,
			RENDER_CORRS_SET	= 1,
			RENDER_CORRS_EDGES	= 2,		// edges between corresponding points
			RENDER_CORRS_PNTS	= 4,		// points colors
			RENDER_CORRS_MVALUE	= 8,		// model value
			RENDER_CORRS_CVALUE	= 16,		// correspondence value
			RENDER_CORRS_SPECIAL_1 = 1024,	// rigid alignment = align models
			RENDER_CORRS_SPECIAL_2 = 2048,	// 
			RENDER_CORRS_SPECIAL_3 = 4096,	// 
			RENDER_CORRS_SPECIAL_4 = 8192	// 
		};
	
		enum RenderDesign
		{
			RENDER_DESIGN_NONE		= 0,
			RENDER_DESIGN_WND		= 1,
			RENDER_DESIGN_CTRL_WND	= 2
		};
	
		enum SpecialUIMode
		{
			UIMode_NONE,
			UIMode_SELECT_WINDOW,
			UIMode_SELECT_MODEL,
			UIMode_SELECT_POINT,
		};
	
		DatabaseVisWindow(ParamParser & visParams, 
						  AnalysisPipeline & pipeline);
	
		virtual void KeyPressed(const char key, int mouseX, int mouseY);
		virtual void MouseButtonEvent(int x, int y, MouseButton mouseButtonID, 
									  MouseButtonState newState);
		virtual void MouseMotionEvent(int x, int y, int dx, int dy);
		virtual void Resize(int w, int h);
	
		virtual void Draw();
		virtual void InitializeGL();	
		virtual void DrawModelListSubwindow();
		
		virtual int RegisterModel(Model3DRenderingParams * modelRenderParams);
		virtual void SetCurrCameraModel(int modelID);
		virtual Model3DRenderingParams * GetCurrCameraModel();
	
		virtual void GetWindowParams(int wndID, double * cX, double * cY, double * scale, 
									 double * dx=NULL, double * dy=NULL);
		virtual void DrawWindow(int wndID);
		virtual int ClickInWindow(int x, int y);
		virtual void ProcessClickInSubwindow(const VKString & wndName,
											 double xWnd, double yWnd, 
											 bool justPressed);
		virtual void ClickInModelList(double xWnd, double yWnd);
	
		virtual bool IsCorrFlagSet(CorrRenderFlags flag);
		virtual void ToggleCorrFlag(CorrRenderFlags flag);
		virtual void SetCorrFlag(CorrRenderFlags flag, bool on);
	
		virtual void SetPendingUpdateToTrue();
		virtual bool PendingCorrValueRenderingUpdate();
		virtual void UpdatedCorrValueRendering();
	
		virtual ModelPoint SelectedPoint();
		virtual VKString CorrValueToRender();
		virtual VKString CollectionToRnder(int * saveCorrID);
		virtual VKString ModelValueToRender(int subwindowID);
		virtual void SetCorrToRender(const VKString & corrCollName, 
									 int corrID, 
									 const VKString & corrValueName);
		virtual void SetDesignRender(bool on);
		virtual int SelectedWindowID();
	
		virtual void FillMostSimilarModels(int firstID, std::vector<int> & ids);

		std::vector<int> m_NewModelOrder;
		int m_iCurrModelTraversalID;
		int m_iPointDescriptorToRender;
	
		enum WriteImgType
		{
			WRITE_SCREENSHOT_NONE,
			WRITE_SCREENSHOT,
			WRITE_SCREENSHOT_ALL,
			WRITE_SCREENSHOT_EACH_AS_MAIN
		};
		void WriteSurfaceValues(const VKString & dirPrefix);
		void WriteImages(const VKString & dirPrefix, WriteImgType tp, 
						 int imgID0 = 0, bool createTimeDir = true, 
						 bool avoidDuplicates = true,
						 VKString * saveDir = NULL, VKStringList * saveImgNames = NULL);
		void WriteCurrentImage(BufferID & bufferID,
							   const VKString & filename, 
							   bool removeWhite,
							   int x0, int y0, int x1, int y1);
		VKString ImageName(int imgID);
		VKString GetWriteToDir(const VKString & prefix, 
							   bool timeSubdir, bool uniqueDir);
		VKString GetTime();
	
		WriteImgType m_WriteImg;
	
		void WriteCameras();
		void ReadCameras();
	
		void DisableApp();
		void EnableApp();

	protected:	
	
		// camera transforms	
		SpecialUIMode m_UIMode;
		int m_iSelectedWindow;

		int m_iCurrentSelection;
	
		AnalysisPipeline & m_Pipeline;
		std::vector<Model3DRenderingParams *> m_pModelsRegisteredForEvents;
		int m_iCurrCamera;
	
		static const int CORRS_SELECTION_ARROWS = 5;
		int m_iSelectedList;
		VKStringList m_sLists;
	
		int m_iCorrRenderFlags;
		VKString m_sSelectedCollectionToRender;
		int m_iCollectionCorrIDToRender;
		bool m_bPendingCorrValueRenderUpdate;
	
		VKString m_sCorrValueToRender;
	
		VKString m_sActiveApp;
		int m_iRenderDesign;
};

#endif

