/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __DistanceGeodesic_H
#define __DistanceGeodesic_H

#include <vector>
#include <map>

#include "ModelDistance.h"

class DistanceGeodesic : public ModelDistance
{
	public:
		struct GeoVertexData 
		{
			R3MeshVertex *vertex;
			double distance;
			RNBoolean selected;
			GeoVertexData **heappointer;
		};	
	
		DistanceGeodesic(Model3D * model, int numCachedRows);
		virtual ~DistanceGeodesic();
	
		virtual double Distance(const ModelPoint & p1, const ModelPoint & p2);
		virtual double Distance(int v1, int v2);
		virtual void GetNeighborhood(const ModelPoint & p1, double threshold, 
									 std::vector<ModelPoint> * trgPnts, 
									 const VKString & pointSet="auto");
		virtual ModelPoint GetClosestPoint(const ModelPoint & p1, 
										   const VKString & pointSet="auto");
		virtual double CalculateStandardNorm();
		virtual VKString DistanceType();

	protected:
		virtual bool IsPrecomputed(int vertexID);
		virtual void PrecomputeRow(int vertexID);
		
		std::vector<int> m_AddedVertices;
		int * m_aVertexIDToRow;
		double ** m_PrecomputedRows;
		int m_iMaxNumCached;
	
		GeoVertexData * m_aPrecompData;
	
	// Static functions: Dijkstra (Tom's implementation)
		static GeoVertexData * InitializeFunkPrecomputation(R3Mesh * mesh, 
															bool verb, 
															GeoVertexData * exists=NULL);	
		static void PrecomputeFillFunkhouser(GeoVertexData * tempVertexData, 
						R3Mesh * mesh, int fromVertex, 
						double * toDistanceArray, 
						const std::vector<int> & toVerticesRemember=std::vector<int>());
		static void FillThresholdedDistances(GeoVertexData * tempVertexData, 
											 R3Mesh * mesh, int fromVertex,
											 std::map<int, double> & verticesWithValues, 
											 double maxDistance);
		static double GetThresholdedDistances(int toVertex, R3Mesh * mesh, 
											  std::map<int, double> & verticesWithValues, 
											  double maxDistance);
};

#endif

