/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __LSFuzzy_H
#define __LSFuzzy_H

#include "LinAlgMatrixSparseReal.h"
#include "VKString.h"

/**
 * Least-square solution for fuzzy correspondence values
 * (note this class is from any model to all other models)
 */ 
class LSFuzzyPerGroup
{
	public:
		LSFuzzyPerGroup(std::vector<int> & grouping, 
						int * globalPntToGroup,
						int groupID, double threshold, 
						double interpWeight);
		~LSFuzzyPerGroup();
		
		void ClearMtxData();
		void InitMtx(std::vector<int> & aij, std::vector<double> & avals);
	
		int FillRow(int pntFrom);
		std::map<int, double> * GetCorrs(int pntFrom);
	
		bool * m_aRowFilled;
		std::map<int, double> ** m_GroupedRowVals;
	
		int m_iMatrixNumAlignEquations;
		LinAlgMatrixSparseReal * m_pLSEquations;
		LinAlgVectorReal * m_pLSEquationsVector;
		LinAlgVectorReal * m_pFuzzyValues;
	
		int m_iGroupOffset;
		double m_fMaxValThreshold;	
		double m_fInterpWeight;	
		int m_iGroupID;
		std::vector<int> & m_Grouping;
		int m_iTotalNPoints;
	
		int * m_aGlobalPntToGroup;
};

/**
 * Least-square solution for fuzzy correspondence values
 */ 
class LSFuzzy
{
	public:
		enum LSFuzzyType
		{
			TP_LS_FUZZY,
			TP_RADIOSITY,
			TP_MATLAB_SCRIPT
		};
		LSFuzzy(LSFuzzyType tp,
				std::vector<int> & grouping);
		~LSFuzzy();
	
		void SetMatlabScript(const VKString & scriptName);
		std::map<int, double> * GetAllCorrs(int pntFrom);
	
		void ClearAll();
		double FuzzyCorr(int pntFrom, int pntTo);
		void FuzzyCorrs(int pntFrom, int ** closestIDs, 
						double ** closestVals, int * nCorrs);
		int BestCorr(int pntFrom, int groupTo, double * value);
		int GetGroup(int pnt);
		
		void AddValue(int p1, int p2, double val);
		void InitGroup(int groupID);
		void Embed(const VKString & prefix, 
				   double * timeInSec);
		void PrecomputeAllFuzzy();
		void FillRow(int pntFrom);
	
		void Write(const char * dir, const char * prefix);
		void Read(const char * dir, const char * prefix);
	
		static void RunMatlab(const VKString & scriptName, 
							  const VKString & argsPrefix );
	protected:
		void WriteAijIntoSparseMatrix(const VKString & sparseMtxStr);
		void WriteGrouping(const VKString & groupingStr);
		void ReadSparseMatrixIntoPrecompCorrs(const VKString & sparseMtxStr);
	
		VKString m_sMatlabScript;
		LSFuzzyType m_Type;
	
		bool m_bComputeAllFuzzy;
		std::vector<int> m_Aij;			// indicies
		std::vector<double> m_Aval;		// values
		
		LSFuzzyPerGroup ** m_FuzzyPerGroup;
		
		int m_iTotalNPoints;
		std::vector<int> m_Grouping;
		int * m_aGlobalPntToGroup;
	
		int m_iCorrSize;
		int * m_aiTempClosestIDs;
		double * m_aiTempClosestVals;
	
		LinAlgMatrixSparseReal * m_pRadiosityFormFactors;
		LinAlgVectorReal * m_pRadiosityEmitted;
		LinAlgVectorReal * m_pRadiosityEnergy;
		
		std::map<int, double> ** m_aPrecomputedCorrs;
		bool * m_aIsRowPrecomputed;
	
};

#endif


