/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifdef OPEN_COLLADA

#include "COLLADAFW.h"
#include "COLLADASaxFWLLoader.h"

#include "VkFunctions.h"

#include "Model3DColladaParser.h"
#include "Model3D.h"

#include "ModelValueLabel.h"

Model3DColladaParser::GeometryIDs::Face::Face(int v1, int v2, int v3)
{
	v[0] = v1;
	v[1] = v2;
	v[2] = v3;
}

int & Model3DColladaParser::GeometryIDs::Face::operator [](int i)
{
	return v[i];
}



Model3DColladaParser::Model3DColladaParser(Model3D * model)
{
	m_pModel = model;
}

void Model3DColladaParser::LoadColladaFile(Model3D * model)
{
	Model3DColladaParser * modelWriter = new Model3DColladaParser(model);
	assert(modelWriter!=NULL);
	COLLADASaxFWL::Loader * colladaLoader = new COLLADASaxFWL::Loader();
	assert(colladaLoader!=NULL);
	COLLADAFW::Root root(colladaLoader, modelWriter);
	bool loaded = root.loadDocument(model->ModelFile().c_str());
	if (!loaded)
		std::cout<<"[WARNING] Collada Parsing Failed"<<std::endl;
	delete modelWriter;
	delete colladaLoader;
	//assert(loaded);
}

void Model3DColladaParser::UnimplementedCall(const VKString & name)
{
	if (!m_UnimplementedStuffCalled.contains(name))
		m_UnimplementedStuffCalled.push_back(name);
}

/** This method will be called if an error in the loading process occurred and the loader cannot
 continue to to load. The writer should undo all operations that have been performed.
 @param errorMessage A message containing informations about the error that occurred.
 */
void Model3DColladaParser::cancel(const COLLADAFW::String& errorMessage)
{
	std::cout<<"[WARNING] Collada parsing failed"<<std::endl;
	std::cout<<errorMessage<<std::endl;
}

/** This is the method called. The writer hast to prepare to receive data.*/
void Model3DColladaParser::start()
{
	m_pModel->m_pMesh = new R3Mesh();
}

/** This method is called after the last write* method. No other methods will be called after this.*/
void Model3DColladaParser::finish()
{
	VKString defaultPrefix = "LabelsFromScene";
	ParseNodeHierarchy();
	
	// NOTE: do NOT delete code below - it should be adjusted to parse labels!
//	int deepestStack = 0;
//	
//	for (std::map<COLLADAFW::UniqueId, GeometryIDs>::iterator iter = m_mIDToGeometry.begin(); 
//		 iter != m_mIDToGeometry.end(); iter++)
//		deepestStack = vkMax(iter->second.m_sNameStack.count(), deepestStack);
//
//	std::vector<ModelValueLabel * > labelings;
//	VKStringList labelingsNames;
//	for (int depth=deepestStack-1; depth>=0; depth--)
//	{
//		std::map<int, VKString> segmentIDToMap;
//		VKString labelsName = defaultPrefix+VKString::number(deepestStack - depth - 1);
//		labelingsNames.push_back(labelsName);
//		ModelValueLabel * labels = new ModelValueLabel(m_pModel, labelsName);
//		for (std::map<COLLADAFW::UniqueId, GeometryIDs>::iterator iter = m_mIDToGeometry.begin(); 
//			 iter != m_mIDToGeometry.end(); iter++)
//		{
//			GeometryIDs & geom = iter->second;
//			VKString label("");
//			int segmentID = -1;
//			
//			if (depth < (int)geom.m_DepthToSegmentID.size())
//				segmentID = geom.m_DepthToSegmentID[depth];
//			
//			if (depth < (int)geom.m_sNameStack.count())
//				label = geom.m_sNameStack[depth];
//
//			if (label!="")
//				labels->AssignLabel(segmentID, label);
//			
//			// iterate through vertices
//			for (int i=0; i<(int)geom.m_Vertices.size(); i++)
//				labels->AssignSegmentID(ModelPoint(geom.m_Vertices[i], m_pModel), segmentID);
//			
//			// iterate through faces
//			for (int i=0; i<(int)geom.m_Faces.size(); i++)
//				labels->AssignSegmentIDToFace(geom.m_Faces[i], segmentID);
//			
//			labelings.push_back(labels);
//		}
//		
//		//m_pModel->m_ModelValuesCameWithExternalFile.push_back(labelsName);
//		//m_pModel->AddModelValue(labelsName, labels);
//	}
//	
//	for (int i=0; i<(int)labelings.size(); i++)
//		labelings[i]->AddAllLabelings(i, labelingsNames);	
//	
//	R3Mesh * newMesh = Model3D::CloneMesh(m_pModel->m_pMesh);
//	delete m_pModel->m_pMesh;
//	m_pModel->m_pMesh = newMesh;
		
//	if (m_UnimplementedStuffCalled.count()>0)
//	{
//		std::cout<<"[WARNING] Parsing COLLADA, unimplemented features: ";
//		for (int i=0; i<m_UnimplementedStuffCalled.count(); i++)
//			std::cout<<m_UnimplementedStuffCalled[i].c_str()<<" ";
//		std::cout<<std::endl;
//    }
}

/** When this method is called, the writer must write the global document asset.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeGlobalAsset ( const COLLADAFW::FileInfo* asset )
{
	UnimplementedCall("FileInfo");
	return true;
}

/** When this method is called, the writer must write the scene.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeScene ( const COLLADAFW::Scene* scene )
{
	// pretty-much pointless - just gives a unique id of VisualScene and KinematicScene
	return true;
}

/** When this method is called, the writer must write the entire visual scene.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeVisualScene ( const COLLADAFW::VisualScene* visualScene )
{
	const COLLADAFW::NodePointerArray & nodes = visualScene->getRootNodes();
	for (int i=0; i<nodes.getCount(); i++)
		m_RootVisualNodes.push_back(AddNode(nodes[i]));
	return true;
}

/** When this method is called, the writer must handle all nodes contained in the 
 library nodes.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeLibraryNodes ( const COLLADAFW::LibraryNodes* libraryNodes )
{
	const COLLADAFW::NodePointerArray & nodes = libraryNodes->getNodes();
	for (int i=0; i<nodes.getCount(); i++)
		m_RootLibraryNodes.push_back(AddNode(nodes[i]));
	return true;
}

COLLADAFW::UniqueId Model3DColladaParser::AddNode(const COLLADAFW::Node * node)
{
	// TODO: transformations of the node?
	COLLADAFW::UniqueId nodeID = node->getUniqueId();
	
	assert(m_mIDToNodeHierarchy.find(nodeID)==m_mIDToNodeHierarchy.end());
	m_mIDToNodeHierarchy[nodeID] = NodeHierarchy();
	
	NodeHierarchy & saveNode = m_mIDToNodeHierarchy[nodeID];
	saveNode.m_sNodeName = node->getName().c_str();
	
	//	std::cout<<"Node name: "<<node->getName()<<" ID="<<nodeID<<std::endl;
	//	std::cout<<"->\t\tNodes="<<node->getChildNodes().getCount()<<std::endl;
	//	std::cout<<" - transformations: "<<node->getTransformations().getCount()<<std::endl;
	//	std::cout<<" - geometries: "<<node->getInstanceGeometries().getCount()<<std::endl;
	//	std::cout<<" - instances: "<<node->getInstanceNodes().getCount()<<std::endl;
	
	COLLADABU::Math::Matrix4 nodeMatrix = node->getTransformationMatrix();
	R4Matrix gapsMatrix;
	for (int i=0; i<4; i++)
		for (int j=0; j<4; j++)
			gapsMatrix[i][j] = nodeMatrix.getElement(i,j);
	saveNode.m_XForm = gapsMatrix;
	
	for (int i=0; i<node->getInstanceNodes().getCount(); i++)
		saveNode.m_NodeIDs.push_back(node->getInstanceNodes()[i]->getInstanciatedObjectId());
		
	for (int i=0; i<node->getInstanceGeometries().getCount(); i++)
		saveNode.m_GeometryIDs.push_back(node->getInstanceGeometries()[i]->getInstanciatedObjectId());
	
	for (int i=0; i<node->getChildNodes().getCount(); i++)
		saveNode.m_ChildrenIDs.push_back(AddNode(node->getChildNodes()[i]));
	
	return nodeID;
}

void Model3DColladaParser::ParseNodeHierarchy()
{
	VKStringList nameStack;
	std::vector<int> depthToSegStack;
	R4Matrix matrixStack = R4identity_matrix;
	for (int i=0; i<(int)m_RootVisualNodes.size(); i++)
		TransformNode(m_RootVisualNodes[i], matrixStack, nameStack, depthToSegStack);
}

void Model3DColladaParser::TransformNode(const COLLADAFW::UniqueId nodeID,
										 const R4Matrix & matrixStack,
										 VKStringList & nameStack,
										 std::vector<int> & depthToSegStack)
{
	std::map<COLLADAFW::UniqueId, NodeHierarchy>::iterator iter = m_mIDToNodeHierarchy.find(nodeID);
	assert(iter!=m_mIDToNodeHierarchy.end());
	NodeHierarchy & node = iter->second;

//	for (int i=0; i<nameStack.count(); i++)
//		std::cout<<"   "<<std::flush;
//	std::cout<<"Node["<<nameStack.count()<<"]: "<<node.m_sNodeName.c_str()<<std::endl;

	nameStack.push_back(node.m_sNodeName);
	
	if (nameStack.count() > (int)depthToSegStack.size())
		depthToSegStack.push_back(0);
	else
		depthToSegStack[nameStack.count()-1]++;
	
	R4Matrix newMtx = matrixStack * node.m_XForm;
	
	for (int i=0; i<(int)node.m_NodeIDs.size(); i++)
		TransformNode(node.m_NodeIDs[i], newMtx, nameStack, depthToSegStack);
	
	for (int i=0; i<(int)node.m_GeometryIDs.size(); i++)
		TransformGeometry(node.m_GeometryIDs[i], newMtx, nameStack, depthToSegStack);
	
	for (int i=0; i<(int)node.m_ChildrenIDs.size(); i++)
		TransformNode(node.m_ChildrenIDs[i], newMtx, nameStack, depthToSegStack);
	
	nameStack.remove(nameStack.count()-1);
}

void Model3DColladaParser::TransformGeometry(const COLLADAFW::UniqueId geometryID,
											 const R4Matrix & matrixStack,
											 VKStringList & nameStack,
											 std::vector<int> & depthToSegStack)
{
	R3Affine gapsXForm(matrixStack);
	std::map<COLLADAFW::UniqueId, GeometryIDs>::iterator iter = m_mIDToGeometry.find(geometryID);
	assert(iter!=m_mIDToGeometry.end());

//	std::cout<<"Transforming "<<m_mIDToGeometry[geometryID].m_sName.c_str()<<std::endl;
//	std::cout<<"mtx="<<std::endl;
//	for (int i=0; i<4; i++)
//	for (int j=0; j<4; j++)
//		std::cout<<matrixStack[i][j]<<(j==3 ? "\n" : " ");
	
	
	iter->second.m_sNameStack = nameStack;
	iter->second.m_DepthToSegmentID = depthToSegStack;

	std::vector<R3MeshVertex*> vts;
	std::vector<R3Point> vPos;
	for (int v=0; v<(int)iter->second.m_Vertices.size(); v++)
	{
		R3Point p = iter->second.m_Vertices[v];
		gapsXForm.Apply(p);
		R3MeshVertex * vtx = m_pModel->m_pMesh->CreateVertex(p);
		vts.push_back(vtx);
		vPos.push_back(p);
		
//		R3MeshVertex * vtx = m_pModel->m_pMesh->Vertex(iter->second.m_Vertices[v]);
//		R3Point p = m_pModel->m_pMesh->VertexPosition(vtx);
//		gapsXForm.Apply(p);
//		m_pModel->m_pMesh->SetVertexPosition(vtx, p);
	}
	
	for (int f=0; f<(int)iter->second.m_Faces.size(); f++)
	{
		int v1=iter->second.m_Faces[f][0];
		int v2=iter->second.m_Faces[f][1];
		int v3=iter->second.m_Faces[f][2];
		if (v1==v2 || v1==v3 || v2==v3)
			continue;	// don't create 
		
		// try to create face
		if (!FaceExists(vts[v1], vts[v2], vts[v3]))
		{
			R3MeshFace * face = m_pModel->m_pMesh->CreateFace(vts[v1], vts[v2], vts[v3]);
			if (face==NULL)	// if face creation failed - revert order?
			{
				if (!FaceExists(vts[v1], vts[v2], vts[v3]))
				{
					face = m_pModel->m_pMesh->CreateFace(vts[v1], vts[v3], vts[v2]);
					if (face==NULL)	// create duplicate vertices (allow triangle soup)
					{
						// create new vertices
						R3MeshVertex * vtx1 = m_pModel->m_pMesh->CreateVertex(vPos[v1]);
						R3MeshVertex * vtx2 = m_pModel->m_pMesh->CreateVertex(vPos[v2]);
						R3MeshVertex * vtx3 = m_pModel->m_pMesh->CreateVertex(vPos[v3]);
						if (!FaceExists(vtx1, vtx2, vtx3))
						{
							CreatedFace(vtx1, vtx2, vtx3);
							face = m_pModel->m_pMesh->CreateFace(vtx1, vtx2, vtx3);				
							assert(face!=NULL);
						}
					}
					else
						CreatedFace(vts[v1], vts[v2], vts[v3]);
				}
			}
			else
				CreatedFace(vts[v1], vts[v2], vts[v3]);
		}
		
//		R3Mesh * msh=m_pModel->m_pMesh;
//		for (int e=0; e<m_pModel->m_pMesh->NEdges(); e++)
//		{
//			R3MeshEdge * edge = msh->Edge(e);
//			bool failed=msh->FaceOnEdge(edge, 0) == m_pModel->m_pMesh->FaceOnEdge(edge, 1);
//			if (failed && msh->FaceOnEdge(edge, 0)!=NULL)
//			{
//				R3MeshFace * face = msh->FaceOnEdge(edge, 0);
//				std::cout<<"Adding Face: "<<f<<" Triangle: "<<v1<<" "<<v2<<" "<<v3<<std::endl;
//				std::cout<<"\tBroken Face="<<msh->FaceID(face)<<std::endl;
//				std::cout<<"\tBroken Vertices="<<msh->VertexID(msh->VertexOnFace(face, 0))<<std::flush;
//				std::cout<<" "<<msh->VertexID(msh->VertexOnFace(face, 1))<<std::flush;
//				std::cout<<" "<<msh->VertexID(msh->VertexOnFace(face, 2))<<std::endl;
//				assert(false);
//			}
//		}		
	}
}

bool Model3DColladaParser::FaceExists(R3MeshVertex * v1, R3MeshVertex * v2, R3MeshVertex * v3)
{
	std::map<R3MeshVertex * , std::map<R3MeshVertex*, std::map<R3MeshVertex*, bool> > >::iterator iter1;
	iter1 = m_CreatedFaces.find(v1);
	
	if (iter1!=m_CreatedFaces.end())
	{
		std::map<R3MeshVertex*, std::map<R3MeshVertex*, bool> >::iterator iter2;
		iter2 = iter1->second.find(v2);
		if (iter2!=iter1->second.end())
		{
			std::map<R3MeshVertex*, bool>::iterator iter3;
			iter3 = iter2->second.find(v3);
			return iter3!=iter2->second.end();
		}
	}
	return false;
}

void Model3DColladaParser::CreatedFace(R3MeshVertex * v1, R3MeshVertex * v2, R3MeshVertex * v3)
{
	m_CreatedFaces[v1][v2][v3] = true;
	m_CreatedFaces[v1][v3][v2] = true;
	m_CreatedFaces[v2][v1][v3] = true;
	m_CreatedFaces[v2][v3][v1] = true;
	m_CreatedFaces[v3][v1][v2] = true;
	m_CreatedFaces[v3][v2][v1] = true;
}

/** When this method is called, the writer must write the geometry.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeGeometry ( const COLLADAFW::Geometry* geometry )
{
	m_mIDToGeometry[geometry->getUniqueId()] = GeometryIDs();
	
	GeometryIDs & geomIDs = m_mIDToGeometry[geometry->getUniqueId()];
	geomIDs.m_sName = geometry->getName().c_str();

	COLLADAFW::Geometry::GeometryType gType = geometry->getType();
	if (gType==COLLADAFW::Geometry::GEO_TYPE_MESH)
	{
		// TODO: deduce label
		COLLADAFW::Mesh * mesh = (COLLADAFW::Mesh *) geometry;
		COLLADAFW::MeshVertexData & vtxPos = mesh->getPositions();
		COLLADAFW::DoubleArray * dblArray = vtxPos.getDoubleValues();
		COLLADAFW::FloatArray * fltArray = vtxPos.getFloatValues();
		assert(dblArray!=NULL || fltArray!=NULL);
		int NVert = vtxPos.getValuesCount();
		assert(NVert%3==0);
		NVert = NVert / 3;

		for (int i=0; i<NVert; i++)
		{
			R3Point p;
			if (dblArray!=NULL)
				p = R3Point((*dblArray)[i*3+0], (*dblArray)[i*3+1], (*dblArray)[i*3+2]);
			else
				p = R3Point((*fltArray)[i*3+0], (*fltArray)[i*3+1], (*fltArray)[i*3+2]);

			geomIDs.m_Vertices.push_back(p);
		}
		
		COLLADAFW::MeshPrimitiveArray & primitives = mesh->getMeshPrimitives();
		
		int NFaces = mesh->getFacesCount();		

		for (int i=0; i<primitives.getCount(); i++)
		{
			COLLADAFW::MeshPrimitive * primitive = primitives[i];
			COLLADAFW::MeshPrimitive::PrimitiveType tp = primitive->getPrimitiveType();
			if (tp==COLLADAFW::MeshPrimitive::TRIANGLES)
			{
				int NFaces = primitive->getFaceCount();
				COLLADAFW::UIntValuesArray & indArray = primitive->getPositionIndices();
				assert(indArray.getCount()==NFaces*3);
				for (int f=0; f<NFaces; f++)
				{
					int v1loc = indArray[3*f+0];
					int v2loc = indArray[3*f+1];
					int v3loc = indArray[3*f+2];
					geomIDs.m_Faces.push_back(GeometryIDs::Face(v1loc, v2loc, v3loc));
				}
			}
			else 
				UnimplementedCall(VKString("NonTriMesh_")+VKString::number((int)tp));
		}
	}
	else 
		std::cout<<"[WARNING] Unknown geometry: "<<gType<<std::endl;
	//	UnimplementedCall(VKString("UnknownGeometry_")+VKString::number((int)gType));
	
	return true;
}

/** When this method is called, the writer must write the material.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeMaterial( const COLLADAFW::Material* material )
{
	UnimplementedCall("Material");	
	return true;
}

/** When this method is called, the writer must write the effect.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeEffect( const COLLADAFW::Effect* effect )
{
	UnimplementedCall("Effect");
	return true;
}

/** When this method is called, the writer must write the camera.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeCamera( const COLLADAFW::Camera* camera )
{
	UnimplementedCall("Camera");
	return true;
}

/** When this method is called, the writer must write the image.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeImage( const COLLADAFW::Image* image )
{
	UnimplementedCall("Image");
	return true;
}

/** When this method is called, the writer must write the light.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeLight( const COLLADAFW::Light* light )
{
	UnimplementedCall("Light");
	return true;
}

/** When this method is called, the writer must write the Animation.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeAnimation( const COLLADAFW::Animation* animation )
{
	UnimplementedCall("Animation");
	return true;
}

/** When this method is called, the writer must write the AnimationList.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeAnimationList( const COLLADAFW::AnimationList* animationList )
{
	UnimplementedCall("AnimationList");
	return true;
}

/** When this method is called, the writer must write the skin controller data.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeSkinControllerData( const COLLADAFW::SkinControllerData* skinControllerData )
{
	UnimplementedCall("SkinControllerData");
	return true;
}

/** When this method is called, the writer must write the controller.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeController( const COLLADAFW::Controller* controller )
{
	UnimplementedCall("Controller");
	return true;
}

/** When this method is called, the writer must write the formulas. All the formulas of the entire
 COLLADA file are contained in @a formulas.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeFormulas( const COLLADAFW::Formulas* formulas )
{
	UnimplementedCall("Formulas");
	return true;
}

/** When this method is called, the writer must write the kinematics scene. 
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeKinematicsScene( const COLLADAFW::KinematicsScene* kinematicsScene )
{
	UnimplementedCall("KinematicsScene");
	return true;
}

#endif

