/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifdef OPEN_COLLADA
#ifndef __Model3DColladaParser_H
#define __Model3DColladaParser_H

#include <vector>
#include <iostream>
#include <assert.h>

#include "gaps.h"

#include "COLLADAFW.h"
#include "COLLADASaxFWLLoader.h"

#include "VKString.h"
#include "ParamParser.h"

class Model3D;
class ModelDatabase;


class Model3DColladaParser : public COLLADAFW::IWriter
{
	public:
		class NodeHierarchy
		{
			public:
				VKString m_sNodeName;
				
				COLLADAFW::UniqueId m_NodeID;
							
				R4Matrix m_XForm;
				
				std::vector<COLLADAFW::UniqueId> m_NodeIDs;
				std::vector<COLLADAFW::UniqueId> m_GeometryIDs;
				std::vector<COLLADAFW::UniqueId> m_ChildrenIDs;
		};
		
		class GeometryIDs
		{
			public:
				struct Face
				{
					Face(int v1, int v2, int v3);
					int & operator [](int i);
					int v[3];
				};
			
				std::vector<R3Point> m_Vertices;
				std::vector<Face> m_Faces;
				VKString m_sName;
				VKStringList m_sNameStack;
				std::vector<int> m_DepthToSegmentID;
		};
		
		Model3DColladaParser(Model3D * model);
	
		static void LoadColladaFile(Model3D * model);
	
		
		Model3D * m_pModel;
	
		VKStringList m_UnimplementedStuffCalled;
		void UnimplementedCall(const VKString & name);
	
		virtual COLLADAFW::UniqueId AddNode(const COLLADAFW::Node * node);
		virtual void TransformNode(const COLLADAFW::UniqueId nodeID,
								   const R4Matrix & matrixStack,
								   VKStringList & nameStack, 
								   std::vector<int> & depthToSegStack);
		virtual void TransformGeometry(const COLLADAFW::UniqueId geometryID,
									   const R4Matrix & matrixStack,
									   VKStringList & nameStack,
									   std::vector<int> & depthToSegStack);
	
		virtual void ParseNodeHierarchy();

		std::map<COLLADAFW::UniqueId, GeometryIDs> m_mIDToGeometry;
		std::map<COLLADAFW::UniqueId, NodeHierarchy> m_mIDToNodeHierarchy;
	
		std::vector<COLLADAFW::UniqueId> m_RootVisualNodes;
		std::vector<COLLADAFW::UniqueId> m_RootLibraryNodes;
	
	// inherited:
		virtual void cancel(const COLLADAFW::String& errorMessage);		
		virtual void start();
		virtual void finish();
		virtual bool writeGlobalAsset ( const COLLADAFW::FileInfo* asset );		
		virtual bool writeScene ( const COLLADAFW::Scene* scene );		
		virtual bool writeVisualScene ( const COLLADAFW::VisualScene* visualScene );
		virtual bool writeLibraryNodes ( const COLLADAFW::LibraryNodes* libraryNodes );
		virtual bool writeGeometry ( const COLLADAFW::Geometry* geometry );
		virtual bool writeMaterial( const COLLADAFW::Material* material );
		virtual bool writeEffect( const COLLADAFW::Effect* effect );
		virtual bool writeCamera( const COLLADAFW::Camera* camera );
		virtual bool writeImage( const COLLADAFW::Image* image );
		virtual bool writeLight( const COLLADAFW::Light* light );
		virtual bool writeAnimation( const COLLADAFW::Animation* animation );
		virtual bool writeAnimationList( const COLLADAFW::AnimationList* animationList );
		virtual bool writeSkinControllerData( const COLLADAFW::SkinControllerData* skinControllerData );
		virtual bool writeController( const COLLADAFW::Controller* controller );
		virtual bool writeFormulas( const COLLADAFW::Formulas* formulas );
		virtual bool writeKinematicsScene( const COLLADAFW::KinematicsScene* kinematicsScene );	
	
		bool FaceExists(R3MeshVertex * v1, R3MeshVertex * v2, R3MeshVertex * v3);
		void CreatedFace(R3MeshVertex * v1, R3MeshVertex * v2, R3MeshVertex * v3);
		std::map<R3MeshVertex * , std::map<R3MeshVertex*, std::map<R3MeshVertex*, bool> > > m_CreatedFaces;
};

#endif
#endif
