/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __ModelPerVertexValues_H
#define __ModelPerVertexValues_H

#include <map>

#include "VKString.h"

#include "ModelPoint.h"



class Model3D;
class ModelDistance;
class VisWindow;
class CollectionIdentityPointsets;

//class ModelValuesInterpolatorDescriptor
//{
//	public:
//		enum ModelValueInterpolatorTypes
//		{
//			MODEL_INTERPOLATE_INVALIDATE,
//			MODEL_INTERPOLATE_SETTO0,
//			MODEL_INTERPOLATE_NEAREST_NEIGHBOR,
//			MODEL_INTERPOLATE_GAUSSIAN 
//		};
//	
//		ModelValuesInterpolatorDescriptor();
//		ModelValuesInterpolatorDescriptor(const VKString & distanceMetric, double sigma);
//		
//		ModelValueInterpolatorTypes m_Interpolate;
//		VKString m_sDistanceMetricName;
//		double m_fSigma;
//	
//		VKString m_sInterpolateFromSet;
//		VKString m_sInterpolateToSet;
//};

/**
 * This class caches per vertex values. 
 * It can interpolate, smooth, etc...
 */
class ModelPerVertexValues
{
	public:
		ModelPerVertexValues(Model3D * model, bool pntCloudOnly, bool allocateAsNeededOnly);
		~ModelPerVertexValues();

		ModelPerVertexValues * CreateCopy() const;
		void CopyTo(ModelPerVertexValues * target) const;
		void TakeMaxValues(ModelPerVertexValues * anotherVals);
	
		void ClearValues();
		void SetVertexValue(int vertexID, double value);
		void SetVertexValue(const ModelPoint & atVertexSample, double value);
		double GetValueAtVertex(int vertexID);
		double GetInterpolatedValue(const ModelPoint & pnt);
		ModelPoint NextAtVertexSampleRequiredForInterpolation(const ModelPoint & pnt);
		int NextVertexRequiredForInterpolation(const ModelPoint & pnt);
		
		void SetValues(const VKString & pointSet, double val );
		
	//	void SmoothSimpleLaplacian(int numIterations);
	//	void SmoothGaussianFromKnown(double sigma);	
	//	void SmoothGaussianFromKnown(const VKString & knownSet, double sigma);
	//	void SmoothGaussianFromKnown(ModelPoint * knownSet, double sigma);
	//	
	//	bool SmoothGaussianFromKnownBySplatting(SurfaceSampleSet * knownSet, double sigma, double splatRadius);
	//	void SmoothGaussianFromKnownExact(SurfaceSampleSet * knownSet, double sigma);
	//	
	//	void SmoothPickNearestNeighborsForUnassigned();
		
	//	bool GetPerVertexValues(std::vector<double> & saveValues, double setMin = -FLT_MAX, 
	//							double setMax = FLT_MAX, bool maxToOne=true);
		
	//	void LocalExtrema(SurfaceSampleSet * minSet, SurfaceSampleSet * maxSet, double minGeodesic);
		
	//	void CutToMaximum(double maxValue);
	//	void CutToMinimum(double minValue);	
		
		double MinValue();
		double MaxValue();
		
	//	void WritePerVertexValuesInArff(const VKString & filename,
	//									const VKString & relation, 
	//									const VKString & attributeName);
	//	void WritePerVertexValues(std::ofstream & textStream);
	//	bool LoadPerVertexValues(std::ifstream & textStream);
		
		bool AllValuesSet();
		
	//	DistanceOnTheFly * GetSurfaceDistance(double maxDistance=-1, 
	//										  const VKString & checkDistance="default");
		
		
	//	static double GetPatchArea(R3Mesh * mesh, R3MeshVertex * vertex);
	//	static double GetPatchArea(R3Mesh * mesh, int vertexID);
		
	//	void Interpolate(const ModelValuesInterpolatorDescriptor & interpolate);
		void Interpolate(CollectionIdentityPointsets * interpolator);
		void Interpolate(const VKString & interpolatorCollection="auto");
	
		void GetAllValues(std::vector<double> & vals);
		bool InitColorsPercentileRange(double bottomPercentile, double topPercentile, bool isocontours,
									   VisWindow * window);
		bool InitColorsValueRange(double minValue, double maxValue, bool isocontours,
								  VisWindow * window);

		bool InitColors(VisWindow * window);
//		bool InitRenderingColors(ModelValuesInterpolatorDescriptor & interpolate, 
//								 const VKString & collectionName, int corrID, 
//								 const VKString & corrValueName, const ModelPoint & selectedFrom, 
//								 double bottomPercentile, double topPercentile, bool isocontours,
//								 VisWindow * window);
		
		//void InitMeshValues(const VKString & distanceMetricName, const VKString & modelValue,
		//					double bottomPercentile, double topPercentile, bool isocontours);
		
		void Draw(VisWindow * window);
	
		void ReadValues(std::ifstream & textStream);
		void WriteValues(std::ofstream & textStream);
	
		virtual void FillColor(int vertexID, GLfloat *material);
	
	protected:
		// rendering stuff
		ModelPerVertexValues * m_pRenderR;
		ModelPerVertexValues * m_pRenderG;
		ModelPerVertexValues * m_pRenderB;
		std::vector<int> m_EdgesIsoContours;
	
		// DATA
		int NEntries() const;
		void AllocatePerVertexStorage();
		void AllocatePerVertexStorageFillFromKnown();

	//	struct VertexData 
	//	{
	//		R3MeshVertex *vertex;
	//		double distance;
	//		RNBoolean selected;
	//		VertexData **heappointer;
	//	};
	//	
	//	double ReestimateSigma(double sigma, int numSamples);
	//	double EstimateRadiusForSplatting(int numSamples);
		
		Model3D * m_pModel;
		
		double m_fMinValue;
		double m_fMaxValue;
		
		bool m_bPointCloudOnly;
		
		bool m_bAllVertexValuesAllocated;
		
		double * m_PerVertexValues;
		bool * m_PerVertexValuesExist;
		
		int m_iNumValuesNotSet;
		
		std::map<int, double> m_VertexToValueMap;
		
	//	static void IsExtrema(R3Mesh * mesh, VertexData * vertex_data, bool & isMin, bool &isMax, 
	//						  R3MeshVertex * vOrg, double ring);
	//	
	//	static double * GetKnownValueArray(int id);
	//	static double * GetNormalizationArray(int id);	
	//	static std::map<int, double *> s_KnownValuesArrays;
	//	static std::map<int, double *> s_NormalizationArrays;
		
	//	struct PrecomputedWeights
	//	{	
	//		// note pointers to a row of data is stored
	//		void GetSampleWeighting(int vID, double ** savedWeights, 
	//								int ** savedIDs, int * savedNumPnts);
	//		double m_fSigma;
	//		VKString m_sSampleSet;
	//		double ** m_WeightArray;
	//		int ** m_SampleIDs;
	//		int * m_NumPointsInNhd;
	//		SampledSurface * m_pSurface;
	//	};
		// surface 
	//	static void ClearPrecomputedWeights();
	//	static PrecomputedWeights * GetInterpolationWeighting(const VKString & sampleSetName,
	//														  SampledSurface * surf, double sigma);
	//	static std::map<SampledSurface *, std::vector<PrecomputedWeights *> > s_InterpolationWeights;
};

#endif

