/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __ModelPoint_H
#define __ModelPoint_H

#include <vector>
#include "gaps.h"

class Model3D;
class ModelDatabase;
class PointSet;

/** 
 * Point on a 3d model (mesh / triangle soup / point cloud)
 */
class ModelPoint
{
	public:
		friend class PointSet;
	
		enum PointType
		{
			PNT_TP_INVALID,
			PNT_TP_ON_MESH_SURFACE,
			PNT_TP_FLATTENED
		};
		
		ModelPoint();
		ModelPoint(const ModelPoint & other);
	
	// from R3Mesh (triangle soup / point cloud)
		ModelPoint(int triangleID, double b1, double b2, double b3, Model3D * model);
		ModelPoint(int vertexID, Model3D * model, int cloudID = -1);
	
	// from flat mesh (no model)
		ModelPoint(int triangleID, double b1, double b2, double b3, R3Mesh * flatMesh, Model3D * model);
		
		virtual ~ModelPoint();
	
		virtual void InvalidateAllFields();
		virtual void InitializeTriangle(int triangleID, double b1, double b2, double b3);
		virtual void InitializeVertex(int vertID);
	
		bool operator == (const ModelPoint & other) const;
		void operator = (const ModelPoint & other);
		bool operator < (const ModelPoint & other) const;
	
		bool Invalid() const;
		int TriID() const;
		double B(int coordID) const;
		int VertID(int coordID) const;
			
		void SetPointCloudID(int id);
		int PntCloudID() const;
	
		int NearestVertex(bool * atVertex=NULL) const;
		double Interpolate(double b1, double b2, double b3) const;
		
		virtual const Model3D * GetModel() const;
		virtual bool SameModel(const ModelPoint & pnt) const;
		
		virtual R3Point GetPosition() const;
		virtual R3Vector Normal() const;
		
		virtual PointType GetPointType() const;
	
		virtual void WritePoint(std::ofstream & textStream) const;
		virtual void ReadPoint(std::ifstream & textStream, 
							   ModelDatabase * db,
							   Model3D * model = NULL);
	
	protected:
		void CalculatePosition();
	
		R3Point m_P;
		
		int m_iPntCloudID;
	
		int m_iVertID;
	
		int m_iTri;
		double m_afB[3];

		Model3D * m_pModel;
		R3Mesh * m_pMesh;
		PointType m_PointTp;
};

#endif

