/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __PointDescriptorNormals_H
#define __PointDescriptorNormals_H

#include "Sortable.h"
#include "gaps.h"

#include "PointSet.h"
#include "ModelPoint.h"
#include "PointDescriptor.h"

class PointDescriptorNormals : public PointDescriptor
{
	public:
		enum NormalDescrType
		{
			NDESCR_AVE_NORMAL,
			NDESCR_HISTOGRAM,
		};
		PointDescriptorNormals();
		PointDescriptorNormals(int histBins1, int histBins2);	
		virtual ~PointDescriptorNormals();
		virtual void AddNorm(const R3Vector & normal, double weight, 
							 double * gaussSplat=NULL, int splatNX=-1, int splatNY=-1);
		virtual double MeasureSimilarity(const PointDescriptor & compareToMe);
		virtual double MeasureSimilarityGivenTransform(const PointDescriptor & p2CompareToMe, 
													   const R3Affine & alignP2ToMe);
		virtual double MeasureSimilarity(const PointDescriptorNormals & compareToMe,
										 const R3Affine & alignP2ToMe);


		virtual void Normalize();
		virtual void Draw(const ModelPoint & pnt, DatabaseVisWindow * window, Model3D * model);

		static void CreateNormalHistogramPntDescriptors(const VKString & descrName, 
														int histBins1, int histBins2, 
														double sigma1, double sigma2,
														ModelDatabase * db, const VKString & modelName);
		static void CreateAveNormalPntDescriptor(const VKString & descrName, 
												 ModelDatabase * db, const VKString & modelName);
	
	protected:
		NormalDescrType m_Type;
		int m_iHistBins1;
		int m_iHistBins2;
	
		R3Vector m_AveNormal;
		double * m_aHistogram;
		double m_fWeights;
};



#endif


