/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/


#ifndef __SinglePairPnt2Pnt_H
#define __SinglePairPnt2Pnt_H

#include <vector>
#include <map>

#include "gaps.h"
#include "VKString.h"

#include "ModelPoint.h"
#include "CollectionSinglePair.h"

class SinglePairPnt2Pnt : public CollectionSinglePair
{
	public:
		enum PntsType
		{
			PNT_CLD,
			PNT_VTX
		};
		SinglePairPnt2Pnt(const VKString & collectionName, ModelDatabase * db, int corrID);
		SinglePairPnt2Pnt(const VKString & modelFrom, const VKString & modelTo,
						  const VKString & collectionName, ModelDatabase * db, int corrID,
						  PntsType tp = PNT_CLD);
		virtual ~SinglePairPnt2Pnt();
		virtual void InitData();
	
		virtual ModelPoint GetBestCorr(const ModelPoint & pnt, 
									   Model3D * toModel, 
									   const VKString & modelValue, 
									   double * value);

		virtual void GetCorrs(const ModelPoint & pnt, 
							  std::vector<ModelPoint> * corrPnts);
	
		virtual void AddCorr(const ModelPoint & p1, 
							 const ModelPoint & p2, double value);
		virtual void AddCorr(int v1ID, int v2ID, double value);
		
		virtual bool ReadVtx2VtxMap(const VKString & filename);
		virtual bool ReadFuzzyMap(FILE * fileStream);
		static int AddBIMCollection(VKStringList &pairs,
									const VKString & collectionName, 
									const VKString & corrValueName,
									ModelDatabase * db, 
									double maxDist,
									int parallel,
									std::vector<std::vector<int> > * saveMaps = NULL);
	
		virtual VKString GetCollectionType();
		virtual bool ReadCollectionChild(std::ifstream & textStream);
		virtual void WriteCollectionChild(std::ofstream & textStream);
		virtual void InitCorrValue(const VKString & corrValName,
								   double globalValue);
	
	protected:
		PntsType m_PntType;
	
		int ** m_aCorrs;
		double ** m_aCorrVals;
		int * m_aNCorrs;
		double m_fGlobalValue;
		
		static bool CheckIfCanBIM(Model3D * model);
};


#endif

