/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/


#ifndef __SinglePairRigidAlign_H
#define __SinglePairRigidAlign_H

#include <map>

#include "gaps.h"
#include "VKString.h"

#include "CollectionSinglePair.h"

class SinglePairRigidAlign : public CollectionSinglePair
{
	public:
		friend class CorrValueAlignDist;
	
		enum InitializationType
		{
			RIGID_ALIGN_INIT_PCA,
			RIGID_ALIGN_INIT_FEATURE_CORRS
		};

		SinglePairRigidAlign(const VKString & collectionName, ModelDatabase * db, int corrID);
		SinglePairRigidAlign(const VKString & modelFrom, const VKString & modelTo, double maxDist,
							 const VKString & collectionName, ModelDatabase * db, int corrID,
							 bool makeIsotropic);

		virtual ~SinglePairRigidAlign();
	
		virtual ModelPoint GetBestCorr(const ModelPoint & pnt, 
									   Model3D * toModel, 
									   const VKString & modelValue, 
									   double * value);
		
		
		virtual void DrawSpecial(DatabaseVisWindow * window);
		virtual void DrawFromMesh(DatabaseVisWindow * window);
		virtual void DrawAlignedToMesh(DatabaseVisWindow * window);
	
		virtual void InitFromTransform(const R3Affine & affine12);
	
		static int AddMapsToCollectionPCA(const VKString & modelFrom, 
										  const VKString & modelTo,
										  const VKString & collectionName, 
										  ModelDatabase * db, 
										  bool pca_translation, 
										  bool pca_rotation, 
										  bool pca_scale,
										  bool icp_translation, 
										  bool icp_rotation, 
										  bool icp_scale,
										  int selectK, 
										  const VKString & pointSetName, 
										  double maxDist,
										  R3Vector * upVec,
										  bool anisotropic);

		static int AddMapsToCollectionFromDiffused(const VKString & modelFrom, 
												   const VKString & modelTo,
												   const VKString & collectionName, 
												   ModelDatabase * db, 
												   const VKString & diffusionGuessMap,
												   const VKString & diffusionModelValue,
												   double maxDist, bool anisotropic,
												   R3Vector * upVec,
												   int replaceID,
												   bool keepHistory);
	
		static void FindCorrs(ModelDatabase * db,
							  const VKString & modelFrom, 
							  const VKString & modelTo,
							  const VKString & diffusionGuessMap, 
							  const VKString & diffusionModelValue,
							  std::vector<PointWithID> * savePnts1, 
							  std::vector<PointWithID> * savePnts2,
							  std::vector<double> * saveWeights, 
							  bool takeAll = true,
							  std::vector<int> * selectedPntsFrom = NULL);
		
		static R4Matrix BestRigidAlignment(std::vector<PointWithID> & pntsFrom, 
										   std::vector<PointWithID> & pntsTo,
										   std::vector<double> & weights, 
										   bool trans, bool rot, bool scale,
										   R3Vector * upVec);
	
		virtual void GetCorrs(const ModelPoint & pnt, std::vector<ModelPoint> * corrPnts);
	
		virtual const R3Affine & GetTransform12();
		virtual R3Point Transform12(const R3Point & p1);
		virtual VKString GetCollectionType();
	
		virtual bool ReadCollectionChild(std::ifstream & textStream);
		virtual void WriteCollectionChild(std::ofstream & textStream);
	
		virtual double GetMaxDistThreshold() const;
		virtual double GetMaxDistThresholdOrigModel() const;
		virtual void SetMaxDistThreshold(double threshold);
	
		static R4Matrix InverseMatrix(const R4Matrix & affine);
		static R3Affine InverseAffine(const R3Affine & affine);
	
		static void PruneSimilar(std::vector<R3Affine> & affineTransforms12,
								 std::vector<Sortable> & affineTransformScores,
								 PointSet * pntsTest, double threshold);
	
//	protected:
		double m_fDistScale;
		double m_fMaxDist;
		bool m_bMakeIsotropic;
	
		static double RMSD(Model3D * modelFrom, Model3D * modelTo,
						   const R3Affine & affine12, const R3Affine & affine21, 
						   const VKString & pntSetName, bool anisotropic);

		R3Affine m_Affine12;
		static R4Matrix R3AlignPointsInXYPlane(int npoints, R3Point *points1, R3Point *points2, 
											   RNScalar *weights, RNBoolean align_center, 
											   RNBoolean align_rotation, int align_scale);
		static R4Matrix R3AlignPointsInXYPlane(const RNArray<R3Point *>& points1, 
											   const RNArray<R3Point *>& points2, 
											   RNScalar* weights = NULL, 
											   RNBoolean align_center = TRUE, 
											   RNBoolean align_rotation = TRUE, 
											   int align_scale = FALSE);

		static bool RefineWithICP(Model3D * modelFrom, Model3D * modelTo,
								  R3Affine & affine12, 
								  R3Affine & affine21, 
								  bool icp_translation, bool icp_rotation, bool icp_scale,
								  const VKString & pointSetName,
								  bool anisotropic,
								  R3Vector * upVector);
		static int CreatePointPointCorrespondences(Model3D * modelFrom, Model3D * modelTo,
												   const VKString & pointSetName,
												   bool anisotropic,
												   const R3Affine& affine12, 
												   const R3Affine& affine21,
												   R3Point *correspondences1, 
												   R3Point *correspondences2, 
												   int max_correspondences);
		static R3Point * m_CorrBuffer1;
		static R3Point * m_CorrBuffer2;
		static int m_iCorrBufferSize;
};


#endif

